// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package common;

import java.util.*;
import java.util.concurrent.*;

/**
 * Pool of worker threads. One thread is created per processor.
 * Submitted instances of Callable are executed one by one in the order they are submitted.
 * Do not submit new Callable from within a thread that is itself runnning in one o the worker threads,
 * since this is with a very high probability causing deadlocks.
 * Basically this class is a wrapper for the ScheduledThreadPoolExecutor-Class.
 * Note: In a future migration to Java7, this should be replaced by the more capable JoinForkPool-Class
 */
public class ThreadPool {

	private static ScheduledThreadPoolExecutor threadPool, secondLevelThreadPool;
	private static int processors = Runtime.getRuntime().availableProcessors();
	
	static {
		ThreadFactory tf = new ThreadFactory() {
			@Override
			public Thread newThread(Runnable r) {
				Thread t = new Thread(r);
				t.setPriority(Thread.NORM_PRIORITY);
				return t;
			}
		};
		threadPool = new ScheduledThreadPoolExecutor(Runtime.getRuntime().availableProcessors(), tf);
		
		ThreadFactory tf2 = new ThreadFactory() {
			@Override
			public Thread newThread(Runnable r) {
				Thread t = new Thread(r);
				t.setPriority(Thread.MIN_PRIORITY);
				return t;
			}
		};
		secondLevelThreadPool = new ScheduledThreadPoolExecutor(Runtime.getRuntime().availableProcessors(), tf2);
	}
	
	/**
	 * Runs a batch of Callables instances in parallel.
	 * The method returns once all callables have been executed
	 * @param <V> The returned type of each Callable instance. If no return is required use Void 
	 * @param c Vector of callables
	 * @return Vector of Future objects, holding the returned values
	 */
	public static <V> Vector<Future<V>> executeParallel(Vector<? extends Callable<V>> c){
		Queue<Future<V>> runningTasks = new LinkedList<Future<V>>();
		Vector<Future<V>> allTasks = new Vector<Future<V>>();
		
		for (Callable<V> ca : c){
			Future<V> f = threadPool.submit(ca);
			runningTasks.add(f);
			allTasks.add(f);
		}
		
		while (!runningTasks.isEmpty()){
			Future<V> f = runningTasks.poll();
			try {
				f.get();
			} catch (InterruptedException e) {
				for (Future<V> t : allTasks)
					t.cancel(true);
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		
		return allTasks;
	}
	
	/**
	 * Submit a single callable that is executed asynchronously,
	 * whenever the ThreadPool has free resources 
	 * @param c
	 * @return
	 */
	public static <V> Future<V> submit(Callable<V> c){		
		Future<V> f = threadPool.submit(c);

		return f;
	}
	
	/**
	 * Runs a batch of Callables instances in parallel.
	 * The method returns once all callables have been executed
	 * The Callables are exectuted with lower priority than the primary ThreadPool.
	 * It should be used to improve the level of parallelism within threads that run in parallel themself
	 * @param <V> The returned type of each Callable instance. If no return is required use Void 
	 * @param c Vector of callables
	 * @return Vector of Future objects, holding the returned values
	 */
	public static <V> Vector<Future<V>> executeParallelSecondLevel(Vector<? extends Callable<V>> c){
		Queue<Future<V>> runningTasks = new LinkedList<Future<V>>();
		Vector<Future<V>> allTasks = new Vector<Future<V>>();
		
		for (Callable<V> ca : c){
			Future<V> f = secondLevelThreadPool.submit(ca);
			runningTasks.add(f);
			allTasks.add(f);
		}
		
		while (!runningTasks.isEmpty()){
			Future<V> f = runningTasks.poll();
			try {
				f.get();
			} catch (InterruptedException e) {
				for (Future<V> t : allTasks)
					t.cancel(true);
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		
		return allTasks;
	}
	
	/**
	 * Submit a single callable that is executed asynchronously,
	 * whenever the second level ThreadPool has free resources 
	 * @param c
	 * @return
	 */
	public static <V> Future<V> submitSecondLevel(Callable<V> c){		
		Future<V> f = secondLevelThreadPool.submit(c);

		return f;
	}
	
	public static int availProcessors(){
		return processors;
	}
}
