// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package common;

import java.util.ArrayList;

import javax.media.opengl.GL2;

public class TubeRenderer {

	public static void drawTube(GL2 gl, ArrayList<float[]> path, ArrayList<float[]> directions, float thickness){
		float[][] pathA = path.toArray(new float[path.size()][]);
		float[][] directionsA = directions.toArray(new float[directions.size()][]);
		drawTube(gl, pathA, directionsA, thickness);
	}
	
	public static void drawTube(GL2 gl, float[][] path, float[][] directions, float thickness){
		if (path.length != directions.length || path.length<2) throw new IllegalArgumentException("Cannot render tube");
		float[] center = path[0];
		float[] normal1 = directions[0];
		float[][] normal = new float[8][3];
		float[][] points = new float[8][3];
		float[] u = new float[3];
		
		if (Math.abs(normal1[0]) >= Math.abs(normal1[1])){
			u[0] = -normal1[2];
			u[1] = 0f;
			u[2] = normal1[0];
		} else {
			u[0] = 0f;
			u[1] = normal1[2];
			u[2] = -normal1[1];
		}
		
		u = VectorOps.multiply(VectorOps.normalize(u), thickness*0.5f);
		float[]	v = VectorOps.crossProduct(normal1, u);
		v = VectorOps.multiply(VectorOps.normalize(v), thickness*0.5f);
		
		float[] uMul = VectorOps.multiply(u, 0.7071067811865475727f);
		float[] vMul = VectorOps.multiply(v, 0.7071067811865475727f);
		
		normal[0] = u;
		points[0] = VectorOps.add(center, normal[0]);
		
		normal[1] = VectorOps.add(uMul, vMul);
		points[1] = VectorOps.add(center, normal[1]);
		
		normal[2] = v;
		points[2] = VectorOps.add(center, normal[2]);
		
		normal[3] = VectorOps.sub(vMul, uMul);
		points[3] = VectorOps.add(center, normal[3]);
		
		normal[4] = VectorOps.invert(u);
		points[4] = VectorOps.add(center, normal[4]);
		
		normal[5] = VectorOps.invert(VectorOps.add(uMul, vMul));
		points[5] = VectorOps.add(center, normal[5]);
		
		normal[6] = VectorOps.invert(v);
		points[6] = VectorOps.add(center, normal[6]);
		
		normal[7] = VectorOps.sub(uMul, vMul);
		points[7] = VectorOps.add(center, normal[7]);
		
		
		drawCap(gl, center, points, VectorOps.invert(normal1), true);
		
		float[][] normal2 = normal;
		float[][] points2 = points;
		
		for (int j=1; j<path.length;j++){
			points2 = points;
			normal2 = normal;
			points = new float[8][3];
			normal = new float[8][3];
			
			float[] dir = directions[j-1];
			float[] n = VectorOps.add(directions[j], dir);
			//Check if two following direction are in opposite direction, in this case to not use average directions
			if (VectorOps.product(directions[j], dir) < -.9f) 
				n = directions[j];
			else
				n = VectorOps.normalize(n);
			
			float d = VectorOps.product(path[j], n);
			float il = 1f / VectorOps.product(dir, n);
			
			for (int i=0; i<8; i++){
				float lambda = (d-VectorOps.product(points2[i], n)) * il;
				points[i][0] = points2[i][0] + dir[0] * lambda;
				points[i][1] = points2[i][1] + dir[1] * lambda;
				points[i][2] = points2[i][2] + dir[2] * lambda;
				
				//Do not normalize normals, this is done in the shader
				normal[i][0] = points[i][0] - path[j][0];
				normal[i][1] = points[i][1] - path[j][1];
				normal[i][2] = points[i][2] - path[j][2];
			}
			
			gl.glBegin(GL2.GL_TRIANGLE_STRIP);
			for (int i=0; i<7; i++){
				gl.glNormal3f(normal[i][0], normal[i][1], normal[i][2]);
				gl.glVertex3f(points[i][0], points[i][1], points[i][2]);
				gl.glNormal3f(normal2[i][0], normal2[i][1], normal2[i][2]);
				gl.glVertex3f(points2[i][0], points2[i][1], points2[i][2]);
			}
			gl.glNormal3f(normal[0][0], normal[0][1], normal[0][2]);
			gl.glVertex3f(points[0][0], points[0][1], points[0][2]);
			gl.glNormal3f(normal2[0][0], normal2[0][1], normal2[0][2]);
			gl.glVertex3f(points2[0][0], points2[0][1], points2[0][2]);
			gl.glEnd();
		}
		
		center[0] = path[path.length-1][0];
		center[1] = path[path.length-1][1];
		center[2] = path[path.length-1][2];
		float l = 1f/VectorOps.length(directions[directions.length-1]);
		normal1[0] = directions[directions.length-1][0]*l;
		normal1[1] = directions[directions.length-1][1]*l;
		normal1[2] = directions[directions.length-1][2]*l;
		
		drawCap(gl, center, points, normal1, false);
	}
	
	private static void drawCap(GL2 gl, float[] center, float[][] points, float[] normal, boolean inverseOrder){
		gl.glBegin(GL2.GL_TRIANGLE_FAN);
		gl.glNormal3f(normal[0], normal[1], normal[2]);
		gl.glVertex3f(center[0], center[1], center[2]);
		if (!inverseOrder){
			gl.glVertex3f(points[0][0], points[0][1], points[0][2]);
			gl.glVertex3f(points[1][0], points[1][1], points[1][2]);
			gl.glVertex3f(points[2][0], points[2][1], points[2][2]);
			gl.glVertex3f(points[3][0], points[3][1], points[3][2]);
			gl.glVertex3f(points[4][0], points[4][1], points[4][2]);
			gl.glVertex3f(points[5][0], points[5][1], points[5][2]);
			gl.glVertex3f(points[6][0], points[6][1], points[6][2]);
			gl.glVertex3f(points[7][0], points[7][1], points[7][2]);
			gl.glVertex3f(points[0][0], points[0][1], points[0][2]);
		} else {
			gl.glVertex3f(points[0][0], points[0][1], points[0][2]);
			gl.glVertex3f(points[7][0], points[7][1], points[7][2]);
			gl.glVertex3f(points[6][0], points[6][1], points[6][2]);
			gl.glVertex3f(points[5][0], points[5][1], points[5][2]);
			gl.glVertex3f(points[4][0], points[4][1], points[4][2]);
			gl.glVertex3f(points[3][0], points[3][1], points[3][2]);
			gl.glVertex3f(points[2][0], points[2][1], points[2][2]);
			gl.glVertex3f(points[1][0], points[1][1], points[1][2]);
			gl.glVertex3f(points[0][0], points[0][1], points[0][2]);
		}
		gl.glEnd();
	}
}
