// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package common;

import model.Coordinate;
import common.VectorOps;

/**
 * Collection for vector calculations (3D). Vectors are represented by arrays, either of double or float values. Their length has to be 3. 
 */
public class VectorOps {

	/**
	 * Cross product of two vectors u x v
	 * @param u
	 * @param v
	 * @return u x v
	 */
	public static double[] crossProduct(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return new double[]{ u[1]*v[2]-u[2]*v[1],
							 u[2]*v[0]-u[0]*v[2],
							 u[0]*v[1]-u[1]*v[0]};
	}
	
	/**
	 * Cross product of two vectors u x v
	 * @param u
	 * @param v
	 * @return u x v
	 */
	public static float[] crossProduct(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return new float[]{ u[1]*v[2]-u[2]*v[1],
							 u[2]*v[0]-u[0]*v[2],
							 u[0]*v[1]-u[1]*v[0]};
	}
	
	/**
	 * Vector product of two vectors u*v
	 * @param u
	 * @param v
	 * @return u*v
	 */
	public static double product(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
	}
	
	/**
	 * Vector product of two vectors u*v
	 * @param u
	 * @param v
	 * @return u*v
	 */
	public static float product(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
	}
	
	/**
	 * Length of vector z
	 * @param u
	 * @return |u|
	 */
	public static double length(double[] u){
		assert (u.length == 3);
		return Math.sqrt(u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
	}
	
	/**
	 * Length of vector z
	 * @param u
	 * @return |u|
	 */
	public static float length(float[] u){
		assert (u.length == 3);
		return (float)Math.sqrt(u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
	}
	
	/**
	 * The square of vector u's length
	 * @param u
	 * @return |u|²
	 */
	public static float lengthSqr(float[] u){
		assert (u.length == 3);
		return (u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
	}
	
	/**
	 * The square of vector u's length
	 * @param u
	 * @return |u|²
	 */
	public static double lengthSqr(double[] u){
		assert (u.length == 3);
		return (u[0]*u[0]+u[1]*u[1]+u[2]*u[2]);
	}
	
	/**
	 * The difference of vector u from vector v
	 * @param u
	 * @return u-v
	 */
	public static double[] sub(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return new double[]{u[0]-v[0], u[1]-v[1], u[2]-v[2]};
	}
	
	/**
	 * The difference of vector u from vector v
	 * @param u
	 * @return u-v
	 */
	public static float[] sub(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return new float[]{u[0]-v[0], u[1]-v[1], u[2]-v[2]};
	}
	
	/**
	 * The difference of vector u from vector v
	 * @param u
	 * @return u-v
	 */
	public static float[] sub(Coordinate u, Coordinate v){
		return new float[]{u.getX()-v.getX(), u.getY()-v.getY(), u.getZ()-v.getZ()};
	}
	
	/**
	 * The sum of vector v and vector u
	 * @param u
	 * @return u+v
	 */
	public static double[] add(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return new double[]{u[0]+v[0], u[1]+v[1], u[2]+v[2]};
	}
	
	/**
	 * The sum of vector v and vector u
	 * @param u
	 * @return u+v
	 */
	public static float[] add(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return new float[]{u[0]+v[0], u[1]+v[1], u[2]+v[2]};
	}
	
	/**
	 * The product of vector u and scalar m
	 * @param u
	 * @return u*m
	 */
	public static double[] multiply(double[] u, double m){
		assert (u.length == 3);
		return new double[]{u[0]*m, u[1]*m, u[2]*m};
	}
	
	/**
	 * The product of vector u and scalar m
	 * @param u
	 * @return u*m
	 */
	public static float[] multiply(float[] u, float m){
		assert (u.length == 3);
		return new float[]{u[0]*m, u[1]*m, u[2]*m};
	}
	
	/**
	 * Vector u * (-1)
	 * @param u
	 * @return u * (-1)
	 */
	public static double[] invert(double[] u){
		assert (u.length == 3);
		return new double[]{-u[0],-u[1],-u[2]};
	}
	
	/**
	 * Vector u * (-1)
	 * @param u
	 * @return u * (-1)
	 */
	public static float[] invert(float[] u){
		assert (u.length == 3);
		return new float[]{-u[0],-u[1],-u[2]};
	}
	
	/**
	 * The cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|
	 */
	public static double dist(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return Math.sqrt((u[0]-v[0])*(u[0]-v[0])+(u[1]-v[1])*(u[1]-v[1])+(u[2]-v[2])*(u[2]-v[2]));
	}
	
	/**
	 * The cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|
	 */
	public static double dist(double[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return Math.sqrt((u[0]-v[0])*(u[0]-v[0])+(u[1]-v[1])*(u[1]-v[1])+(u[2]-v[2])*(u[2]-v[2]));
	}
	
	/**
	 * The cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|
	 */
	public static float dist(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return (float)Math.sqrt((u[0]-v[0])*(u[0]-v[0])+(u[1]-v[1])*(u[1]-v[1])+(u[2]-v[2])*(u[2]-v[2]));
	}
	
	/**
	 * The cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|
	 */
	public static float dist(Coordinate u, Coordinate v){
		return (float)Math.sqrt((u.getX()-v.getX())*(u.getX()-v.getX())+
				(u.getY()-v.getY())*(u.getY()-v.getY())+(u.getZ()-v.getZ())*(u.getZ()-v.getZ()));
	}
	
	/**
	 * The squared cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|²
	 */
	public static double sqrDist(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		return (u[0]-v[0])*(u[0]-v[0])+(u[1]-v[1])*(u[1]-v[1])+(u[2]-v[2])*(u[2]-v[2]);
	}
	
	/**
	 * The squared cartesian distance between vectors u and v
	 * @param u
	 * @param v
	 * @return |u-v|²
	 */
	public static float sqrDist(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		return (u[0]-v[0])*(u[0]-v[0])+(u[1]-v[1])*(u[1]-v[1])+(u[2]-v[2])*(u[2]-v[2]);
	}

	/**
	 * The squared cartesian distance between vectors u and Coordinate v
	 * @param u
	 * @param v
	 * @return |u-v|²
	 */
	public static float sqrDist(float[] u, Coordinate v){
		assert (u.length == 3);
		return (u[0]-v.getX())*(u[0]-v.getX())+(u[1]-v.getY())*(u[1]-v.getY())+(u[2]-v.getZ())*(u[2]-v.getZ());
	}
	
	/**
	 * The squared cartesian distance between Coordinate u and v
	 * @param u
	 * @param v
	 * @return |u-v|²
	 */
	public static float sqrDist(Coordinate u, Coordinate v){
		return (u.getX()-v.getX())*(u.getX()-v.getX())
			+ (u.getY()-v.getY())*(u.getY()-v.getY())
			+ (u.getZ()-v.getZ())*(u.getZ()-v.getZ());
	}
	
	/**
	 * The angle between vectors u and v
	 * @param u
	 * @param v
	 * @return acos( (u*v)/(|u|*|v|))
	 */
	public static double angle(float[] u, float[] v){
		assert (u.length == 3 && v.length == 3);
		float a = VectorOps.product(u, v) /(VectorOps.length(u) * VectorOps.length(v));
		if (a>1) a = 1;		//Rounding errors produce in some cases something like 1.00000001, acos will return NaN otherwise
		if (a<-1) a =- 1;
		return Math.acos(a);
	}
	
	/**
	 * The angle between vectors u and v
	 * @param u
	 * @param v
	 * @return acos( (u*v)/(|u|*|v|))
	 */
	public static double angle(double[] u, double[] v){
		assert (u.length == 3 && v.length == 3);
		double a = VectorOps.product(u, v) /(VectorOps.length(u) * VectorOps.length(v));
		if (a>1) a = 1;		//Rounding errors produce in some cases something like 1.00000001, acos will return NaN otherwise
		if (a<-1) a =- 1;
		return Math.acos(a);
	}
	
	/**
	 * Vector u with length = 1
	 * @param u
	 * @return u/|u|
	 */
	public static double[] normalize(double[] u){
		assert (u.length == 3);
		double l = VectorOps.length(u);
		return new double[]{u[0]/l, u[1]/l, u[2]/l};  
	}
	
	/**
	 * Vector u with length = 1
	 * @param u
	 * @return u/|u|
	 */
	public static float[] normalize(float[] u){
		assert (u.length == 3);
		float l = VectorOps.length(u);
		return new float[]{u[0]/l, u[1]/l, u[2]/l};  
	}
	
	/**
	 * Compute a normal from the vectors p2-p1 and p3-p1 
	 * @param p1
	 * @param p2
	 * @param p3
	 * @return normal with length 1 
	 */
	public static float[] makeNormal(float[] p1, float[] p2, float[] p3){
		float[] u = VectorOps.sub(p2, p1);
		float[] v = VectorOps.sub(p3, p1);
		float[] n = VectorOps.crossProduct(u, v);
		return VectorOps.normalize(n);
	}
	
	/**
	 * Compute a normal from the vectors between coordinates p2-p1 and p3-p1 
	 * @param p1
	 * @param p2
	 * @param p3
	 * @return normal with length 1 
	 */
	public static float[] makeNormal(Coordinate p1, Coordinate p2, Coordinate p3){
		float[] u = VectorOps.sub(p2, p1);
		float[] v = VectorOps.sub(p3, p1);
		float[] n = VectorOps.crossProduct(u, v);
		return VectorOps.normalize(n);
	}
}
