// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.awt.Color;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.*;
import java.util.concurrent.CyclicBarrier;
import java.util.regex.Pattern;

import common.ColorTable;
import crystalStructures.CrystalStructureProperties.CrystalProperty;

import model.*;
import model.BurgersVector.BurgersVectorType;
import model.MDFileLoader.ImportDataContainer;
import model.polygrain.Grain;
import model.polygrain.mesh.Mesh;
import model.skeletonizer.Skeletonizer;
import model.skeletonizer.processors.BurgersVectorAnalyzer;
import model.skeletonizer.processors.BurgersVectorAnalyzer.ClassificationPattern;
import model.skeletonizer.processors.SkeletonPreprocessor;

/**
 * All crystal structure depended subroutines are collected here

 */
public abstract class CrystalStructure{

	float latticeConstant;
	float nearestNeighborSearchRadius;
	ArrayList<CrystalProperty> crystalProperties = new ArrayList<CrystalStructureProperties.CrystalProperty>();
	float[][] defaultColors;
	
	private static final ArrayList<CrystalStructure> structures = new ArrayList<CrystalStructure>();
	/**
	 * Initialize a list of all supported crystal structures
	 * all CrystalStructures in this list must implement deriveNewInstance() and getIDName() correctly
	 */
	static {
		structures.add(new FCCStructure());
		structures.add(new FCCTwinnedStructure());
		structures.add(new BCCStructure());
		structures.add(new BCCExperimentalStructure());
		structures.add(new B2NiTi());
		structures.add(new B2());
		structures.add(new L12_Ni3AlStructure());
		structures.add(new L12_Ni3AlStructureExperimental());
		structures.add(new FeCStructure());
		structures.add(new DiamondCubicStructure());
		structures.add(new SiliconStructure());
		structures.add(new UndefinedCrystalStructure());
	}
	
	public CrystalStructure() {
		loadDefaultColorScheme();
		//if no default file exists, use default values
		if (this.defaultColors == null)
			this.defaultColors = getDefaultColors();
	}
	
	public static List<CrystalStructure> getCrystalStructures(){
		return Collections.unmodifiableList(structures);
	}
	
	public static CrystalStructure createCrystalStructure(String cs, float latticeConstant, float nearestNeighborSearchRadius) {
		CrystalStructure struct = null;
		//Check if there is a structures with the same identifier
		for (CrystalStructure c : structures){
			if (c.toString().equalsIgnoreCase(cs)) struct = c;
		}
		
		//if a structure with the given identifier is found, derive a new instance of this class or create a default one
		if (struct == null) struct = new UndefinedCrystalStructure();
		else struct = struct.deriveNewInstance();
			
		if (nearestNeighborSearchRadius == 0f) 
			struct.nearestNeighborSearchRadius = struct.getDefaultNearestNeighborSearchScaleFactor() * latticeConstant;
		else struct.nearestNeighborSearchRadius = nearestNeighborSearchRadius;
		
		struct.latticeConstant = latticeConstant;
		return struct;
	}
	
	/* **********************************************************
	 * Abstract methods
	 * **********************************************************/
	/**
	 * An unique ID labeling the crystal structure
	 * MUST BE ALWAYS overridden / implemented in each subclass
	 * @return
	 */
	protected abstract String getIDName();
	
	/**
	 * Create a new instance of the same class
	 * Basically it is just cloning, but avoiding casts.
	 * Additional operations can be performed in the subclass as well.
	 * MUST BE ALWAYS overridden / implemented in each subclass
	 * @return
	 */
	protected abstract CrystalStructure deriveNewInstance();
	
	/**
	 * Sets the maximum number of classes in which atoms can be assigned to
	 * Number should correspond to the number of labels in getNameForType(i)
	 * optionally override getColor(index) and getGLColor(index) if other color schemes
	 * are needed
	 * @return number of different atom classes in this structure
	 */
	public abstract int getNumberOfTypes();
	
	/**
	 * Identifies the atom type for a single atom by Bond angle analysis, Common neighbor analysis, whatever...
	 * @see getAtomTypeKeyword(), if the types are read from file, this method is not being used
	 * @param atom atoms to assign a atom type
	 * @param nnb The nearest neighbor graph for this task
	 * @return atom type, must be smaller than getNumberOfTypes()
	 */
	public abstract int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb);
	
	/**
	 * Return a list of patterns to map numerical Burgers vectors to the crystallographic ones.
	 * Each desired mapping must be defined in the list 
	 * Defines at the same time, the assignment of types for the Burgers vectors in 
	 * "identifyBurgersVectorType(BurgersVector bv)"
	 * @return
	 */
	public abstract List<ClassificationPattern> getBurgersVectorClassificationPattern();
	
	/**
	 * The number of nearest neighbor atoms in a perfect lattice
	 * Used in computing grain rotations
	 * @return
	 */
	public abstract int getNumberOfNearestNeighbors();
	
	
	/**
	 * Creates a subset of atoms belonging to dislocations that are considered during skeletonization
	 * @param data
	 * @return
	 */
	public abstract List<Atom> getDislocationDefectAtoms(AtomData data);
	
	/**
	 * Test if a Burgers vector should be calculated for an atom
	 * @param a
	 * @return
	 */
	public abstract boolean isRBVToBeCalculated(Atom a);
	
	/**
	 * Return the names for atoms of type i
	 * @param i
	 * @return
	 */
	public abstract String getNameForType(int i);
	
	/**
	 * The indicator for a atom type for atoms in (near) perfect lattice sites.
	 * This value is primarily needed to dispose atoms during loading files, that are not associated with defects
	 * @return Indicator for atoms in (near) perfect lattice sites
	 */
	public abstract int getDefaultType();
	
	/**
	 * Each crystal structure has to provide a indicator to identify a class of atoms assigned as a free surface.
	 * This indicator is required to filter defect structures on free surfaces.
	 * @return Indicator for atoms associated with a free surface
	 */
	public abstract int getSurfaceType();
	
	
	/**
	 * The length of a perfect Burgers vector, required to estimate a scalar value of GND densities.
	 * @return length of a perfect Burgers vector
	 */
	public abstract float getPerfectBurgersVectorLength();
	
	/**
	 * The perfect nearest neighbors in an arbitrary orientation 
	 * @return
	 */
	public abstract double[][] getPerfectNearestNeighborsUnrotated();
	
	
	/**
	 * The radius of the integrated sphere during the calculation of RBVs.
	 * Usually somewhere between the first and second nearest neighbor distance 
	 * @return
	 */
	public abstract float getRBVIntegrationRadius();
	
	/**
	 * Default scaling factor to search nearest neighbors relative to the lattice constant
	 * @return
	 */
	public abstract float getDefaultNearestNeighborSearchScaleFactor();
	
	
	/* ******************************
	 * Final methods
	 ********************************/
	/**
	 * Return the color associated for the given atom class index  
	 * @param index
	 * @return
	 */
	public final Color getColor(int index){
		float[] c = getGLColor(index); 
		return new Color(c[0], c[1],c[2]);
	}
	
	@Override
	public final String toString() {
		return getIDName();
	}
	
	/**
	 * Perform an identification of the atom types (Bond angle analysis, Common neighbor analysis, whatever...)
	 * @param atoms A list of all atoms
	 * @param nnb The nearest neighbor graph for this task
	 * @param start Perform the identification for all atom within the range of start and end
	 * The calculation has to be able to performed in parallel
	 * @param barrier If the analysis consists of several phases, this barrier can be used to synchronize the threads
	 * @param end see start
	 */
	public void identifyDefectAtoms(List<Atom> atoms, NearestNeighborBuilder<Atom> nnb, int start, int end, CyclicBarrier barrier) {
		for (int i=start; i<end; i++){
			if (Thread.interrupted()) return;
			Atom a = atoms.get(i);
			a.setType(identifyAtomType(a, nnb));
		}
	}
	
	public final float getLatticeConstant() {
		return latticeConstant;
	}
	
	/**
	 * Identifies a rotation matrix from the current configuration of bonds towards a reference
	 * @param a Atom to derive the rotation from
	 * @param nnb  NearestNeighborBuilder that can provide the required neighbors
	 * @return either the rotation matrix or null if the atom
	 *  is not suited to get rotation from (atom is defect atom or similar)
	 */
	public final float[][] identifyRotation(Atom a, NearestNeighborBuilder<Atom> nnb){
		ArrayList<Atom> neigh = nnb.getNeigh(a);
		//Identify the local grain rotation
		if (neigh.size() == getNumberOfNearestNeighbors()) {
			ArrayList<float[]> neighList = nnb.getNeighVec(a);
			float[][] neighVec = neighList.toArray(new float[neighList.size()][]);
			
			float[][] rot = CrystalRotationTools.getLocalRotationMatrix(a, neighVec, this);
			return rot;
		}
		return null;
	}
	
	
	/**
	 * The perfect nearest neighbors rotated into the grain orientation 
	 * @param g
	 * @return
	 */
	public final float[][] getPerfectNearestNeighbors(Grain g) {
		return getPerfectNearestNeighbors(g.getCystalRotationTools());
	}
	
	/**
	 * The perfect nearest neighbors rotated into the default orientation 
	 * @return
	 */
	public final float[][] getPerfectNearestNeighbors(){
		return getPerfectNearestNeighbors(Configuration.getCrystalRotationTools());
	}
	
	/**
	 * The nearest neighbors bonds (relative to a central atom) in the current crystal rotation
	 * @param crt The orientation of the crystal 
	 * @return array of float[3], each a bond in xyz
	 */
	public final float[][] getPerfectNearestNeighbors(CrystalRotationTools crt){
		double[][] neighPerf = getPerfectNearestNeighborsUnrotated();
		float[][] neighPerfFloat = new float[neighPerf.length][];
		
		float[][] rot = crt.getDefaultRotationMatrix();
		
		for (int i=0; i<neighPerf.length;i++){
			float[] n = new float[3];
			n[0] = (float)(neighPerf[i][0] * rot[0][0] + neighPerf[i][1] * rot[1][0] + neighPerf[i][2] * rot[2][0]) * latticeConstant; 
			n[1] = (float)(neighPerf[i][0] * rot[0][1] + neighPerf[i][1] * rot[1][1] + neighPerf[i][2] * rot[2][1]) * latticeConstant;
			n[2] = (float)(neighPerf[i][0] * rot[0][2] + neighPerf[i][1] * rot[1][2] + neighPerf[i][2] * rot[2][2]) * latticeConstant;
			neighPerfFloat[i] = n;
		}
		
		return neighPerfFloat;
	}
	
	/**
	 * Identifies the type of a Burgers vector based on the list of pattern from 
	 * "getBurgersVectorClassificationPattern()"
	 * @param bv
	 * @return
	 */
	public final BurgersVectorType identifyBurgersVectorType(BurgersVector bv){
		if (bv.getFraction() == 0) return BurgersVectorType.UNDEFINED;
		if (bv.getDirection()[0] == 0 && bv.getDirection()[1] == 0 && bv.getDirection()[2] == 0)
			return BurgersVectorType.ZERO;
		
		for (ClassificationPattern cp : getBurgersVectorClassificationPattern())
			if (cp.typeMatch(bv)) return cp.getType();
		
		return BurgersVectorType.OTHER;
	}
	
	/**
	 * Distance in which nearest neighbors are to be searched.
	 * Usually slightly larger than the nearest neighbor distance in a perfect crystals
	 * @see getDefaultNearestNeighborSearchScaleFactor()
	 * @return
	 */
	public final float getNearestNeighborSearchRadius() {
		return nearestNeighborSearchRadius;
	}
	
	
	/* ****************************************************************
	 * Methods that can be overridden if required, controlling details
	 * of the Nye tensor analysis, skeletonization and handling of 
	 * grain- or phase-boundaries 
	 ******************************************************************/
	
	/**
	 * A list of special options for a CrystalStructure
	 * Handle via CrystalStructureProperties 
	 * @return
	 */
	public ArrayList<CrystalProperty> getCrystalProperties() {
		return crystalProperties;
	}
	
	/**
	 * Return the color associated for the given atom class index  
	 * @param index float[3] array to be used in OpenGl
	 * @return
	 */
	public final float[] getGLColor(int index){
		assert (index<defaultColors.length && index>=0);
		return defaultColors[index];		
	}
	
	/**
	 * Provides the default color scheme for this crystal structure,
	 * length of the returned array must be at least of getNumberOfTypes() 
	 * @return
	 */
	public float[][] getDefaultColors(){
		return ColorTable.getGLColors();
	}
	
	
	public final void resetColors(){
		this.defaultColors = getDefaultColors();
		this.saveDefaultColorScheme();
	}
	
	/**
	 * Set a color for a certain type
	 * @param index
	 * @param color
	 */
	public final void setGLColors(int index, float[] color){
		assert (index<defaultColors.length && index >=0);
		assert(color.length>=3);
		
		defaultColors[index][0] = Math.min(1, Math.max(0, color[0]));
		defaultColors[index][1] = Math.min(1, Math.max(0, color[1]));
		defaultColors[index][2] = Math.min(1, Math.max(0, color[2]));
		
		this.saveDefaultColorScheme();
	}
	
	public final void loadColorsFromFile(File f) throws IOException {
		ArrayList<float[]> c = new ArrayList<float[]>();
		LineNumberReader lnr = new LineNumberReader(new FileReader(f));
		
		String s = lnr.readLine();
		Pattern p = Pattern.compile("\\s+");
		
		try {
			while (s!=null){
				String[] parts = p.split(s);
				if (parts.length<3) return;
				float[] c1 = new float[3];
				c1[0] = Float.parseFloat(parts[0]);
				c1[1] = Float.parseFloat(parts[1]);
				c1[2] = Float.parseFloat(parts[2]);
				c.add(c1);
				
				s = lnr.readLine();
			}	
		} catch (NumberFormatException nf){
			return;
		} finally {
			if (lnr!=null) lnr.close();
		}
		
		if (c.size()>=this.getNumberOfTypes())
			this.defaultColors = (float[][]) c.toArray(new float[c.size()][]);
		
		return;
	}
	
	
	private final void loadDefaultColorScheme(){
		File colorSchemeFile; 
		if (Configuration.RUN_AS_STICKWARE){
			colorSchemeFile = new File("ColorSchemes", this.getIDName()+".color");
		} else {
			String userHome = System.getProperty("user.home");
			File dir = new File(userHome+"/.AtomViewer/ColorSchemes");
			colorSchemeFile = new File(dir, this.getIDName()+".color");
		}
		if (colorSchemeFile.exists()){
			try {
				loadColorsFromFile(colorSchemeFile);
			} catch (IOException e) {}
		}
	}
	
	private final void saveDefaultColorScheme(){
		File colorSchemeFile; 
		if (Configuration.RUN_AS_STICKWARE){
			File subDir = new File("ColorSchemes");
			if (!subDir.exists()) subDir.mkdir();
			colorSchemeFile = new File(subDir, this.getIDName()+".color");
		} else {
			String userHome = System.getProperty("user.home");
			File dir = new File(userHome+"/.AtomViewer");
			File subDir = new File(dir+"ColorSchemes");
			if (!subDir.exists()) subDir.mkdir();
			colorSchemeFile = new File(subDir, this.getIDName()+".color");
		}
		try {
			saveColorsToFile(colorSchemeFile);
		} catch (IOException e) {}
	}
	
	public final void saveColorsToFile(File f) throws IOException {
		PrintWriter pw = new PrintWriter(f);
		for (int i=0; i<defaultColors.length; i++)
			pw.printf("%f %f %f\n", defaultColors[i][0], defaultColors[i][1], defaultColors[i][2]);
		pw.close();
	}
	
	/**
	 * Number of different elements in the crystal.
	 * Monoatomic structures return 1, Biatomic structures 2...
	 * If virtual atom types are used, the same scheme as in the IMD format for n elements has to be used
	 * 0 -> element 0
	 * 1 -> element 1
	 * ...
	 * n -> element n
	 * n+1 -> element 0
	 * n+2 -> element 1
	 * ...
	 * 2n -> element n
	 * 2n+1 -> element 0
	 * ...
	 * @return
	 */
	public int getNumberOfElements(){
		return 1;
	}
	
	/**
	 * Scaling factor to display atoms of different elements with different sizes 
	 * @return scaling factors, array has the same size as the value returned by getNumberOfElements()
	 */
	public float[] getSphereSizeScalings(){
		float[] size = new float[getNumberOfElements()];
		for (int i=0; i<getNumberOfElements(); i++)
			size[i] = 1f;
		return size;
	}
	
	/**
	 * If the atom types are given in the imported files, the returned string identifies the label under which
	 * the values are found
	 * if the label is found during import "identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb)" is not 
	 * going to be used
	 * @return ID identifying the atom type during import
	 */
	public String getAtomTypeKeyword(){
		return "ada_type";
	}
	
	/**
	 * Process a single line in the meta data field
	 * If it can be processed, the result is to be stored into the metaContainer
	 * @param s
	 * @param metaContainer
	 * @param lnr
	 * @param idc
	 * return true if processed
	 */
	public boolean processMetadataLine(String s, Map<String, Object> metaContainer,
			LineNumberReader lnr, ImportDataContainer idc) throws IOException{
		if (ImportStates.POLY_MATERIAL.isActive())
			return PolygrainMetadata.processMetadataLine(s, metaContainer, lnr, idc);
		return false;
	}
	
	/* **********************************
	 * skeletonization related methods
	 ************************************/
	/**
	 * Perform analysis on a finalized skeleton, e.g. mapping of Burgers vectors 
	 * @param skel
	 */
	public void analyse(Skeletonizer skel) {
		if (ImportStates.BURGERS_VECTORS.isActive())
			new BurgersVectorAnalyzer(this).analyse(skel);
	}
	
	/**
	 * Define a set of PreProcessors for the skeletonizer
	 * @return
	 */
	public List<SkeletonPreprocessor> getSkeletonizerPreProcessors(){
		return new Vector<SkeletonPreprocessor>();
	}
	
	/**
	 * Starts the creation of nearest neighbors in the field neigh of Atom.
	 * TODO: This method is implemented in a fairly limited way. Should be replaced one day.
	 * @param a
	 * @param nnb
	 */
	public void buildNeighborsForRBV(Atom a, NearestNeighborBuilder<Atom> nnb){
		a.buildNeigh(nnb);
	}
	
	/**
	 * Initial threshold to connect nodes in the skeleton.
	 * Larger values lead to smoother curves but may destroy details
	 * and increase time to compute networks 
	 * @return
	 */
	public float getSkeletonizationMeshingThreshold(){
		return getNearestNeighborSearchRadius();
	}
	
	/* ****************************
	 * Stacking fault methods, currently only in use in FCC 
	 *******************************/
	
	/**
	 * Defines if stacking faults exist in the structure.
	 * If true, the method getStackingFaultNormals requires to return the possible stacking fault normals
	 * @return
	 */
	public boolean hasStackingFaults(){
		return false;
	}
	
	/**
	 * If stacking faults exist in the crystal structure, this method provides their normals
	 * (in xyz-coordinates)
	 * @return normals of all possible stacking fault planes
	 */
	public float[][] getStackingFaultNormals(CrystalRotationTools crt){
		return new float[0][0];
	}
	
	/**
	 * Creates a subset of atoms belonging to stacking faults (if not existing, the list is empty)
	 * Should be linked with the hasStackingFault() method
	 * @param data
	 * @return
	 */
	public TreeSet<Atom> getStackingFaultAtoms(AtomData data){
		return new TreeSet<Atom>();
	};
	
	/**
	 * Each stacking fault consists of atoms all classified with the same type
	 * @return If there are multiple types that can form stacking faults, the methods returns true
	 */
	public boolean hasMultipleStackingFaultTypes(){
		return false;
	}
	
	/* *************************************
	 * Polycrystalline / Polyphase related methods
	 **************************************** */
	
	/**
	 * Crystal structure specific routines to identify different grains
	 * @param data
	 * @return
	 */
	public List<Grain> identifyGrains(AtomData data) {
		List<Grain> grains = new Vector<Grain>();
		
		if (!data.isGrainsImported()){
			/**
			 * HIGHLY EXPERIMENTAL STUFF
			 * do not expect to work properly
			 * In fact the opposite is true 
			 */
			final int defaultType = this.getDefaultType();
			final int surfaceType = this.getSurfaceType();
			
			data.findNearestNeighbors(getNearestNeighborSearchRadius(), new NearestNeighborFilter() {
				@Override
				public boolean accept(Atom a) {
					if (a.getType()!=surfaceType) return true;
					return false;
				}
			});
			
			
			//Detect grains from data
			TreeSet<Atom> notAssignedAtoms = new TreeSet<Atom>();
			for (Atom a : data.getAtoms()) {
				if ( a.getNeigh()!=null ) notAssignedAtoms.add(a);
			}
			while (!notAssignedAtoms.isEmpty()) {
				Atom a = notAssignedAtoms.pollFirst();
				if (a.getNeigh().length <= 9) continue;
				
				ArrayList<Atom> grain = new ArrayList<Atom>();
				//From each item in the grain array the bfs is continued
				//There is no need to keep a second list, the index of processed
				//atoms is enough
				int queuePosition = 0;
				int grainID = grains.size();
				grain.add(a);
				a.setGrain(grainID);
				
				while (queuePosition < grain.size()) {
					Atom c = grain.get(queuePosition++);
					int count = 0;
					for (int j=0; j<c.getNeigh().length; j++)
						if (c.getNeigh()[j].getType() == defaultType) count++;
					
					// in case some same element neighbors exists, continue a breath
					// first search at this atoms
					if (count > 9) {
						for (int i = 0; i < c.getNeigh().length; i++) {
							if (c.getNeigh()[i].getGrain() == Atom.DEFAULT_GRAIN) {
								c.getNeigh()[i].setGrain(grainID);
								grain.add(c.getNeigh()[i]);
								notAssignedAtoms.remove(c.getNeigh()[i]);
							}
						}
					}
				}
	
				if (grain.size() > 20) {
					Mesh mesh = new Mesh(grain, this);
					Grain g = new Grain(mesh, grain, grains.size(), this, data.getBounds());
					grains.add(g);
				} else {
					for (int i=0; i<grain.size(); i++) {
						grain.get(i).setGrain(Atom.DEFAULT_GRAIN);
					}
				}
			}
		} else {
			Object o = data.getFileMetaData("grain");
			PolygrainMetadata meta = null;
			if (o!=null && o instanceof PolygrainMetadata){
				meta = (PolygrainMetadata)o;
			}
			
			if (meta != null && meta.meshes.size() == meta.grainOrientation.size()){
				for (int i : meta.meshes.keySet()){
					Mesh mesh = new Mesh(meta.meshes.get(i).triangles, meta.meshes.get(i).vertices);
					Grain g;
					if (meta!=null && meta.grainOrientation.containsKey(i))
						g = new Grain(mesh, new ArrayList<Atom>(), i, this, meta.grainOrientation.get(i));
					else g = new Grain(mesh, new ArrayList<Atom>(), i, this, data.getBounds());
					grains.add(g);
					if (meta.numAtoms.get(i)!=null)
						g.setNumberOfAtoms(meta.numAtoms.get(i));
				}
			} else {
				//Grains are given in input, just process the sets
				HashMap<Integer, ArrayList<Atom>> grainsAtomLists = new HashMap<Integer, ArrayList<Atom>>();
				for (int i=0; i<data.getAtoms().size(); i++){
					Atom a = data.getAtoms().get(i);
					if (a.getGrain() != Atom.IGNORED_GRAIN){	
						if (!grainsAtomLists.containsKey(a.getGrain())){
							grainsAtomLists.put(a.getGrain(), new ArrayList<Atom>());
						}
						grainsAtomLists.get(a.getGrain()).add(a);
					}
				}
						
				//Mesh creation
				for (Integer i : grainsAtomLists.keySet()){
					ArrayList<Atom> grain = grainsAtomLists.get(i);
					if (grain!= null &&!grain.isEmpty()){
						Mesh mesh = new Mesh(grain, this);
						Grain g;
						if (meta!=null && meta.grainOrientation.containsKey(i))
							g = new Grain(mesh, grain, i, this, meta.grainOrientation.get(i));
						else g = new Grain(mesh, grain, i, this, data.getBounds());
						
						grains.add(g);
					}
				}
			}
		}
		return grains;
	}
	
	/**
	 * In true polycrystalline materials, grains must be identified beforehand the
	 * Nye tensor method can be applied. In some polyphase material like in Shape Memory Alloys
	 * the lattice structure is not deviating that much and it is better to compute RBV first and 
	 * filter dislocations
	 * @return
	 */
	public boolean createRBVbeforeGrains(){
		return false;
	}
	
	/**
	 * Defines if connection in the dislocation structure are possible
	 * over multiple grains (phases) or not
	 * @return
	 */
	public boolean skeletonizeOverMultipleGrains(){
		return false;
	}
	
	/**
	 * Grains are colored by size, enables a consistent coloring scheme
	 * @return
	 */
	public boolean orderGrainsBySize(){
		return false;
	}
	
	/**
	 * If an atom is closer to a grain boundary than this threshold it 
	 * will be assigned as IGNORED_GRAIN
	 * @return the filter threshold, filtering is disabled by setting to zero
	 */
	public float getGrainBoundaryFilterDistance(){
		return 0f;
	}
	
	/**
	 * Initial size to approximate a grain with a mesh
	 * Larger values result in coarser meshes and faster execution  
	 * @return
	 */
	public float getMeshingCellSize(){
		return 3f*nearestNeighborSearchRadius;
	}
}
