// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/>

package crystalStructures;

import java.util.*;

import common.Tupel;

import model.Atom;
import model.AtomData;
import model.ImportStates;
import model.NearestNeighborBuilder;
import model.NearestNeighborFilter;
import model.polygrain.Grain;
import model.polygrain.mesh.Mesh;

public class L12_Ni3AlStructure extends FCCStructure{
	
	private static final float[][] bondsAngleClasses;
	
	static{
		float[] tol_l = new float[]{0f, 7f, 6f, 5f, 4f, 3f};
		float[] tol_u = new float[]{13f, 10f, 6f, 5f, 4f, 3f};
		float[] angles = new float[]{180f, 146.5f, 120f, 109.5f, 90f, 60f};
		
		
		
		bondsAngleClasses = new float[6][2];
		
		// Lower bonds
		bondsAngleClasses[0][0] = -1.001f;		//Should be exactly -1, but there might be rounding errors
		bondsAngleClasses[1][0] = (float)Math.cos((angles[1]+tol_l[1])*Math.PI/180.);
		bondsAngleClasses[2][0] = (float)Math.cos((angles[2]+tol_l[2])*Math.PI/180.);
		bondsAngleClasses[3][0] = (float)Math.cos((angles[3]+tol_l[3])*Math.PI/180.);
		bondsAngleClasses[4][0] = (float)Math.cos((angles[4]+tol_l[4])*Math.PI/180.);
		bondsAngleClasses[5][0] = (float)Math.cos((angles[5]+tol_l[5])*Math.PI/180.);
		
		
		//Upper bonds
		bondsAngleClasses[0][1] = (float)Math.cos((angles[0]-tol_u[0])*Math.PI/180.);
		bondsAngleClasses[1][1] = (float)Math.cos((angles[1]-tol_u[1])*Math.PI/180.);
		bondsAngleClasses[2][1] = (float)Math.cos((angles[2]-tol_u[2])*Math.PI/180.);
		bondsAngleClasses[3][1] = (float)Math.cos((angles[3]-tol_u[3])*Math.PI/180.);
		bondsAngleClasses[4][1] = (float)Math.cos((angles[4]-tol_u[4])*Math.PI/180.);
		bondsAngleClasses[5][1] = (float)Math.cos((angles[5]-tol_u[5])*Math.PI/180.);
		
	}
	
	public L12_Ni3AlStructure() {
		super();
	}
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new L12_Ni3AlStructure();
	}
	
	@Override
	protected String getIDName() {
		return "Ni3Al";
	}
	
	@Override
	public int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb) {
		ArrayList<Tupel<Atom, float[]>> d_cont  = nnb.getNeighAndNeighVec(atom);
		
		/*
		 * type=0: L12
		 * type=1: APB
		 * type=2: CSF
		 * type=3: SISF
		 * type=4: 12 neighbors, defect
		 * type=5: 10-11 neighbors
		 * type=6: >=13 neighbors
		 * type=7: <10 neighbors (surface)
		 * type=8: pseudoTwin
		 */
		if (d_cont.size() < 10) return 7;
		else if (d_cont.size() < 11) return 5;
		else if (d_cont.size() > 13) return 6;
		else {
			//There are three type of bond within the binary system with an atom of type A in the middle of the bond
			//type 0: B-A-B (both neighbor the same, but other type)
			//type 1: A-A-B (one neighbor of each element)
			//type 2: A-A-A (bond consists of only one type)
			int bonds[][] = new int[bondsAngleClasses.length][3];
			
			//Element of the central atom in the bond pair
			int type_atom = atom.getElement()%2;
			
			
			for (int i = 0; i < d_cont.size(); i++) {
				//Element of the first bond atom
				int type_i = d_cont.get(i).getO1().getElement()%2;
				
				float v_x = d_cont.get(i).getO2()[0];
				float v_y = d_cont.get(i).getO2()[1];
				float v_z = d_cont.get(i).getO2()[2];
				float v_length = (float)Math.sqrt(v_x*v_x+v_y*v_y+v_z*v_z);
				
				for (int j = 0; j < i; j++) {
					//Element of the second bond atom
					int type_j = d_cont.get(j).getO1().getElement()%2;
					
					int bondType = 0;
					//Identify the type of bond 
					if (type_j!=type_i) bondType = 1;
					else {
						if (type_i == type_atom) bondType = 2;
						else bondType = 0;
					}
					
					float u_x = d_cont.get(j).getO2()[0];
					float u_y = d_cont.get(j).getO2()[1];
					float u_z = d_cont.get(j).getO2()[2];
					
					float u_length = (float)Math.sqrt(u_x*u_x+u_y*u_y+u_z*u_z);
					float a = v_x*u_x+v_y*u_y+v_z*u_z;
					a /= (v_length*u_length);
					
					for (int k=0; k<bondsAngleClasses.length; k++){
						if (a>=bondsAngleClasses[k][0] && a<=bondsAngleClasses[k][1]) bonds[k][bondType]++;
					}
				}
			}
			
			//type 0: B-A-B (both neighbor the same, but other type)
			//type 1: A-A-B (one neighbor of each element)
			//type 2: A-A-A (bond consists of only one type)
			
			//Nickel
			if (type_atom == 0){
				if (bonds[0][0] == 2 && bonds[0][2] == 4) return 0; //L12
				if (bonds[0][2] == 3 && bonds[0][1] == 2 && bonds[0][0] == 1) return 1; //APB
				if (bonds[0][2] == 4 && bonds[0][1] == 1 && bonds[0][0] == 1) return 1; //APB
				if (bonds[0][2] == 2 && bonds[0][0] == 1 && bonds[1][2] == 3 && bonds[1][1] == 2 && bonds[1][0] == 1) return 2; //CSF
				if (bonds[0][2] == 2 && bonds[0][0] == 1 && bonds[1][2] == 4 && bonds[1][1] == 2) return 2; //CSF
				if (bonds[0][2] == 2 && bonds[0][0] == 1 && bonds[1][2] == 2 && bonds[1][1] == 4) return 3; //SISF
				
				if (bonds[0][2] == 2 && bonds[0][0] == 1 && bonds[1][2]+bonds[1][1]+bonds[1][0] >= 3) return 2; //CSF
				
				if (bonds[0][0] == 1 && bonds[0][2] == 5  && bonds[1][0] == 0) return 8; //Pseudo-twin
				
			} else {	//Aluminum	
				if (bonds[0][0] == 6) return 0; //L12
				if (bonds[0][0] == 5 && bonds[0][1] == 1) return 1; //APB
				if (bonds[0][0] == 3 && bonds[1][1] == 2 && bonds[1][0] == 4) return 2; //CSF
				if (bonds[0][0] == 3 && bonds[1][0] == 6) return 3; //SISF
				
				if (bonds[0][0] == 3 && bonds[1][2]+bonds[1][1]+bonds[1][0] >=3 ) return 2; //CSF
				if (bonds[0][2] == 1 && bonds[0][0] == 5  && bonds[1][2] == 0) return 8; //Pseudo-twin
			}
			
			if (d_cont.size() < 12) return 5;
			else if (d_cont.size() > 12) return 6;
			else return 4;
		}
	}
	
	@Override
	public ArrayList<Atom> getDislocationDefectAtoms(AtomData data) {
		ArrayList<Atom> defectAtoms = new ArrayList<Atom>();
		
		float minRBVLength = latticeConstant*0.7071067f*0.25f; //latticeConstant*0.7071067f -> perfect burgers vector 1/2[110]
		minRBVLength *= minRBVLength;
		float maxRBVLength = latticeConstant*0.7071067f*2.5f;
		maxRBVLength *= maxRBVLength;
		
		for (Atom a : data.getAtoms()) {
			if ( (a.getType() != 0 && a.getType() != 7) && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN)){
				if (data.isRbvAvailable()){
					if (a.getRBV()!=null && a.getRBV().getSquaredRBVLengh()>minRBVLength && a.getRBV().getSquaredRBVLengh()<maxRBVLength)
						defectAtoms.add(a);
				}
				else defectAtoms.add(a);
			}
		}
		
		return defectAtoms;
	}
	
	
	@Override
	public String getNameForType(int i) {
			switch (i) {
			case 0: return "L12";
			case 1: return "APB";
			case 2: return "CSF";
			case 3: return "SISF";
			case 4: return "12 neigbors, defect";
			case 5: return "10-11 neighbors";
			case 6: return ">=13 neighbors";
			case 7: return "<10 neighbors";
			case 8: return "PseudoTwin";
			
			default: return "unknown";
		}
	}
	
	@Override
	public boolean hasMultipleStackingFaultTypes(){
		return true;
	}
	
	@Override
	public TreeSet<Atom> getStackingFaultAtoms(AtomData data){
		TreeSet<Atom> sfAtoms = new TreeSet<Atom>();
		for (int i=0; i<data.getAtoms().size(); i++){
			Atom a = data.getAtoms().get(i);
			if (a.getType() >= 1 && a.getType() <= 3 && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN))
				sfAtoms.add(a);
		}
		return sfAtoms;
	}
	
	@Override
	public int getSurfaceType() {
		return 7;
	}
	
	@Override
	public int getDefaultType(){
		return 0;
	}
	
	@Override
	public int getNumberOfTypes() {
		return 9;
	}
	
	@Override
	public int getNumberOfElements(){
		return 2;
	}
	
	@Override
	public float[] getSphereSizeScalings(){
		float[] size = new float[getNumberOfElements()];
		size[0] = 1f;
		size[1] = 0.79f;
		return size;
	}
	
	@Override
	public float getRBVIntegrationRadius() {
		return 0.707107f*latticeConstant*1.1f;
	}
	
	@Override
	public boolean isRBVToBeCalculated(Atom a) {
		return true;
	}
	
	@Override
	public List<Grain> identifyGrains(AtomData data) {
		List<Grain> grains = new Vector<Grain>();
		
		if (!data.isGrainsImported()){
			/**
			 * HIGHLY EXPERIMENTAL STUFF
			 * do not expect to work properly
			 * In fact the opposite is true 
			 */
			final int defaultType = this.getDefaultType();
			final int surfaceType = this.getSurfaceType();
			
			data.findNearestNeighbors(getNearestNeighborSearchRadius(), new NearestNeighborFilter() {
				@Override
				public boolean accept(Atom a) {
					if (a.getType()!=surfaceType) return true;
					return false;
				}
			});
			
			
			//Detect grains from data
			TreeSet<Atom> notAssignedAtoms = new TreeSet<Atom>();
			for (Atom a : data.getAtoms()) {
				if ( a.getNeigh()!=null ) notAssignedAtoms.add(a);
			}
			while (!notAssignedAtoms.isEmpty()) {
				Atom a = notAssignedAtoms.pollFirst();
				if (a.getNeigh().length <= 9) continue;
				
				ArrayList<Atom> grain = new ArrayList<Atom>();
				//From each item in the grain array the bfs is continued
				//There is no need to keep a second list, the index of processed
				//atoms is enough
				int queuePosition = 0;
				int grainID = grains.size();
				grain.add(a);
				a.setGrain(grainID);
				
				while (queuePosition < grain.size()) {
					Atom c = grain.get(queuePosition++);
					int count = 0;
					for (int j=0; j<c.getNeigh().length; j++)
						if (c.getNeigh()[j].getType() == defaultType) count++;
					
					// in case some same element neighbors exists, continue a breath
					// first search at this atoms
					if (count > 9) {
						for (int i = 0; i < c.getNeigh().length; i++) {
							if (c.getNeigh()[i].getGrain() == Atom.DEFAULT_GRAIN) {
								c.getNeigh()[i].setGrain(grainID);
								grain.add(c.getNeigh()[i]);
								notAssignedAtoms.remove(c.getNeigh()[i]);
							}
						}
					}
				}
	
				if (grain.size() > 20) {
					Mesh mesh = new Mesh(grain, this);
					Grain g = new Grain(mesh, grain, grains.size(), this, data.getBounds());
					grains.add(g);
				} else {
					for (int i=0; i<grain.size(); i++) {
						grain.get(i).setGrain(Atom.DEFAULT_GRAIN);
					}
				}
			}
		} else {
			return super.identifyGrains(data);
		}
		return grains;
	}
}
