// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.*;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.regex.Pattern;

import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import model.Configuration;
import model.CustomColumnInfo;

import common.CommonUtils;
import crystalStructures.*;

public class JCrystalConfigurationDialog extends JDialog{
	private static final long serialVersionUID = 1L;

	private DecimalFormat decimalFormat = new DecimalFormat("0.###");
	private JFormattedTextField latticeConstTextField = new JFormattedTextField(decimalFormat);
	private JFormattedTextField[][] crystalOrientation = new JFormattedTextField[3][3];
	private JComboBox crystalStructureComboBox = new JComboBox();
	private CrystalStructure crystalStructure;
	private JButton okButton = new JButton("OK");
	private JButton cancelButton = new JButton("cancel");
	private Container tableContainer = new Container();
	private Container crystalPropertiesContainer = new Container();
	private CrystalConfContent t;
	private JButton addRawValueButton = new JButton("add");
	
	private JTextField[] nameTextFields, idTextFields, unitTextFields;
	private JCheckBox[] activeCheckBoxes;
	
	private boolean isSavedSuccessfully = false;
	
	public JCrystalConfigurationDialog(final File folder, boolean showCancelButton){
		createDialog(folder, showCancelButton);
	}
	
	public JCrystalConfigurationDialog(Window owner, final File folder, boolean showCancelButton){
		super(owner);
		createDialog(folder, showCancelButton);
	}
	
	private void createDialog(final File folder, boolean showCancelButton){
		this.setTitle("Crystal configuration");
		this.setLayout(new GridBagLayout());
		this.setModalityType(Dialog.ModalityType.APPLICATION_MODAL);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 0; gbc.gridy = 0;
		gbc.weightx = 1;
		
		this.add(new JLabel("Crystal orientation"), gbc); gbc.gridy++;
		Container c = new Container();
		c.setLayout(new GridLayout(3,4));
		
		CrystalOrientationListener col = new CrystalOrientationListener();
		
		for (int i=0; i<3; i++){
			if (i == 0) c.add(new Label("X-Axis"));
			if (i == 1) c.add(new Label("Y-Axis"));
			if (i == 2) c.add(new Label("Z-Axis"));
			for (int j=0; j<3; j++){
				crystalOrientation[i][j] = new JFormattedTextField(decimalFormat);
				c.add(crystalOrientation[i][j]);
				if (i == j) crystalOrientation[i][j].setValue(1);
				else crystalOrientation[i][j].setValue(0);
				if (i == 2) crystalOrientation[i][j].setEditable(false); 
				else {
					crystalOrientation[i][j].addActionListener(col);
					crystalOrientation[i][j].addPropertyChangeListener(col);
				}
			}
		}
		
		gbc.gridwidth = 2;
		this.add(c, gbc); gbc.gridy++;
		gbc.gridwidth = 1;
		
		this.add(new JLabel(" "), gbc); gbc.gridy++;
		this.add(new JLabel("Crystal structure"), gbc); gbc.gridx++;
		this.add(crystalStructureComboBox, gbc); gbc.gridx = 0; gbc.gridy++;
		
		if (Configuration.Options.SIMPLE.isEnabled()){
			crystalStructureComboBox.addItem(new FCCStructure());
			crystalStructureComboBox.addItem(new BCCStructure());
			crystalStructureComboBox.addItem(new B2());
			crystalStructureComboBox.addItem(new UndefinedCrystalStructure());
		} else {
			for (CrystalStructure cs : CrystalStructure.getCrystalStructures())
				crystalStructureComboBox.addItem(cs);
		}
		
		this.add(new JLabel("Lattice constant"), gbc); gbc.gridx++;
		this.add(latticeConstTextField, gbc); gbc.gridy++;
		latticeConstTextField.setValue(3);
		gbc.gridx = 0;
		
		crystalStructureComboBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				crystalStructure = (CrystalStructure)crystalStructureComboBox.getSelectedItem();
				createCrystalPropertiesContainer();
			}
		});
		crystalStructureComboBox.setSelectedIndex(0);
		
		gbc.gridx = 0;
		gbc.gridy++;
		
		okButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					readRawValueTable();
					writeConfigurationFile(folder);
				} catch (IOException e1) {
					JOptionPane.showMessageDialog(null, "Error writing crystal.conf", "Error", JOptionPane.ERROR_MESSAGE);
					dispose();
				}
				isSavedSuccessfully = true;
				dispose();
			}
		});
		
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				isSavedSuccessfully = false;
				dispose();
			}
		});
		
		//Test if crystal.conf exists
		try {
			File conf = new File(folder, "crystal.conf");
			if (conf.exists()){
				t = readConfigurationFile(conf);
				if (t != null) {
					CrystalStructure cs = t.cs;
					float[][] o = t.orientation;
					
					crystalStructureComboBox.setSelectedIndex(0);
					for (int i = 0; i<crystalStructureComboBox.getItemCount(); i++){
						if (cs.toString().equalsIgnoreCase(crystalStructureComboBox.getItemAt(i).toString())){
							crystalStructureComboBox.setSelectedIndex(i);
							break;
						}
					}
					
					crystalOrientation[0][0].setValue(o[0][0]);
					crystalOrientation[0][1].setValue(o[0][1]);
					crystalOrientation[0][2].setValue(o[0][2]);
					
					crystalOrientation[1][0].setValue(o[1][0]);
					crystalOrientation[1][1].setValue(o[1][1]);
					crystalOrientation[1][2].setValue(o[1][2]);
					
					crystalOrientation[2][0].setValue(o[2][0]);
					crystalOrientation[2][1].setValue(o[2][1]);
					crystalOrientation[2][2].setValue(o[2][2]);
					
					latticeConstTextField.setValue(cs.getLatticeConstant());
				}
				crystalStructure = t.cs;
			}
			if (t == null) t = new CrystalConfContent(new float[][]{{1f,0f,0f},{1f,1f,0f},{0f,0f,1f}}, 
					new FCCStructure(), new ArrayList<CustomColumnInfo>());
		} catch (IOException ex){
			
		}
		
		addRawValueButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				readRawValueTable();
				t.rawColumns.add(new CustomColumnInfo("Name", "ID", "Unit", 0));
				createTable();
			}
		});
		
		
		createTable();
		JScrollPane sp = new JScrollPane(tableContainer);
		sp.setMinimumSize(new Dimension(200,200));
		sp.setPreferredSize(new Dimension(200,200));
		sp.setBorder(new TitledBorder(new EtchedBorder(1), "Import Raw Values"));
		gbc.gridwidth = 2;
		this.add(sp, gbc); gbc.gridy++;
		
		if (!Configuration.Options.SIMPLE.isEnabled()){
			createCrystalPropertiesContainer();
			sp = new JScrollPane(crystalPropertiesContainer);
			sp.setMinimumSize(new Dimension(400,200));
			sp.setPreferredSize(new Dimension(400,200));
			gbc.gridwidth = 2;
			this.add(sp, gbc); gbc.gridy++;
		}
		
		gbc.gridwidth = 1;
		
		gbc.gridx = 0;
		gbc.gridy++;
		
		if (!showCancelButton){
			gbc.gridwidth = 2;
			this.add(okButton, gbc);
		} else {
			this.add(okButton, gbc); gbc.gridx++;
			this.add(cancelButton, gbc);
		}
		
		this.pack();
		
		GraphicsDevice gd = this.getOwner().getGraphicsConfiguration().getDevice();
		this.setLocation( (gd.getDisplayMode().getWidth()-this.getWidth())>>1, 
				(gd.getDisplayMode().getHeight()-this.getHeight())>>1);
		this.setVisible(true);
	}
	
	
	
	private void createTable(){
		nameTextFields = new JTextField[t.rawColumns.size()];
		idTextFields = new JTextField[t.rawColumns.size()];
		unitTextFields = new JTextField[t.rawColumns.size()];
		activeCheckBoxes = new JCheckBox[t.rawColumns.size()];
		
		tableContainer.removeAll();
		tableContainer.setLayout(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 0; gbc.gridy = 0;
		gbc.weightx = 1;

		tableContainer.add(new JLabel("Name"), gbc); gbc.gridx++;
		tableContainer.add(new JLabel("ID in file"), gbc); gbc.gridx++;
		tableContainer.add(new JLabel("Unit"), gbc); gbc.gridx++;
		tableContainer.add(new JLabel("Active"), gbc);
		gbc.gridy++; gbc.gridx = 0;
		
		for (int i=0; i<t.rawColumns.size(); i++){
			nameTextFields[i] = new JTextField(t.rawColumns.get(i).getName());
			tableContainer.add(nameTextFields[i], gbc); gbc.gridx++;
			idTextFields[i] = new JTextField(t.rawColumns.get(i).getId());
			tableContainer.add(idTextFields[i], gbc); gbc.gridx++;
			unitTextFields[i] = new JTextField(t.rawColumns.get(i).getUnit());
			tableContainer.add(unitTextFields[i], gbc); gbc.gridx++;
			activeCheckBoxes[i] = new JCheckBox("", true);
			tableContainer.add(activeCheckBoxes[i], gbc);
			gbc.gridx = 0; gbc.gridy++;
		}
	
		gbc.gridwidth = 4;
		tableContainer.add(addRawValueButton, gbc);
		tableContainer.invalidate();
		this.validate();
	}
	
	private void createCrystalPropertiesContainer(){
		crystalPropertiesContainer.removeAll();
		crystalPropertiesContainer.setLayout(new GridLayout(1,1));
		crystalPropertiesContainer.add(CrystalStructureProperties.createPropertyContainer(crystalStructure.getCrystalProperties()));
		crystalPropertiesContainer.invalidate();
		this.validate();
	}
	
	private void readRawValueTable(){
		t.rawColumns.clear();
		
		for (int i=0; i<nameTextFields.length; i++){
			if (activeCheckBoxes[i].isSelected()){
				String name = nameTextFields[i].getText();
				String id = idTextFields[i].getText();
				String unit = unitTextFields[i].getText();
				CustomColumnInfo cci = new CustomColumnInfo(name, id, unit, t.rawColumns.size());
				t.rawColumns.add(cci);
			}
		}
	}
	
	public static CrystalConfContent readConfigurationFile(File confFile) throws IOException{
		CrystalStructure cs;
		float[][] crystalOrientation = new float[3][3];
		ArrayList<CustomColumnInfo> customColumns = new ArrayList<CustomColumnInfo>(); 
		
		LineNumberReader lnr = new LineNumberReader(new FileReader(confFile));
		
		float lattice = 0f;
		float nnd = 0f;
		String struct = "fcc";
		
		boolean structureFound = false, latticeConstFound = false;
		boolean orientationXfound = false, orientationYfound = false, orientationZfound = false;
		
		Pattern p = Pattern.compile("[ \t]+");
		String line = lnr.readLine();
		while (line!=null && !line.startsWith("#CrystalStructureOptions")){
			line = line.trim();
			if (!line.isEmpty() && !line.startsWith("#")){
				String[] parts = p.split(line);
				
				if (parts[0].toLowerCase().equals("structure")){
					struct = parts[1].toLowerCase();
					structureFound = true;
				} else if (parts[0].toLowerCase().equals("latticeconst")){
					if (parts.length != 1){
						lattice = Float.parseFloat(parts[1]);
						latticeConstFound = true;
					}
				} else if (parts[0].toLowerCase().equals("nearestneighcutoff")){
					if (parts.length != 1) nnd = Float.parseFloat(parts[1]);
				} else if (parts[0].toLowerCase().equals("orientation_x")){
					if (parts.length>=4){
						crystalOrientation[0][0] = Float.parseFloat(parts[1]);
						crystalOrientation[0][1] = Float.parseFloat(parts[2]);
						crystalOrientation[0][2] = Float.parseFloat(parts[3]);
						orientationXfound = true;
					}
				} else if (parts[0].toLowerCase().equals("orientation_y")){
					if (parts.length>=4){
						crystalOrientation[1][0] = Float.parseFloat(parts[1]);
						crystalOrientation[1][1] = Float.parseFloat(parts[2]);
						crystalOrientation[1][2] = Float.parseFloat(parts[3]);
						orientationYfound = true;
					}
				} else if (parts[0].toLowerCase().equals("orientation_z")){
					if (parts.length>=4){
						crystalOrientation[2][0] = Float.parseFloat(parts[1]);
						crystalOrientation[2][1] = Float.parseFloat(parts[2]);
						crystalOrientation[2][2] = Float.parseFloat(parts[3]);
						orientationZfound = true;
						
						
					}
				} else if (parts[0].toLowerCase().equals("raw")){
					if (parts.length>=3){
						CustomColumnInfo cci;
						if (parts.length>=4)
							cci = new CustomColumnInfo(parts[1], parts[2], parts[3], customColumns.size());
						else cci = new CustomColumnInfo(parts[1], parts[2], "", customColumns.size());
						customColumns.add(cci);
					}
				}
			}			
			
			line = lnr.readLine();
		}
			
		if (structureFound != true || latticeConstFound != true || 
				orientationXfound != true || orientationYfound != true || orientationZfound != true){
			lnr.close();
			return null;
		}
			
		
		cs = CrystalStructure.createCrystalStructure(struct, lattice, nnd);
		
		CrystalStructureProperties.readProperties(cs.getCrystalProperties(), lnr);
		
		lnr.close();
		return new CrystalConfContent(crystalOrientation, cs, customColumns);
	}
	
	public boolean isSavedSuccessfully() {
		return isSavedSuccessfully;
	}
	
	private class CrystalOrientationListener implements ActionListener, PropertyChangeListener{
		@Override
		public void actionPerformed(ActionEvent arg0) {
			update();
		}
		
		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			update();
		}
		
		
		private void update(){
			int[] x = new int[3];
			int[] y = new int[3];
			
			x[0] = ((Number) crystalOrientation[0][0].getValue()).intValue();
			x[1] = ((Number) crystalOrientation[0][1].getValue()).intValue();
			x[2] = ((Number) crystalOrientation[0][2].getValue()).intValue();
			
			y[0] = ((Number) crystalOrientation[1][0].getValue()).intValue();
			y[1] = ((Number) crystalOrientation[1][1].getValue()).intValue();
			y[2] = ((Number) crystalOrientation[1][2].getValue()).intValue();
			
			int[] z = new int[]{ x[1]*y[2]-x[2]*y[1],
					             x[2]*y[0]-x[0]*y[2],
					             x[0]*y[1]-x[1]*y[0]};
			
			int gcd = CommonUtils.gcd(z);
			if (gcd != 0){
				z[0] /= gcd; z[1] /= gcd; z[2] /= gcd;
			}
			
			crystalOrientation[2][0].setValue(z[0]);
			crystalOrientation[2][1].setValue(z[1]);
			crystalOrientation[2][2].setValue(z[2]);
		}
	}
	
	private void writeConfigurationFile(File folder) throws IOException {
		File f = new File(folder, "crystal.conf");
		if (!f.exists()){
			boolean created = f.createNewFile();
			if (!created) return;
		}
		if (!f.canWrite()) return;
		PrintWriter pw = new PrintWriter(f);
		
		int[] l = new int[3];
		l[0] = ((Number) crystalOrientation[0][0].getValue()).intValue();
		l[1] = ((Number) crystalOrientation[0][1].getValue()).intValue();
		l[2] = ((Number) crystalOrientation[0][2].getValue()).intValue();
		int gcd = CommonUtils.gcd(l);
		if (gcd != 0){
			l[0] /= gcd; l[1] /= gcd; l[2] /= gcd;
		}
		pw.println(String.format("orientation_x %d %d %d", l[0],l[1], l[2]));
		
		l[0] = ((Number) crystalOrientation[1][0].getValue()).intValue();
		l[1] = ((Number) crystalOrientation[1][1].getValue()).intValue();
		l[2] = ((Number) crystalOrientation[1][2].getValue()).intValue();
		gcd = CommonUtils.gcd(l);
		if (gcd != 0){
			l[0] /= gcd; l[1] /= gcd; l[2] /= gcd;
		}
		pw.println(String.format("orientation_y %d %d %d", l[0],l[1], l[2]));
		
		l[0] = ((Number) crystalOrientation[2][0].getValue()).intValue();
		l[1] = ((Number) crystalOrientation[2][1].getValue()).intValue();
		l[2] = ((Number) crystalOrientation[2][2].getValue()).intValue();
		gcd = CommonUtils.gcd(l);
		if (gcd != 0){
			l[0] /= gcd; l[1] /= gcd; l[2] /= gcd;
		}
		pw.println(String.format("orientation_z %d %d %d", l[0],l[1], l[2]));
		
		CrystalStructure cs = (CrystalStructure)crystalStructureComboBox.getSelectedItem();
		
		pw.println("structure\t"+cs.toString().toLowerCase());
		
		float latticeConst = ((Number) latticeConstTextField.getValue()).floatValue();
		if (latticeConst<0f) latticeConst = -latticeConst;
		
		float nnbDist = cs.getDefaultNearestNeighborSearchScaleFactor() * latticeConst;
		
		pw.println(String.format("latticeconst %.4f", latticeConst));
		pw.println(String.format("# Modify slighty if needed"));
		pw.println(String.format("nearestneighcutoff %.4f", nnbDist));
		
		for (int i=0; i<t.rawColumns.size(); i++){
			CustomColumnInfo c = t.rawColumns.get(i);
			if (!c.isSpecialColoumn())
				pw.println(String.format("raw %s %s %s", c.getName(), c.getId(), c.getUnit()));
		}
		
		CrystalStructureProperties.storeProperties(crystalStructure.getCrystalProperties(), pw);
		
		pw.close();	
	}
	
	public static class CrystalConfContent{
		float[][] orientation;
		CrystalStructure cs;
		ArrayList<CustomColumnInfo> rawColumns;
		public CrystalConfContent(float[][] orientation, CrystalStructure cs, ArrayList<CustomColumnInfo> rawColumns) {
			this.orientation = orientation;
			this.cs = cs;
			this.rawColumns = rawColumns;
		}
		
		public CrystalStructure getCrystalStructure() {
			return cs;
		}
		
		public float[][] getOrientation() {
			return orientation;
		}
		
		public ArrayList<CustomColumnInfo> getRawColumns() {
			return rawColumns;
		}
	}
}
