// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import javax.media.opengl.GL2;
import javax.media.opengl.GL3;

import com.jogamp.common.nio.Buffers;


public class ShaderCollection {

	public final static int ATTRIB_VERTEX = 0;
	public final static int ATTRIB_COLOR = 1;
	public final static int ATTRIB_TEX0 = 2;
	public final static int ATTRIB_VERTEX_OFFSET = 3;
	
	private static GL3Shader lastUsedGL3Shader = null;
	
	public enum Shader {
		TEXTURE_RENDER_SHADER(defaultVertexShader, anaglyphFragmentShader),
		PICKING_SHADER(defaultVertexShader, colorPickingFragmentShader),
		SPHERE_SHADER(translateSphereVertexShader, pplFragmentShader),
		SPHERE_ADS_SHADER(translateSphereVertexShader, pplADSFragmentShader), 
		
		BILLBOARD_SHADER(translateBillboardVertexShader, billboardFragmentShader),
		BILLBOARD_PICKING_SHADER(translateBillboardVertexShader, billboardPickingFragmentShader),
		
		SPHERE_PICKING_SHADER(translateSphereVertexPickingShader, colorPickingFragmentShader),
		DEFAULT_SHADER(defaultVertexShader, defaultFragmentShader),
		DEFAULT_PPL_SHADER(defaultPPLVertexShader, pplFragmentShader),
		DEFAULT_ADS_PPL_SHADER(defaultPPLVertexShader, pplADSFragmentShader),
		DEFAULT_TEXTURE_SHADER(defaultVertexShader, defaultTexturedFragmentShader);
	
		private int shaderProgram = -1;
		private String[] vertexShader, fragmentShader;
		
		private Shader(String[] vertexShader, String[] fragmentShader){
			this.vertexShader = vertexShader;
			this.fragmentShader = fragmentShader;
		}
		
		
		public void compile(GL2 gl){
			shaderProgram = createShaderProgram(vertexShader, fragmentShader, gl);
		}
		
		public void delete(GL2 gl){
			if (shaderProgram != -1) gl.glDeleteProgram(shaderProgram);
		}
		
		public boolean isAvailable(){
			return (shaderProgram!=-1);
		}
		
		public int getProgram(){
			return shaderProgram;
		}
		
		public void enable(GL2 gl){
			if (lastUsedGL3Shader!=null){
				System.out.println("Cannot switch from GL3 shader to GL2 shader without disabling the old shader explicitely");
				System.exit(-1);
			}
			gl.glUseProgram(this.shaderProgram);
		}
		
		public void enable(GL3 gl){
			if (lastUsedGL3Shader!=null){
				lastUsedGL3Shader.disable(gl);
			}
			gl.glUseProgram(this.shaderProgram);
		}
	};
	
	public enum GL3Shader {
		SPHERE_SHADER(translateSphereVertexInstancedShaderGL3, pplFragmentShaderGL3,
			 new int[]{ATTRIB_VERTEX, ATTRIB_COLOR, ATTRIB_VERTEX_OFFSET}, new String[]{"p", "Color", "Move"}),
		SPHERE_ADS_SHADER(translateSphereVertexInstancedShaderGL3, pplFragmentADSShaderGL3, 
				new int[]{ATTRIB_VERTEX, ATTRIB_COLOR, ATTRIB_VERTEX_OFFSET}, new String[]{"p", "Color", "Move"}),
		SPHERE_PICKING_SHADER(translateSphereVertexInstancedShaderGL3, spherePickingFragmentShaderGL3, 
				new int[]{ATTRIB_VERTEX, ATTRIB_COLOR, ATTRIB_VERTEX_OFFSET}, new String[]{"p", "Color", "Move"}),
	
		BILLBOARD_SHADER(translateBillboardVertexShaderGL3, billboardFragmentShaderGL3, 
				new int[]{ATTRIB_VERTEX, ATTRIB_COLOR, ATTRIB_VERTEX_OFFSET, ATTRIB_TEX0}, new String[]{"p", "Color", "Move", "tex"}),
		BILLBOARD_PICKING_SHADER(translateBillboardVertexShaderGL3, billboardPickingFragmentShaderGL3, 
				new int[]{ATTRIB_VERTEX, ATTRIB_COLOR, ATTRIB_VERTEX_OFFSET, ATTRIB_TEX0},new String[]{"p", "Color", "Move", "tex"}),
		
		VISIBILTY_TEST_SHADER(visiblityTestVertexGL3, visiblityTestFragmentGL3,
				new int[]{ATTRIB_VERTEX}, new String[]{"p"});
		
		private int shaderProgram = -1;
		private String[] vertexShader, fragmentShader;
		private int[] indices;
		private String[] attribs;
		
		private GL3Shader(String[] vertexShader, String[] fragmentShader, int[] indices, String[] attribs){
			this.vertexShader = vertexShader;
			this.indices= indices;
			this.fragmentShader = fragmentShader;
			this.attribs = attribs;
		}
		
		public void compile(GL3 gl){
			shaderProgram = createShaderProgram(vertexShader, fragmentShader, gl, indices, attribs);
		}
		
		public void delete(GL3 gl){
			if (shaderProgram != -1) gl.glDeleteProgram(shaderProgram);
		}
		
		public boolean isAvailable(){
			return (shaderProgram!=-1);
		}
		
		public int getProgram(){
			return shaderProgram;
		}
		
		public String[] getAttribs() {
			return attribs;
		}
		
		public void enable(GL3 gl){
			if (lastUsedGL3Shader == this) return;
			gl.glUseProgram(this.shaderProgram);
			if (lastUsedGL3Shader!=null){
				lastUsedGL3Shader.disable(gl);
			}
			lastUsedGL3Shader = this;
			
			for (int i: indices){
				gl.glEnableVertexAttribArray(i);
			}
		}
		
		public void disable(GL3 gl){
			for (int i: indices){
				gl.glDisableVertexAttribArray(i);
			}
			lastUsedGL3Shader = null;
		}
	};
	
	private static boolean initializedGL2 = false;
	private static boolean initializedGL3 = false;

	private static String[] defaultVertexShader = {
		"varying vec4 FrontColor;" +
		"void main(void)"+
		"{"+
		"gl_Position     = ftransform();"+
		"FrontColor      = gl_Color;"+
		"gl_TexCoord[0]  = gl_MultiTexCoord0;"+
		"}"
	};
	
	private static String[] defaultFragmentShader = {
		"varying vec4 FrontColor;"+
		"void main(void)"+
		"{"+
			"gl_FragColor = FrontColor;"+
		"}"
	};
	
	private static String[] defaultTexturedFragmentShader = {
		"uniform sampler2D Texture0;"+
		"varying vec4 FrontColor;"+
				
		"void main(void)"+
		"{"+
			"gl_FragColor = FrontColor * texture2D(Texture0, gl_TexCoord[0].xy);"+
		"}"
	};
	
	private static String[] defaultPPLVertexShader = {
		"varying vec4 FrontColor;" +
		"varying vec3 lightvec;"+
		"varying vec3 normal;"+
		
		"void main(void)"+
		"{"+
		" gl_Position    = gl_ModelViewProjectionMatrix * gl_Vertex;"+
		" normal         = normalize(gl_NormalMatrix * gl_Normal);"+
		" vec3 v         = vec3(gl_ModelViewMatrix * gl_Vertex);"+
		" lightvec       = normalize( vec3(-3.0, 0.0, 5.) - v);"+
		" FrontColor     = gl_Color;"+
		"}"
	};
	
	private static String[] anaglyphFragmentShader = {
		"uniform sampler2D left;"+
		"uniform sampler2D right;"+
		"uniform sampler2D back;"+
		"uniform int stereo;"+
	 
		"void main(void)"+
		"{"+
		"vec4 ColorL = texture2D(left, gl_TexCoord[0].st);"+
		"vec4 Background = texture2D(back, gl_TexCoord[0].st);"+
		"if (stereo == 1) {"+
		   "vec4 ColorR = texture2D(right, gl_TexCoord[0].st);"+
		   "float GrayR = dot(vec3(0.3, 0.59, 0.11), vec3(ColorR));"+
		   "gl_FragColor = vec4( GrayR, ColorL.g, ColorL.b, 1.) + Background*vec4(1.-ColorR.a, 1.-ColorL.a, 1.-ColorL.a, 1);"+
		"} else {"+
		   "gl_FragColor = ColorL + Background*vec4(1.-ColorL.a,1.-ColorL.a,1.-ColorL.a, 1.-ColorL.a);"+
		"}"+
		"}"
	};
	
	private static String[] colorPickingFragmentShader = {
		"uniform vec4 pickcolor;"+
		 
		"void main(void)"+
		"{"+
		"  gl_FragColor = pickcolor;"+
		"}"
	};
	
	private static String[] translateSphereVertexShader = {
		"varying vec3 lightvec;"+
		"varying vec3 normal;"+
		"varying vec4 FrontColor;"+
		"uniform vec4 Color;"+
		"uniform vec4 Move;"+
		"uniform mat4 mvm;"+
		"uniform mat4 mvpm;"+
		"uniform mat3 nm;"+

		"void main(void) {"+
		"  normal         = normalize(nm * gl_Vertex.xyz);"+
		"  vec4 scale     = vec4(Move[3], Move[3], Move[3], 1.);"+
		"  vec3 v         = vec3(mvm * (gl_Vertex*scale + vec4(Move.xyz, 0.)));"+
		"  lightvec       = normalize( vec3(-3.0, 0.0, 5.) - v);"+
		"  gl_Position    = mvpm * ( gl_Vertex*scale + vec4(Move.xyz, 0.) );"+
		"  FrontColor     = Color;"+
		"}"
	};
	
	private static String[] translateSphereVertexInstancedShaderGL3 = {
		"#version 330\n"+
		"in vec4 Move;"+
		"in vec4 Color;"+
		"in vec3 p;"+
		"out vec3 lightvec;"+
		"out vec3 normal;"+
		"out vec4 FrontColor;"+
		"uniform mat4 mvm;"+
		"uniform mat4 mvpm;"+
		"uniform mat3 nm;"+

		"void main(void) {"+
		"  FrontColor     = Color;\n"+
		"  vec4 vp        = vec4(p*Move[3] + Move.xyz, 1.);\n"+
		"  vec3 v         = vec3(mvm * vp);\n"+
		"  normal         = normalize(nm * p);\n"+
		"  lightvec       = normalize( vec3(-3.0, 0.0, 5.) - v);\n"+
		"  gl_Position    = mvpm * vp;\n"+
		"}"
	};
	
	private static String[] visiblityTestVertexGL3 = {
		"#version 330\n"+
		"in vec3 p;"+
		"uniform mat4 mvpm;"+
		"uniform vec3 shift;"+
		"uniform vec3 scale;"+

		"void main(void) {"+
		"  gl_Position    = mvpm * (vec4((p*scale+shift),1.));"+
		"}"
	};
	
	private static String[] visiblityTestFragmentGL3 = {
		"#version 330\n"+
		"out vec4 vFragColor;"+
				
		"void main(void) {"+
		"vFragColor   = vec4(1., 0., 0., 1.);"+
		"}"
	};
	
	private static String[] spherePickingFragmentShaderGL3 = {
		"#version 330\n"+
		"in vec4 FrontColor;"+
		"out vec4 vFragColor;"+
		
		"void main(void) {"+
			"vFragColor   = FrontColor;"+
		"}"
	};
	
	private static String[] pplFragmentShaderGL3 = {
		"#version 330\n"+
		"in vec3 lightvec;"+
		"in vec3 normal;"+
		"in vec4 FrontColor;"+
		"out vec4 vFragColor;"+
		
		"void main(void) {"+
		"vFragColor   = (( 0.8 * max(dot(normal, lightvec), 0.))+ vec4(0.5, 0.5, 0.5, 1.)) * FrontColor;"+
		"}"
	};
	
	private static String[] pplFragmentADSShaderGL3 = {
		"#version 330\n"+
		"in vec3 lightvec;"+
		"in vec3 normal;"+
		"in vec4 FrontColor;"+
		"out vec4 vFragColor;"+
		
		"void main(void) {"+
		"float diff = max(0.3, dot(normalize(normal), normalize(lightvec)));"+
		"vFragColor = vec4(diff,diff,diff,1.) * FrontColor;"+
//		"vFragColor   = (( 0.5 * max(dot(normalize(normal), normalize(lightvec)), 0.))+ vec4(0.5, 0.5, 0.5, 1.)) * FrontColor;"+
		"vec3 vReflection = normalize(reflect(-normalize(lightvec), normalize(normal)));"+
		"float spec = max(0.0, dot(normalize(normal), vReflection));"+
		"float fSpec = pow(spec, 96.0);"+
		"vFragColor.rgb += vec3(fSpec, fSpec, fSpec);"+
		"vFragColor += vec4(fSpec, fSpec, fSpec,1.)*FrontColor;"+
		"}"
	};
	
	private static String[] translateBillboardVertexShaderGL3 = {
		"#version 330\n"+
		"in vec4 Move;"+
		"in vec4 Color;"+
		"in vec3 p;"+
		"in vec2 tex;"+
		"out vec2 vTexCoord;"+
		"out vec4 FrontColor;"+
		"uniform mat4 mvm;"+
		"uniform mat4 mvpm;"+
		"uniform mat3 nm;"+

		"void main(void) {"+
		"  FrontColor     = Color;"+
		"  gl_Position    = mvpm * vec4(p*Move[3] + Move.xyz, 1.);"+
		"  vTexCoord      = tex;"+
		"}"
	};
	
	private static String[] billboardFragmentShaderGL3 = {
		"#version 330\n"+
		"uniform sampler2D Texture0;"+
		"in vec4 FrontColor;"+
		"in vec2 vTexCoord;"+
		"out vec4 vFragColor;"+
		
		"void main(void)"+
		"{"+
			"vec4 col = FrontColor * texture2D(Texture0, vTexCoord.xy);"+
			"if (col[3]<=0.75) {discard;}" +
			"else { col[3]=1.; }"+
			"vFragColor = col;"+
		"}"
	};
	
	private static String[] billboardPickingFragmentShaderGL3 = {
		"#version 330\n"+
		"uniform sampler2D Texture0;"+
		"in vec4 FrontColor;"+
		"in vec2 vTexCoord;"+
		"out vec4 vFragColor;"+
		
		"void main(void)"+
		"{"+
			"vec4 col = texture2D(Texture0, vTexCoord.xy);"+
			"if (col[3]<=0.75) {discard;}" +
			"vFragColor = FrontColor;"+
		"}"
	};
	
	private static String[] translateBillboardVertexShader= {
		"varying vec4 FrontColor;"+
		"uniform vec4 Color;"+
		"uniform vec4 Move;"+
		"uniform mat4 mvm;"+
		"uniform mat4 mvpm;"+

		"void main(void) {"+
		"  vec4 scale     = vec4(Move[3], Move[3], Move[3], 1.);"+
		"  gl_Position    = mvpm * ( gl_Vertex*scale + vec4(Move.xyz, 0.) );"+
		"  FrontColor     = Color;"+
		"  gl_TexCoord[0]  = gl_MultiTexCoord0;"+
		"}"
	};
	
	private static String[] billboardFragmentShader = {
		"uniform sampler2D Texture0;"+
		"varying vec4 FrontColor;"+
				
		"void main(void)"+
		"{"+
			"vec4 col = FrontColor * texture2D(Texture0, gl_TexCoord[0].xy);"+
			"if (col[3]<=0.75) {discard;}" +
			"else { col[3]=1.; }"+
			"gl_FragColor = col;"+
		"}"
	};
	
	
	private static String[] billboardPickingFragmentShader = {
		"uniform sampler2D Texture0;"+
		"uniform vec4 pickcolor;"+
				
		"void main(void)"+
		"{"+
			"vec4 col = texture2D(Texture0, gl_TexCoord[0].xy);"+
			"if (col[3]<=0.75) {discard;}" +
			"gl_FragColor = pickcolor;"+
		"}"
	};
	
	private static String[] translateSphereVertexPickingShader = {
		"uniform vec4 Move;"+

		"void main(void) {"+
		"  vec4 scale     = vec4(Move[3], Move[3], Move[3], 1.);"+
		"  gl_Position    = gl_ModelViewProjectionMatrix * ( gl_Vertex*scale + vec4(Move.xyz, 0.) );"+
		"}"
	};
	
	private static String[] pplADSFragmentShader = {
		"varying vec3 lightvec;"+
		"varying vec3 normal;"+
		"varying vec4 FrontColor;"+
		
		"void main(void) {"+
		//Complete ADS shader
//		"gl_FragColor   = (( 0.5 * max(dot(normalize(normal), normalize(lightvec)), 0.))+ vec4(0.5, 0.5, 0.5, 1.)) * FrontColor;"+
		"float diff = max(0.3, dot(normalize(normal), normalize(lightvec)));"+
		"gl_FragColor = vec4(diff,diff,diff,1.) * FrontColor;"+
		"vec3 vReflection = normalize(reflect(-normalize(lightvec), normalize(normal)));"+
		"float spec = max(0.0, dot(normalize(normal), vReflection));"+
		"float fSpec = pow(spec, 96.0);"+
		"gl_FragColor.rgb += vec3(fSpec, fSpec, fSpec);"+
		"gl_FragColor += vec4(fSpec, fSpec, fSpec,1.)*FrontColor;"+
		"}"
	};
	
	private static String[] pplFragmentShader = {
		"varying vec3 lightvec;"+
		"varying vec3 normal;"+
		"varying vec4 FrontColor;"+
		
		"void main(void) {"+
		"gl_FragColor   = (( 0.8 * max(dot(normal, lightvec), 0.))+ vec4(0.5, 0.5, 0.5, 1.)) * FrontColor;"+
		"}"
	};
	
	public static void init(GL2 gl){
		if (initializedGL2){
			//Reinit
			for (Shader s : Shader.values()) s.delete(gl);
		}
		for (Shader s : Shader.values()) s.compile(gl);
		
		initializedGL2 = true;
	}
	
	public static void initGL3(GL3 gl){
		if (initializedGL3){
			//Reinit
			for (GL3Shader s : GL3Shader.values()) s.delete(gl);
		}
		for (GL3Shader s : GL3Shader.values()) s.compile(gl);
		
		initializedGL3 = true;
	}
	
	public static void dispose(GL2 gl){
		if (initializedGL2){
			//Reinit
			for (Shader s : Shader.values()) s.delete(gl);
		}
		initializedGL2 = false;
	}
	
	public static void dispose(GL3 gl){
		if (initializedGL3){
			//Reinit
			for (GL3Shader s : GL3Shader.values()) s.delete(gl);
		}
		initializedGL3 = false;
	}
	
	private static int createShaderProgram(String[] vertexShader, String[] fragmentShader,GL2 gl){
		int v = gl.glCreateShader(GL2.GL_VERTEX_SHADER);
		int f = gl.glCreateShader(GL2.GL_FRAGMENT_SHADER);
		int program = gl.glCreateProgram();
		gl.glShaderSource(v, 1, vertexShader, null);
		gl.glCompileShader(v);
		gl.glShaderSource(f, 1, fragmentShader, null);
		gl.glCompileShader(f);
		gl.glAttachShader(program, v);
		gl.glAttachShader(program, f);
		
		gl.glLinkProgram(program);
		gl.glValidateProgram(program);
		
		checkShaderLogInfo(gl,v,vertexShader);
		checkShaderLogInfo(gl,f,fragmentShader);
		checkShaderLinkingLogInfo(gl, program, vertexShader, fragmentShader);
		
		gl.glDeleteShader(v);
		gl.glDeleteShader(f);
		
		return program;
	}
	
	
	private static int createShaderProgram(String[] vertexShader, String[] fragmentShader,GL3 gl, int[] indices, String[] attribs){
		if (indices.length != attribs.length){
			System.out.println("Number of  indices mismatches number of attributes");
			System.exit(1);
		}
			
		int v = gl.glCreateShader(GL3.GL_VERTEX_SHADER);
		int f = gl.glCreateShader(GL3.GL_FRAGMENT_SHADER);
		int program = gl.glCreateProgram();
		gl.glShaderSource(v, 1, vertexShader, null);
		gl.glCompileShader(v);
		gl.glShaderSource(f, 1, fragmentShader, null);
		gl.glCompileShader(f);
		gl.glAttachShader(program, v);
		gl.glAttachShader(program, f);
		
		for (int i=0; i<attribs.length; i++){
			gl.glBindAttribLocation(program, indices[i], attribs[i]);
		}
		
		gl.glLinkProgram(program);
		gl.glValidateProgram(program);
		
		checkShaderLogInfo(gl.getGL2(),v,vertexShader);
		checkShaderLogInfo(gl.getGL2(),f,fragmentShader);
		checkShaderLinkingLogInfo(gl.getGL2(), program, vertexShader, fragmentShader);
		
		gl.glDeleteShader(v);
		gl.glDeleteShader(f);
		
		return program;
	}
	
	
	private static void checkShaderLogInfo(GL2 gl, int shaderID, String[] shaderCode) {
		IntBuffer info = Buffers.newDirectIntBuffer(1);
		gl.glGetShaderiv(shaderID, GL2.GL_COMPILE_STATUS, info);
		if (info.get(0) == GL2.GL_FALSE) {
			gl.glGetShaderiv(shaderID, GL2.GL_INFO_LOG_LENGTH, info);
			final int length = info.get(0);
			String out = null;
			if (length > 0) {
				final ByteBuffer infoLog = Buffers.newDirectByteBuffer(length);
				gl.glGetShaderInfoLog(shaderID, infoLog.limit(), info, infoLog);
				final byte[] infoBytes = new byte[length];
				infoLog.get(infoBytes);
				out = new String(infoBytes);
				
				System.out.println("ERROR: Shader Compilation Error\n");
				System.out.print(out);
				System.out.println();
				System.out.print(shaderCode[0]);
				System.exit(0);
			}
		}
	}
	
	private static void checkShaderLinkingLogInfo(GL2 gl, int programID, String[] vShaderCode, String[] fShaderCode) {
		IntBuffer info = Buffers.newDirectIntBuffer(1);
		gl.glGetProgramiv(programID, GL2.GL_COMPILE_STATUS, info);
		if (info.get(0) == GL2.GL_FALSE) {
			gl.glGetShaderiv(programID, GL2.GL_INFO_LOG_LENGTH, info);
			final int length = info.get(0);
			String out = null;
			if (length > 0) {
				final ByteBuffer infoLog = Buffers.newDirectByteBuffer(length);
				gl.glGetShaderInfoLog(programID, infoLog.limit(), info, infoLog);
				final byte[] infoBytes = new byte[length];
				infoLog.get(infoBytes);
				out = new String(infoBytes);
				
				System.out.println("ERROR: Shader Linking Error\n");
				System.out.print(out);
				System.out.println();
				System.out.print(vShaderCode[0]);
				System.out.print(vShaderCode[1]);
				System.exit(0);
			}
		}
	}
	
	public static int loadShader(File vertexShader, File fragmentShader, GL2 gl) throws IOException{
		String[] vsrc = {""};
		String[] fsrc = {""};
		BufferedReader brv = null;
		BufferedReader brf = null;
		
		try {
			brf = new BufferedReader(new FileReader(vertexShader));
			
			String line;
			while ((line=brf.readLine()) != null)
			  fsrc[0] += line + "\n";
			
			brv = new BufferedReader(new FileReader(vertexShader));
			
			while ((line=brv.readLine()) != null)
			  vsrc[0] += line + "\n";
		} catch (IOException ex){
			throw ex; 
		}
		finally {
			if (brv != null) brv.close();
			if (brf != null) brf.close();
		}
		
		return createShaderProgram(vsrc, fsrc, gl);
	}
	
	public static void deleteShader(int shaderProgram, GL2 gl){
		gl.glDeleteProgram(shaderProgram);
	}
}
