// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Properties;

import javax.swing.JOptionPane;

import common.ColorTable.ColorBarScheme;

import crystalStructures.B2NiTi;
import crystalStructures.CrystalStructure;
import gui.ViewerGLJPanel;

public class Configuration {	
	
	/**
	 * If compiled while this flag is true, all config-files are placed in the program directory
	 * otherwise the config files are stored in the users home-directory
	 */
	public static final boolean RUN_AS_STICKWARE = true;
	
	private static File configFile;
	
	private static File lastOpenedFolder = null;
	private static File lastOpenedExportFolder = null;
	private static boolean headless = false;
	
	/**
	 * Global options for AtomViewer, accessible in the settings-menu
	 */
	public static enum Options {
		SIMPLE(true, "Basic features only", "AtomViewer must be restarted to apply all changes", 
				"Enables/Disables all implemented features. Non basic features may not work as expected, use at own risk!"),
		BILLBOARD_RENDERING(false, "Render spheres as flat textures", "", "Rendering flat textures can be much faster on some machines. However, this method can cause heavy graphical glitches."),
		HQ_RENDERING(false, "High quality rendering", "", "Slower rendering, but looks nicer."),
		ADS_SHADING(false, "Specular lighting", "", "Different lighting model");
		
		private Options(boolean enabled, String name, String message, String infoMessage){
			this.enabled = enabled;
			this.name = name;
			this.activateMessage = message;
			this.infoMessage = infoMessage;
		}
		
		private boolean enabled;
		private String activateMessage;
		private String infoMessage;
		private String name;
		private static ViewerGLJPanel viewer;
		
		public static void setViewerPanel(ViewerGLJPanel viewer){
			Options.viewer = viewer;
		}
		
		public void setEnabled(boolean enabled){
			this.enabled = enabled;
			if (viewer != null) {
				viewer.updateBillboards();
				viewer.reDraw();
			}
		}
		
		public String getName() {
			return name;
		}
		
		public String getActivateMessage() {
			return activateMessage;
		}
		
		public String getInfoMessage() {
			return infoMessage;
		}
		
		public boolean isEnabled(){
			return enabled;
		}
		
		static{
			loadProperties();
		}
		
		private static void loadProperties() {
			if (headless) return;
			Properties prop = new Properties();
			
			try {
				if (Configuration.RUN_AS_STICKWARE){
					configFile = new File("viewerSettings.conf");
				} else {
					String userHome = System.getProperty("user.home");
					File dir = new File(userHome+"/.AtomViewer");
					if (!dir.exists()) dir.mkdir();
					configFile = new File(dir, "viewerSettings.conf");
				}
				if (!configFile.exists()) saveProperties();
				prop.load(new FileReader(configFile));
			} catch (IOException e){
				e.printStackTrace();
			}
		
			for (Options i: Options.values()){
				if (prop.getProperty(i.toString()) != null)
					i.setEnabled(Boolean.parseBoolean(prop.getProperty(i.toString())));
			}
		}
		
		public static void saveProperties() {
			if (headless) return;
			Properties prop = new Properties();
		
			for (Options i: Options.values()){
				prop.setProperty(i.toString(), Boolean.toString(i.isEnabled()));
			}
			
			try {
				if (!configFile.exists()) configFile.createNewFile();
				if (configFile.canWrite()){
					prop.store(new FileWriter(configFile), "Viewer settings config file");
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			
		}
	}
	
	private static boolean[] pbc;
	private static CrystalStructure crystalStructure;
	private static CrystalRotationTools crystalRotationTools;
	private static float[][] crystalOrientation;
	private static AtomData currentAtomData;
	//Caching values for faster access, are often needed
	private static float currentAtomDataBound_x, currentAtomDataBound_y, currentAtomDataBound_z;
	private static float currentAtomDataHalfBound_x, currentAtomDataHalfBound_y, currentAtomDataHalfBound_z;
	
	private static int[] dislocationDensitySlices = new int[]{1,1,1};
	private static CustomColumnInfo selectedColumn = null;
	private static boolean filterRange = false;
	private static boolean filterInversed = false;
	private static HashMap<Integer, CustomColumnInfo> customColumns = new HashMap<Integer, CustomColumnInfo>();
	private static int latticeRotationCustomColumn = -1;
	private static int numElements = 1;
	private static HashSet<Integer> grainIndices = new HashSet<Integer>();
	
	public static boolean create(){
		Configuration.pbc = ImportStates.getPeriodicBoundaryConditions();
		Configuration.crystalStructure = ImportStates.getCrystalStructure();
		Configuration.crystalOrientation = ImportStates.getCrystalOrientation();
		
		Configuration.crystalRotationTools = new CrystalRotationTools(crystalStructure, crystalOrientation);
		
		//Disable custom columns if no values are found in the crystal.conf file
		customColumns.clear();
		latticeRotationCustomColumn = -1;
		
		grainIndices.clear();
		grainIndices.add(Atom.IGNORED_GRAIN);
		grainIndices.add(Atom.DEFAULT_GRAIN);
		
		if (ImportStates.getCustomColumns().size() == 0)
			ImportStates.CUSTOM_VALUES.setState(false);
		else if (ImportStates.CUSTOM_VALUES.isActive()) {
			for (int i=0; i<ImportStates.getCustomColumns().size(); i++){
				CustomColumnInfo c = ImportStates.getCustomColumns().get(i);
				customColumns.put(c.getColumn(), c);
			}
		}
		
		//TODO: Bah, instanceof!
		if (ImportStates.DETECT_MARTENSITE_VARIANTS.isActive() && crystalStructure instanceof B2NiTi){
        	ImportStates.CUSTOM_VALUES.setState(true);
        	CustomColumnInfo c = new CustomColumnInfo("Variant", "variant", "", customColumns.size(), 0f, 13f, ColorBarScheme.MARTENSITE_COLORING);
        	customColumns.put(c.getColumn(), c);
        } else ImportStates.DETECT_MARTENSITE_VARIANTS.setState(false);
		
		if (ImportStates.LATTICE_ROTATION.isActive()){
        	ImportStates.CUSTOM_VALUES.setState(true);
        	CustomColumnInfo c = new CustomColumnInfo("EulerX", "eulerX", "", customColumns.size(), true);
        	customColumns.put(c.getColumn(), c);
        	latticeRotationCustomColumn = c.getColumn();
        	c = new CustomColumnInfo("EulerY", "eulerY", "", customColumns.size(), true);
        	customColumns.put(c.getColumn(), c);
        	c = new CustomColumnInfo("EulerZ", "eulerZ", "", customColumns.size(), true);
        	customColumns.put(c.getColumn(), c);
        }
		
		if (ImportStates.DISLOCATION_DENSITY.isActive()){
			dislocationDensitySlices[0] = Integer.parseInt(JOptionPane.showInputDialog("How many slices for the dislocation density in x-direction?", 10));
			dislocationDensitySlices[1] = Integer.parseInt(JOptionPane.showInputDialog("How many slices for the dislocation density in y-direction?", 10));
			dislocationDensitySlices[2] = Integer.parseInt(JOptionPane.showInputDialog("How many slices for the dislocation density in z-direction?", 10));
		}
		
		return true;
	}
	
	public static File getLastOpenedExportFolder() {
		return lastOpenedExportFolder;
	}
	
	public static File getLastOpenedFolder() {
		return lastOpenedFolder;
	}
	
	public static void setLastOpenedExportFolder(File lastOpenedExportFolder) {
		Configuration.lastOpenedExportFolder = lastOpenedExportFolder;
	}
	
	public static void setLastOpenedFolder(File lastOpenedFolder) {
		Configuration.lastOpenedFolder = lastOpenedFolder;
	}
	
	public static boolean[] getPbc() {
		return pbc;
	}
	
	public static AtomData getCurrentAtomData() {
		return currentAtomData;
	}
	
	public static void setPBC(boolean[] pbc){
		Configuration.pbc[0] = pbc[0];
		Configuration.pbc[1] = pbc[1];
		Configuration.pbc[2] = pbc[2];
	}
	
	public static boolean isFilterRange() {
		return filterRange;
	}
	
	public static boolean isFilterInversed() {
		return filterInversed;
	}
	
	public static void setNumElements(int numElements) {
		Configuration.numElements = numElements;
	}
	
	public static int getNumElements() {
		return numElements;
	}
	
	public static CustomColumnInfo getSelectedColumn() {
		return selectedColumn;
	}
	
	public static boolean addGrainIndex(int index){
		return grainIndices.add(index);
	}
	
	public static boolean isHeadless() {
		return headless;
	}
	
	public static void setHeadless(boolean headless) {
		Configuration.headless = headless;
		Configuration.Options.SIMPLE.setEnabled(false);
	}
	
	/**
	 * Returns a copy of the grainIndices Set 
	 * @return
	 */
	public static HashSet<Integer> getGrainIndices(){
		return new HashSet<Integer>(grainIndices);
	}
	
	/**
	 * The first index for lattice rotation angles or -1 if not existing
	 * The other two angles are stored at the consequent indices
	 * @return
	 */
	public static int getLatticeRotationColumn(){
		return latticeRotationCustomColumn;
	}
	
	public static void setSelectedColumn(CustomColumnInfo selectedColumn) {
		Configuration.selectedColumn = selectedColumn;
	}
	
	public static void setFilterRange(boolean filterRange) {
		Configuration.filterRange = filterRange;
	}
	
	public static void setFilterInversed(boolean filterInversed) {
		Configuration.filterInversed = filterInversed;
	}
	
	public static void setCurrentAtomData(AtomData currentAtomData) {
		assert (currentAtomData.getBounds() != null) : "AtomData bounding box is null";
		assert (currentAtomData.getBounds()[0] != 0 ||
				currentAtomData.getBounds()[1] != 0 ||
				currentAtomData.getBounds()[2] != 0 ) : "AtomData bounding box has size 0";
		
		Configuration.currentAtomData = currentAtomData;
		Configuration.currentAtomDataBound_x = currentAtomData.getBounds()[0];
		Configuration.currentAtomDataBound_y = currentAtomData.getBounds()[1];
		Configuration.currentAtomDataBound_z = currentAtomData.getBounds()[2];
		
		Configuration.currentAtomDataHalfBound_x = currentAtomData.getBounds()[0] * 0.5f;
		Configuration.currentAtomDataHalfBound_y = currentAtomData.getBounds()[1] * 0.5f;
		Configuration.currentAtomDataHalfBound_z = currentAtomData.getBounds()[2] * 0.5f;
	}
	
	public static CrystalStructure getCrystalStructure() {
		return crystalStructure;
	}
	
	public static float[][] getCrystalOrientation() {
		return crystalOrientation;
	}
	
	public static CrystalRotationTools getCrystalRotationTools() {
		return crystalRotationTools;
	}
	
	public static int[] getDislocationDensitySlices() {
		return dislocationDensitySlices;
	}
	
	public static CustomColumnInfo getCustomColumnInfo(int i){
		return customColumns.get(i);
	}
	
	public static int getSizeCustomColumns(){
		return customColumns.size();
	}
	
	public static float[] pbcCorrectedDirection(Coordinate c1, Coordinate c2){
		float[] dir = new float[]{c1.getX()-c2.getX(), c1.getY()-c2.getY(), c1.getZ()-c2.getZ()};
		if (pbc[0]){
			if (dir[0]> currentAtomDataHalfBound_x) dir[0] -= currentAtomDataBound_x;
			else if (dir[0]<-currentAtomDataHalfBound_x) dir[0] += currentAtomDataBound_x;
		}
		if (pbc[1]){                                      
			if (dir[1]> currentAtomDataHalfBound_y) dir[1] -= currentAtomDataBound_y;
			else if (dir[1]<-currentAtomDataHalfBound_y) dir[1] += currentAtomDataBound_y;
		}
		if (pbc[2]){
			if (dir[2]> currentAtomDataHalfBound_z) dir[2] -= currentAtomDataBound_z;
			else if (dir[2]<-currentAtomDataHalfBound_z) dir[2] += currentAtomDataBound_z;
		}
		return dir;
	}
	
	public static float pbcCorrectedDistanceSqr(Coordinate c1, Coordinate c2){
		float x = c1.getX()-c2.getX();
		float y = c1.getY()-c2.getY();
		float z = c1.getZ()-c2.getZ();
		if (pbc[0]){
			if ( x > currentAtomDataHalfBound_x) x -= currentAtomDataBound_x;
			else if ( x <-currentAtomDataHalfBound_x) x += currentAtomDataBound_x;
		}                          
		if (pbc[1]){               
			if ( y > currentAtomDataHalfBound_y) y -= currentAtomDataBound_y;
			else if ( y <-currentAtomDataHalfBound_y) y += currentAtomDataBound_y;
		}                           
		if (pbc[2]){                
			if ( z > currentAtomDataHalfBound_z) z -= currentAtomDataBound_z;
			else if ( z <-currentAtomDataHalfBound_z) z += currentAtomDataBound_z;
		}
		
		return x*x + y*y + z*z;
	}
}
