// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.util.ArrayList;

import Jama.*;
import common.CommonUtils;
import common.VectorOps;
import crystalStructures.CrystalStructure;

public class CrystalRotationTools {
	private final float[][] thompsonTetraeder = new float[4][3];
	private final float[][] xyzToCrystal = new float[3][3];
	private final float[][] crystalToXyz = new float[3][3];
	private final float[][] rotationMatrix = new float[3][3];
	
	private final CrystalStructure cs;
	
	public CrystalRotationTools(CrystalStructure cs, float[][] crystalOrientation) {
		this.cs = cs;
		float lc = cs.getLatticeConstant();
		
		xyzToCrystal[0][0] = crystalOrientation[0][0]/VectorOps.length(crystalOrientation[0]) / lc;
		xyzToCrystal[0][1] = crystalOrientation[0][1]/VectorOps.length(crystalOrientation[0]) / lc;
		xyzToCrystal[0][2] = crystalOrientation[0][2]/VectorOps.length(crystalOrientation[0]) / lc;
		                   
		xyzToCrystal[1][0] = crystalOrientation[1][0]/VectorOps.length(crystalOrientation[1]) / lc;
		xyzToCrystal[1][1] = crystalOrientation[1][1]/VectorOps.length(crystalOrientation[1]) / lc;
		xyzToCrystal[1][2] = crystalOrientation[1][2]/VectorOps.length(crystalOrientation[1]) / lc;
		                   
		xyzToCrystal[2][0] = crystalOrientation[2][0]/VectorOps.length(crystalOrientation[2]) / lc;
		xyzToCrystal[2][1] = crystalOrientation[2][1]/VectorOps.length(crystalOrientation[2]) / lc;
		xyzToCrystal[2][2] = crystalOrientation[2][2]/VectorOps.length(crystalOrientation[2]) / lc;
		
		crystalToXyz[0][0] = xyzToCrystal[0][0] * lc * lc;
		crystalToXyz[0][1] = xyzToCrystal[1][0] * lc * lc;
		crystalToXyz[0][2] = xyzToCrystal[2][0] * lc * lc;
		                   
		crystalToXyz[1][0] = xyzToCrystal[0][1] * lc * lc;
		crystalToXyz[1][1] = xyzToCrystal[1][1] * lc * lc;
		crystalToXyz[1][2] = xyzToCrystal[2][1] * lc * lc;
		                   
		crystalToXyz[2][0] = xyzToCrystal[0][2] * lc * lc;
		crystalToXyz[2][1] = xyzToCrystal[1][2] * lc * lc;
		crystalToXyz[2][2] = xyzToCrystal[2][2] * lc * lc;
		
		thompsonTetraeder[2] = VectorOps.normalize(getInXYZ(new float[]{1f,0f,1f}));
		thompsonTetraeder[1] = VectorOps.normalize(getInXYZ(new float[]{0f,1f,1f}));
		thompsonTetraeder[0] = VectorOps.normalize(getInXYZ(new float[]{1f,1f,0f}));
		thompsonTetraeder[3] = new float[]{0f,0f,0f};
		
		rotationMatrix[0][0] = xyzToCrystal[0][0] * lc;
		rotationMatrix[0][1] = xyzToCrystal[1][0] * lc;
		rotationMatrix[0][2] = xyzToCrystal[2][0] * lc;
		
		rotationMatrix[1][0] = xyzToCrystal[0][1] * lc;
		rotationMatrix[1][1] = xyzToCrystal[1][1] * lc;
		rotationMatrix[1][2] = xyzToCrystal[2][1] * lc;
		
		rotationMatrix[2][0] = xyzToCrystal[0][2] * lc;
		rotationMatrix[2][1] = xyzToCrystal[1][2] * lc;
		rotationMatrix[2][2] = xyzToCrystal[2][2] * lc;
	}
	
	/**
	 * The rotation matrix in order to match the default atom positions of a crystalStructure
	 * into the positions of the current orientation
	 * @return
	 */
	public float[][] getDefaultRotationMatrix(){
		return rotationMatrix;
	}
	
	/**
	 * Transform a vector in XYZ-coordinates in crystal coordinates
	 * @param v
	 * @return
	 */
	public float[] getInCrystalCoordinates(float[] v){
		float[] f = new float[3];
		f[0] = xyzToCrystal[0][0]*v[0]+xyzToCrystal[1][0]*v[1]+xyzToCrystal[2][0]*v[2];
		f[1] = xyzToCrystal[0][1]*v[0]+xyzToCrystal[1][1]*v[1]+xyzToCrystal[2][1]*v[2];
		f[2] = xyzToCrystal[0][2]*v[0]+xyzToCrystal[1][2]*v[1]+xyzToCrystal[2][2]*v[2];
		return f;
	}

	/**
	 * Transform a vector in crystal coordinates in XYZ-coordinates
	 * @param v
	 * @return
	 */
	public float[] getInXYZ(float[] v){
		float[] f = new float[3];
		f[0] = crystalToXyz[0][0]*v[0]+crystalToXyz[1][0]*v[1]+crystalToXyz[2][0]*v[2];
		f[1] = crystalToXyz[0][1]*v[0]+crystalToXyz[1][1]*v[1]+crystalToXyz[2][1]*v[2];
		f[2] = crystalToXyz[0][2]*v[0]+crystalToXyz[1][2]*v[1]+crystalToXyz[2][2]*v[2];
		return f;
	} 
	
	public float[][] getThompsonTetraeder() {
		return thompsonTetraeder;
	}
	
	/**
	 * Return a least square rotation matrix for a most similar mapping to the default crystal orientation
	 * and the given atom and its neighbors
	 * @param a
	 * @param neighVec
	 * @param cs
	 * @return
	 */
	public static float[][] getLocalRotationMatrix(Atom a, float[][] neighVec, CrystalStructure cs){
		Matrix t = getAffineTransformationMatrix(a, neighVec, cs);
		if (t == null) return null;
		SingularValueDecomposition svd = new SingularValueDecomposition(t);
		Matrix rot = svd.getU().times(svd.getV().transpose());
		
		float[][] f = new float[3][3];
		f[0][0] = (float)rot.get(0, 0); f[0][1] = (float)rot.get(0, 1); f[0][2] = (float)rot.get(0, 2);
		f[1][0] = (float)rot.get(1, 0); f[1][1] = (float)rot.get(1, 1); f[1][2] = (float)rot.get(1, 2);
		f[2][0] = (float)rot.get(2, 0); f[2][1] = (float)rot.get(2, 1); f[2][2] = (float)rot.get(2, 2);
		return f;
	}
	
	/**
	 * Pick a basis of three distinct vectors from all nearest neighbor vectors
	 * Used to identify the crystal orientation
	 * @param unsortedBonds
	 * @return
	 */
	private static double[][] sortBonds(double[][] unsortedBonds){
		double[][] sort = new double[3][];
		
		//Put all neighbors in a list 
		ArrayList<double[]> ub = new ArrayList<double[]>();
		for (int i=0; i<unsortedBonds.length; i++) ub.add(unsortedBonds[i]);
		
		//Search the nearest one to the center
		double closestDist = Double.POSITIVE_INFINITY;
		int num = 0;
		for (int i=1; i<ub.size(); i++){
			double dist = VectorOps.lengthSqr(ub.get(i));
			if (dist < closestDist) {
				closestDist = dist;
				num = i;
			}
			
		}
		sort[0] = ub.remove(num);
		
		//Search the closest one to sort[0]
		closestDist = Double.POSITIVE_INFINITY;
		num = 0;
		for (int i=0; i<ub.size();i++){
			double dist = VectorOps.sqrDist(ub.get(i), sort[0]);
			if (dist < closestDist) {
				closestDist = dist;
				num = i;
			}
		}
		sort[1] = ub.remove(num);
		
		//Search the closest one to sort[0] and sort[1]
		closestDist = Double.POSITIVE_INFINITY;
		num = 0;
		for (int i=0; i<ub.size();i++){
			double dist = VectorOps.dist(ub.get(i), sort[0]) + VectorOps.dist(ub.get(i), sort[1]);
			if (dist < closestDist) {
				closestDist = dist;
				num = i;
			}
		}
		sort[2] = ub.get(num);
		//Finally check if the orientation of the system is positive, otherwise swap the order of the points
		if (VectorOps.product(VectorOps.crossProduct(sort[0], sort[1]), sort[2]) < 0f){
			double[] tmp = sort[1];
			sort[1] = sort[2];
			sort[2] = tmp;
		}
		
		return sort;
	}
	
	private static Matrix getAffineTransformationMatrix(Atom a, float[][] neighVec, CrystalStructure cs){
		if (neighVec == null || a.getType() != cs.getDefaultType())
			return null;
		
		double[][] neigh = new double[neighVec.length][3];
		
		for (int i=0; i<neighVec.length; i++){
			neigh[i][0] = neighVec[i][0]; 
			neigh[i][1] = neighVec[i][1];
			neigh[i][2] = neighVec[i][2];
		}
		
		Matrix trans = null;
		neigh = CrystalRotationTools.sortBonds(neigh);
		double[][] def = CrystalRotationTools.sortBonds(cs.getPerfectNearestNeighborsUnrotated());
		
		for (int i=0; i<def.length; i++)
			def[i] = VectorOps.normalize(def[i]);
		for (int i=0; i<neigh.length; i++)
			neigh[i] = VectorOps.normalize(neigh[i]);
		
		trans = new Matrix(neigh).solve(new Matrix(def));
		
		return trans;
	}
	
	/**
	 * Return the most likely true Burgers vector from the resultant burgers vector
	 * that is possible in the current crystal system
	 * @param rbv
	 * @return
	 */
	public BurgersVector rbvToBurgersVector(RBV rbv){
		return rbvToBurgersVector(rbv.getRBVAsVector());
	}

	/**
	 * Return the most likely true Burgers vector from the resultant burgers vector
	 * that is possible in the current crystal system
	 * @param rbv
	 * @return
	 */
	public BurgersVector rbvToBurgersVector(float[] rbv){
		if (rbv[0] == 0 && rbv[1] == 0 && rbv[2] == 0) return new BurgersVector(cs);
		float[] rbv_c = this.getInCrystalCoordinates(rbv);
		float l = VectorOps.length(rbv_c);
		
		int[] bv = new int[3];
		
		float max = Math.max(Math.max(Math.abs(rbv_c[0]), Math.abs(rbv_c[1])), Math.abs(rbv_c[2]));
		
		rbv_c[0] /= max;
		rbv_c[1] /= max;
		rbv_c[2] /= max;
		
		for (int i=0; i<3; i++){
			if (Math.abs(rbv_c[i])>0.75f) bv[i] = 6*(int)Math.signum(rbv_c[i]);
			else if (Math.abs(rbv_c[i])>0.412f) bv[i] = 3*(int)Math.signum(rbv_c[i]);
			else if (Math.abs(rbv_c[i])>0.166f) bv[i] = 2*(int)Math.signum(rbv_c[i]);
			else bv[i] = 0;
		}
		
		int gcd = CommonUtils.gcd(bv);
		
		bv[0] /= gcd;
		bv[1] /= gcd;
		bv[2] /= gcd;
		int f = (int)Math.round(Math.sqrt(bv[0]*bv[0] + bv[1]*bv[1] + bv[2]*bv[2]) / l);
		if (f==0) f=1;
		
		return new BurgersVector(f, bv[0], bv[1], bv[2], cs);
	}
}
