// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import gui.JMDFileChooser;
import javax.swing.SwingWorker;

import model.polygrain.Grain;

public abstract class MDFileLoader extends SwingWorker<AtomData, String> {
	public enum InputFormat {IMD, LAMMPS};
	
	protected JMDFileChooser chooser;
	
	public MDFileLoader(JMDFileChooser chooser) {
		this.chooser = chooser;
	}
	
	@Override
	protected final AtomData doInBackground() throws Exception{
		return readInputData();
	}
	
	/**
	 * Reading of MD input must be implemented here
	 * The recommend way to create is first buffering the data read in an
	 * instance of ImportDataContainer and passing this container into the 
	 * AtomData constructor.
	 * Generating a set of AtomData instances should be supported as well
	 * if this option is selected in the JMDFileChooser instance.
	 * In this case the set of data are connected as a double linked list
	 * by passing the previous created instance of AtomData as an argument in
	 * the constructor. 
	 * @return An instance of AtomData read from file, possibly linking to additional data sets 
	 * @throws Exception May be any kind of error during reading and processing
	 */
	protected abstract AtomData readInputData() throws Exception;
	
	/**
	 * Data from MD input must be stored in this container and then passed into 
	 */
	public static class ImportDataContainer{
		public float[] boxSize = new float[3];
		/**
		 * The coordinate system of AtomViewer does not support
		 * negative coordinates with periodic boundary conditions.
		 * If such an file is to be imported, all atoms must be shifted.
		 * The offset must be stored in this array. Printing coordinates of
		 * atoms will be corrected to match the input data.
		 */
		public float[] offset = new float[3];
		/**
		 * All atoms are stored in this list 
		 */
		public ArrayList<Atom> atoms = new ArrayList<Atom>();
		/**
		 * Only used in polycrystalline / polyphase material:
		 * Each subgrain is assigned a number, each atom is assigned a number of the grain its belongs to
		 */
		public HashMap<Integer, Grain> grains = new HashMap<Integer, Grain>();
		/**
		 * The largest (virtual) elements number found in all imported files
		 */
		public int maxNumElements = 1;
		//Some flags for imported or calculated values
		//Set to true if values are imported from files
		public boolean rbvAvailable = false;
		public boolean atomTypesAvailable = false;
		public boolean grainsImported = false;
		public boolean meshImported = false;
		public boolean latticeRotationImported = false;
		/**
		 * Meta data found in the file header
		 */
		Map<String, Object> fileMetaData = null;
		/**
		 * Identifier of the simulation, usually the filename
		 */
		String name;
		
		
		/**
		 * Moves a vector back into the simulation box in case of PBC 
		 * @param pos
		 */
		protected final void backInBox(float[] pos){
			if (Configuration.getPbc()[0]){
				while (pos[0] < 0f) pos[0] += this.boxSize[0];
				while (pos[0] >= this.boxSize[0]) pos[0] -= this.boxSize[0];
			}
			if (Configuration.getPbc()[1]){
				while (pos[1] < 0f) pos[1] += this.boxSize[1];
				while (pos[1] >= this.boxSize[1]) pos[1] -= this.boxSize[1];
			}
			if (Configuration.getPbc()[2]){
				while (pos[2] < 0f) pos[2] += this.boxSize[2];
				while (pos[2] >= this.boxSize[2]) pos[2] -= this.boxSize[2];
			}
		}
	}
}
