// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.util.*;
import java.util.concurrent.*;

import model.polygrain.Grain;
import common.*;
import crystalStructures.CrystalStructure;

/**
 * Computing numerically resultant Burgers vectors
 * Reference implementation for the algorithm published in 
 * 
 * C. Begau, J. Hua, A. Hartmaier 
 * A novel approach to study dislocation density tensors and lattice rotation patterns in atomistic simulations
 * Journal of the Mechanics and Physics of Solids, vol. 60, pp. 711-722, (2012)
 * 
 * Computing the lattice correspondence and the Nye tensor is implemented as described in
 * 
 * C. Hartley and Y. Mishin
 * Characterization and visualization of the lattice misfit associated with dislocation cores
 * Acta Mater., vol. 53, no. 5, pp. 1313–1321, (2005)
 *  
 * @author Christoph Begau
 *
 */
public class RbvBuilder {
	
	/**
	 * Maximum deviation allowed to detect matching bonds (=20°)
	 */
	private static final double PHI_MAX = Math.cos(20*Math.PI/180.);
	
	private final float[][] neighPerf;
	
	//Icoseader data
	private static final float icoConstX  = 1f;
	private static final float icoConstZ  = 1.6180339887f;
	private final double[][] icoVertices;
	
	private final double[][] icoNormals;
	
	private static final int[] icoFaces = {
		0,4,1,  0,9,4,  9,5,4,  4,5,8,  4,8,1,
		8,10,1, 8,3,10, 5,3,8,  5,2,3,  2,7,3,
		7,10,3, 7,6,10, 7,11,6, 11,0,6, 0,1,6,
		6,1,10, 9,0,11, 9,11,2, 9,2,5,  7,2,11
	};
	
	private float[] pnl;
	private float integrationRadius;
	private ConcurrentHashMap<Atom,RbvInfo> atomToRbvInfoMap = new ConcurrentHashMap<Atom, RbvInfo>(16, 0.75f, ThreadPool.availProcessors());
	private final boolean pbc_x, pbc_y, pbc_z;
	private float[] bounds;
	//The half of the box is needed so often, that a precalculated value is useful to save time 
	private float halfBound_x, halfBound_y, halfBound_z;
	
	private NearestNeighborBuilder<Atom> nnb;
	
	/**
	 * 
	 * @param data
	 * @param nearestneighborDistance nearestNeighborDistance used in simulation
	 */
	public static void createRBV(final AtomData data){
		if (data.getGrains().size() == 0){
			new RbvBuilder(data.getAtoms(), data.getBounds(),
					Configuration.getCrystalStructure().getPerfectNearestNeighbors(),
					Configuration.getCrystalStructure().getRBVIntegrationRadius());
		} else {
			ArrayList<Atom> atomsInNoGrain = new ArrayList<Atom>();
			for (int i=0; i<data.getAtoms().size(); i++){
				Atom a = data.getAtoms().get(i);
				if (a.getGrain() == Atom.DEFAULT_GRAIN){
					boolean neighInSameGrain = true;
					if (a.getNeigh() == null) neighInSameGrain = false;
					else {
						for (int j=0; j<a.getNeigh().length; j++){
							if (a.getNeigh()[j].getGrain() != Atom.DEFAULT_GRAIN){
								neighInSameGrain = false;
								break;
							}
						}
					}
					if (neighInSameGrain) atomsInNoGrain.add(a);
				}
			}
			new RbvBuilder(data.getAtoms(), data.getBounds(),
					Configuration.getCrystalStructure().getPerfectNearestNeighbors(),
					Configuration.getCrystalStructure().getRBVIntegrationRadius());
		}
	}
	
	public static void createRBV(final Grain grain, float[] boundingBox){
		new RbvBuilder(grain.getAtomsInGrain(), boundingBox, 
				grain.getCrystalStructure().getPerfectNearestNeighbors(grain),
				grain.getCrystalStructure().getRBVIntegrationRadius());
		for (int i=0; i<grain.getAtomsInGrain().size(); i++){
			grain.getAtomsInGrain().get(i).removeNeigh();
		}
	}
	
	private RbvBuilder(List<Atom> atoms, float[] boundingBox, float[][] perfNeighbors, float integrationRadius) {
		this.nnb = new NearestNeighborBuilder<Atom>(boundingBox, 
				Configuration.getCrystalStructure().getNearestNeighborSearchRadius(), Configuration.getPbc());
		
		for (int i=0; i<atoms.size(); i++){
			Atom a = atoms.get(i);
			if (a.getGrain() != Atom.IGNORED_GRAIN) nnb.add(a);
		}
		
		this.pnl = new float[perfNeighbors.length];
		for (int i=0; i<this.pnl.length; i++){
			this.pnl[i] = VectorOps.length(perfNeighbors[i]);
		}
		
		this.integrationRadius = integrationRadius;
		
		this.pbc_x = Configuration.getPbc()[0];
		this.pbc_y = Configuration.getPbc()[1];
		this.pbc_z = Configuration.getPbc()[2];
		this.bounds = boundingBox;
		
		this.halfBound_x = bounds[0]*0.5f;
		this.halfBound_y = bounds[1]*0.5f;
		this.halfBound_z = bounds[2]*0.5f;
		
		this.neighPerf = perfNeighbors;

		//Create icoseader vertices to approximate a sphere with a radius of the integration radius
		double r = integrationRadius;
		//Icosaeder edge length
		r = (2*r)/Math.sqrt(10+2*Math.sqrt(5));
		
		icoVertices = new double[][] {
			{-icoConstX*r, 0f, icoConstZ*r}, {icoConstX*r, 0f, icoConstZ*r},
			{-icoConstX*r, 0f, -icoConstZ*r}, {icoConstX*r, 0f, -icoConstZ*r},
			{0f, icoConstZ*r, icoConstX*r}, {0f, icoConstZ*r, -icoConstX*r},
			{0f, -icoConstZ*r, icoConstX*r}, {0f, -icoConstZ*r, -icoConstX*r},
			{icoConstZ*r, icoConstX*r, 0f}, {-icoConstZ*r, icoConstX*r, 0f},
			{icoConstZ*r, -icoConstX*r, 0f}, {-icoConstZ*r, -icoConstX*r, 0f}
		};
		
		icoNormals = new double[12][];
		for (int i=0; i<12; i++){
			icoNormals[i] = VectorOps.normalize(icoVertices[i]);
		}
		
				
		List<RbvInfo> infosList = new ArrayList<RbvInfo>();
		CrystalStructure cs = Configuration.getCrystalStructure();
		
		for (int i=0; i<atoms.size(); i++){
			Atom a = atoms.get(i);
			
			if (cs.isRBVToBeCalculated(a)){
				RbvInfo info = new RbvInfo();
				info.atom = atoms.get(i);
				atomToRbvInfoMap.put(atoms.get(i), info);
				infosList.add(info);
			}
		}
		
		
		Vector<AnalyseCallable> tasks = new Vector<AnalyseCallable>();
		
		for (int i = 0; i < ThreadPool.availProcessors(); i++) {
			int start = (infosList.size() * i) / ThreadPool.availProcessors();
			int end = (infosList.size() * (i + 1)) / ThreadPool.availProcessors();
			tasks.add(new AnalyseCallable(start, end, infosList));
		}
		
		ThreadPool.executeParallel(tasks);
	}

	/**
	 * Calculate the volume of an hexaedron
	 * Integrate the nye tensor over a triangle by giving the triangle vertices as p0-p2
	 * for p3-p4 provide p0+normal*nyeTensor
	 * the returned volume is the integrated nye tensor for the triangle
	 * @param p0
	 * @param p1
	 * @param p2
	 * @param p3
	 * @param p4
	 * @param p5
	 * @return
	 */
	private static double integral(double[] p0, double[] p1, double[] p2,
			double[] p3, double[] p4, double[] p5){
		//The integral is the volume of an irregular hexaedron
		//Decompose this hexaedron into three tetrahedrons and sum their volumes
		double volume = tetraederVolume(p3, p5, p1, p4);
		volume += tetraederVolume(p3, p5, p2, p1);
		volume += tetraederVolume(p3, p2, p0, p1);
		
		return volume;
	}

//  Default integration, much slower
//	private static double integral(double[] p0, double[] p1, double[] p2 ){
//		double[] p3 = new double[]{p0[0],p0[1],0f};
//		double[] p4 = new double[]{p1[0],p1[1],0f};
//		double[] p5 = new double[]{p2[0],p2[1],0f};
//		
////		//The integral forms a irregular hexaedron
////		//Decompose it into three tetrahedrons and sum their volume
//		double volume = tetraederVolume(p3, p5, p1, p4);
//		volume += tetraederVolume(p3, p5, p2, p1);
//		volume += tetraederVolume(p3, p2, p0, p1);
//		
//		return volume;
//	}
//	
	private static double tetraederVolume(double[] p1, double[] p2, double[] p3, double[] p4){
//		double[] dir1 = VectorOps.sub(p2, p1);
//		double[] dir2 = VectorOps.sub(p3, p1);
//		double[] dir3 = VectorOps.sub(p4, p1);
//		double[] dir2CrossDir3 = VectorOps.crossProduct(dir2, dir3);
//
//		return VectorOps.product(dir1, dir2CrossDir3)/6.;
		
		return ((p2[0]-p1[0])*((p3[1]-p1[1])*(p4[2]-p1[2])-(p3[2]-p1[2])*(p4[1]-p1[1]))
		          +(p2[1]-p1[1])*((p3[2]-p1[2])*(p4[0]-p1[0])-(p3[0]-p1[0])*(p4[2]-p1[2]))
		          +(p2[2]-p1[2])*((p3[0]-p1[0])*(p4[1]-p1[1])-(p3[1]-p1[1])*(p4[0]-p1[0])))*0.16666667;
	}
	
	private class RbvInfo{
		Atom atom;
		//Packed arrays
		// [0 1 2]
		// [3 4 5]
		// [6 7 8]
		float[] nyeTensor;
		float[] lcm;
		
		private synchronized void makeNyeTensor() {
			if (lcm==null) calculateLcm();
			if (nyeTensor==null) calculateNye();
		}
		
		private double[] interpolateNye(double x, double y, double z, double distSqr, Atom[] possibleNeighbors, float[] norm){
			Atom[] a = new Atom[possibleNeighbors.length];
			double[] distances = new double[possibleNeighbors.length];
			
			int numNeight = 0;
			double sum = 0.;
			
			for (int i=0; i<possibleNeighbors.length;i++){
				double d_x = x-possibleNeighbors[i].getX();
				double d_y = y-possibleNeighbors[i].getY();
				double d_z = z-possibleNeighbors[i].getZ();
				
				if (pbc_x){
					if (d_x < -halfBound_x) d_x += bounds[0];
					else if (d_x > halfBound_x) d_x -= bounds[0];
				}
				if (pbc_y){
					if (d_y < -halfBound_y) d_y += bounds[1];
					else if (d_y > halfBound_y) d_y -= bounds[1];
				}
				if (pbc_z){
					if (d_z < -halfBound_z) d_z += bounds[2];
					else if (d_z > halfBound_z) d_z -= bounds[2];
				}
				
				double dis = d_x*d_x + d_y*d_y + d_z*d_z; 
				if ( dis < distSqr){
					a[numNeight] = possibleNeighbors[i];
					distances[numNeight] = 1./dis;
					sum += distances[numNeight];
					numNeight++;
				}
			}
			
			double[] nyeInter = new double[3];
			if (numNeight>0) {
				for (int i=0; i<numNeight; i++){
					float[] nyeTensor = atomToRbvInfoMap.get(a[i]).nyeTensor;
					
					nyeInter[0] += (norm[0]*nyeTensor[0] + norm[1]*nyeTensor[3] + norm[2]*nyeTensor[6]) * distances[i]; 
					nyeInter[1] += (norm[0]*nyeTensor[1] + norm[1]*nyeTensor[4] + norm[2]*nyeTensor[7]) * distances[i];
					nyeInter[2] += (norm[0]*nyeTensor[2] + norm[1]*nyeTensor[5] + norm[2]*nyeTensor[8]) * distances[i];
				}
				double d = 1./sum;
				nyeInter[0] *= d;
				nyeInter[1] *= d;
				nyeInter[2] *= d;
			} else return null;
			return nyeInter;
		}
		
		private double[] interpolateLineDirection(double x, double y, double z, double distSqr, Atom[] possibleNeighbors){
			Atom[] a = new Atom[possibleNeighbors.length];
			double[] distances = new double[possibleNeighbors.length];
			
			int numNeight = 0;
			double sum = 0.;
			
			for (int i=0; i<possibleNeighbors.length;i++){
				double d_x = x-possibleNeighbors[i].getX();
				double d_y = y-possibleNeighbors[i].getY();
				double d_z = z-possibleNeighbors[i].getZ();
				
				if (pbc_x){
					if (d_x < -halfBound_x) d_x += bounds[0];
					else if (d_x > halfBound_x) d_x -= bounds[0];
				}
				if (pbc_y){
					if (d_y < -halfBound_y) d_y += bounds[1];
					else if (d_y > halfBound_y) d_y -= bounds[1];
				}
				if (pbc_z){
					if (d_z < -halfBound_z) d_z += bounds[2];
					else if (d_z > halfBound_z) d_z -= bounds[2];
				}
				
				double dis = d_x*d_x + d_y*d_y + d_z*d_z; 
				if ( dis < distSqr){
					a[numNeight] = possibleNeighbors[i];
					distances[numNeight] = 1./dis;
					sum += distances[numNeight];
					numNeight++;
				}
			}
			
			double[] ls = new double[9];
			
			if (numNeight>0){
				for (int i=0; i<numNeight; i++){
					RbvInfo info = atomToRbvInfoMap.get(a[i]);
					
					if (info == null){
						info = new RbvInfo();
						info.atom = a[i];
						atomToRbvInfoMap.put(a[i], info);
					}
					
					float[] nyeTensor = info.nyeTensor;
					if (nyeTensor == null){
						info.makeNyeTensor();
						nyeTensor = info.nyeTensor;
					}
					ls[0] += (nyeTensor[0]) * distances[i]; 
					ls[1] += (nyeTensor[3]) * distances[i];
					ls[2] += (nyeTensor[6]) * distances[i];
					
					ls[3] += (nyeTensor[1]) * distances[i]; 
					ls[4] += (nyeTensor[4]) * distances[i];
					ls[5] += (nyeTensor[7]) * distances[i];
					
					ls[6] += (nyeTensor[2]) * distances[i]; 
					ls[7] += (nyeTensor[5]) * distances[i];
					ls[8] += (nyeTensor[8]) * distances[i];
				}
				double d = 1./sum;
				ls[0] *= d; ls[3] *= d; ls[6] *= d;
				ls[1] *= d; ls[4] *= d; ls[7] *= d;
				ls[2] *= d; ls[5] *= d; ls[8] *= d;
			} else return null;
			return ls;
		}
		
		public void calculateBurgersVector(double nnDist) {
			float[] bv;
			float[] lineDirection;
			
			//get all first and second order neighbors 
			ArrayList<Atom> atomsList = new ArrayList<Atom>();
			this.makeNyeTensor();
			atomsList.add(atom);
			for (int i=0; i<atom.getNeigh().length; i++){
				atomsList.add(atom.getNeigh()[i]);
				
				RbvInfo info = atomToRbvInfoMap.get(atom.getNeigh()[i]);
				if (info == null){
					info = new RbvInfo();
					info.atom = atom.getNeigh()[i];
					atomToRbvInfoMap.put(atom.getNeigh()[i], info);
				}
				
				info.makeNyeTensor();
			}
			for (int i=0; i<atom.getNeigh().length; i++){
				Atom a = this.atom.getNeigh()[i];
				if (a.getNeigh()!=null)
				for (int j=0; j<a.getNeigh().length; j++){
					if (!atomsList.contains(a.getNeigh()[j])){
						atomsList.add(a.getNeigh()[j]);
						RbvInfo info = atomToRbvInfoMap.get(a.getNeigh()[j]);
						if (info == null){
							info = new RbvInfo();
							info.atom = a.getNeigh()[j];
							atomToRbvInfoMap.put(a.getNeigh()[j], info);
						}
						
						info.makeNyeTensor();
					}
				}
			}
			
			Atom[] atoms = atomsList.toArray(new Atom[atomsList.size()]);
			
			//Test assumed Burgers vector (1,0,0) (in cartesian space!)
			lineDirection = calculateLineDirection(nnDist, atoms, new int[]{1,0,0});
			
			if (lineDirection == null) return;
			
			//Test alternating assumed Burgers vector if vector is too small 
			if (VectorOps.lengthSqr(lineDirection)<0.1f)
				lineDirection = calculateLineDirection(nnDist, atoms, new int[]{0,1,0});
			
			//Test alternating assumed Burgers vector if vector is too small 
			if (VectorOps.lengthSqr(lineDirection)<0.1f)
				lineDirection = calculateLineDirection(nnDist, atoms, new int[]{0,0,1});
			
			lineDirection = VectorOps.normalize(lineDirection);
			
			bv = calculateBurgersVector(nnDist, atoms, lineDirection);
			if (bv == null) return;
			
			if (VectorOps.lengthSqr(bv) > integrationRadius*integrationRadius*0.02f)
				atom.setRBV(bv, lineDirection);
			
			//local correspondence matrix is needed to calc the nye tensor.
			//because the nye tensors of all neighbors has been calculated
			//the lcm is not needed anymore
//			lcm = null;
		}
		
		private float[] calculateBurgersVector(double nnDist, Atom[] atoms, float[] lineSense) {
			double sqr = nnDist*nnDist;
			
			double[][] nyeInter = new double[12][3];
			for (int i = 0; i<12; i++){
				nyeInter[i] = interpolateNye(atom.getX()+icoVertices[i][0], 
						atom.getY()+icoVertices[i][1],
						atom.getZ()+icoVertices[i][2], sqr, atoms, lineSense);
				if (nyeInter[i] == null) return null;
			}
			float[] bv = new float[3];
			
			for (int i = 0; i<20; i++){
				int v1 = icoFaces[i*3+0];
				int v2 = icoFaces[i*3+1];
				int v3 = icoFaces[i*3+2];
				bv[0] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],
				        VectorOps.add(icoVertices[v1], VectorOps.multiply(icoNormals[v1], nyeInter[v1][0])),
				        VectorOps.add(icoVertices[v2], VectorOps.multiply(icoNormals[v2], nyeInter[v2][0])),
				        VectorOps.add(icoVertices[v3], VectorOps.multiply(icoNormals[v3], nyeInter[v3][0]))
				);
				bv[1] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],
						VectorOps.add(icoVertices[v1], VectorOps.multiply(icoNormals[v1], nyeInter[v1][1])),
				        VectorOps.add(icoVertices[v2], VectorOps.multiply(icoNormals[v2], nyeInter[v2][1])),
				        VectorOps.add(icoVertices[v3], VectorOps.multiply(icoNormals[v3], nyeInter[v3][1]))
		        );
				bv[2] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],
						VectorOps.add(icoVertices[v1], VectorOps.multiply(icoNormals[v1], nyeInter[v1][2])),
				        VectorOps.add(icoVertices[v2], VectorOps.multiply(icoNormals[v2], nyeInter[v2][2])),
				        VectorOps.add(icoVertices[v3], VectorOps.multiply(icoNormals[v3], nyeInter[v3][2]))
		        );
			}
			
			return bv;
		}
		
		private float[] calculateLineDirection(double nnDist, Atom[] atoms, int[] dir) {
			double sqr = nnDist*nnDist;
			
			double[][] lineDirectionInter = new double[12][];
			for (int i = 0; i<12; i++){ 
				lineDirectionInter[i] = interpolateLineDirection(
						atom.getX()+icoVertices[i][0],
						atom.getY()+icoVertices[i][1],
						atom.getZ()+icoVertices[i][2], sqr-0.001, atoms);
				if (lineDirectionInter[i] == null) return null;
			}
			
			float[] ld = new float[3];
			double[] i1 = new double[3];
			double[] i2 = new double[3];
			double[] i3 = new double[3];
			
			for (int i = 0; i<20; i++){
				int v1 = icoFaces[i*3+0];
				int v2 = icoFaces[i*3+1];
				int v3 = icoFaces[i*3+2];
				
				double m = dir[0]*lineDirectionInter[v1][0]+dir[1]*lineDirectionInter[v1][3]+dir[2]*lineDirectionInter[v1][6];
				i1[0] = icoVertices[v1][0]+icoNormals[v1][0]*m; 
				i1[1] = icoVertices[v1][1]+icoNormals[v1][1]*m; 
				i1[2] = icoVertices[v1][2]+icoNormals[v1][2]*m;
				
				i2[0] = icoVertices[v2][0]+icoNormals[v2][0]*m;
				i2[1] = icoVertices[v2][1]+icoNormals[v2][1]*m; 
				i2[2] = icoVertices[v2][2]+icoNormals[v2][2]*m;
				
				i3[0] = icoVertices[v3][0]+icoNormals[v3][0]*m;
				i3[1] = icoVertices[v3][1]+icoNormals[v3][1]*m; 
				i3[2] = icoVertices[v3][2]+icoNormals[v3][2]*m;
				                                                                                                  
				ld[0] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],i1, i2, i3);
				
				
				m = dir[0]*lineDirectionInter[v1][1]+dir[1]*lineDirectionInter[v1][4]+dir[2]*lineDirectionInter[v1][7];
				i1[0] = icoVertices[v1][0]+icoNormals[v1][0]*m; 
				i1[1] = icoVertices[v1][1]+icoNormals[v1][1]*m; 
				i1[2] = icoVertices[v1][2]+icoNormals[v1][2]*m;
				
				i2[0] = icoVertices[v2][0]+icoNormals[v2][0]*m;
				i2[1] = icoVertices[v2][1]+icoNormals[v2][1]*m; 
				i2[2] = icoVertices[v2][2]+icoNormals[v2][2]*m;
				
				i3[0] = icoVertices[v3][0]+icoNormals[v3][0]*m;
				i3[1] = icoVertices[v3][1]+icoNormals[v3][1]*m; 
				i3[2] = icoVertices[v3][2]+icoNormals[v3][2]*m;
				
				ld[1] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],i1, i2, i3);
				
				
				m = dir[0]*lineDirectionInter[v1][2]+dir[1]*lineDirectionInter[v1][5]+dir[2]*lineDirectionInter[v1][8];
				i1[0] = icoVertices[v1][0]+icoNormals[v1][0]*m; 
				i1[1] = icoVertices[v1][1]+icoNormals[v1][1]*m; 
				i1[2] = icoVertices[v1][2]+icoNormals[v1][2]*m;
				
				i2[0] = icoVertices[v2][0]+icoNormals[v2][0]*m;
				i2[1] = icoVertices[v2][1]+icoNormals[v2][1]*m; 
				i2[2] = icoVertices[v2][2]+icoNormals[v2][2]*m;
				
				i3[0] = icoVertices[v3][0]+icoNormals[v3][0]*m;
				i3[1] = icoVertices[v3][1]+icoNormals[v3][1]*m; 
				i3[2] = icoVertices[v3][2]+icoNormals[v3][2]*m;
				
				ld[2] += integral(icoVertices[v1], icoVertices[v2], icoVertices[v3],i1, i2, i3);
			}

			return ld;
		}
		
		private synchronized void calculateLcm(){
			if (lcm!=null) return;
			if (atom.getNeigh() == null) Configuration.getCrystalStructure().buildNeighborsForRBV(atom, nnb);
			float[][] neigh = new float[atom.getNeigh().length][3];
			
			for (int i=0; i<atom.getNeigh().length; i++){
				neigh[i][0] = atom.getNeigh()[i].getX() - atom.getX(); 
				neigh[i][1] = atom.getNeigh()[i].getY() - atom.getY();
				neigh[i][2] = atom.getNeigh()[i].getZ() - atom.getZ();
				
				if (pbc_x){
					if (neigh[i][0] < -halfBound_x)
						neigh[i][0] += bounds[0];
					else if (neigh[i][0] > halfBound_x)
						neigh[i][0] -= bounds[0];
				}
				if (pbc_y){
					if (neigh[i][1] < -halfBound_y)
						neigh[i][1] += bounds[1];
					else if (neigh[i][1] > halfBound_y)
						neigh[i][1] -= bounds[1];
				}
				if (pbc_z){
					if (neigh[i][2] < -halfBound_z)
						neigh[i][2] += bounds[2];
					else if (neigh[i][2] > halfBound_z)
						neigh[i][2] -= bounds[2];
				}
			}
			
//			Matrix a = new Matrix(3, 3, 0.);
//			Matrix b = new Matrix(3, 3, 0.);
			
			double[] a = new double[9];
			double[] b = new double[9];

			for (int i=0; i<neigh.length; i++){
				float bestAngle = -1;
				int best = 0;
				float[] n = neigh[i];
				float l = VectorOps.length(n);
				for (int j=0; j<neighPerf.length; j++){
					float angle = VectorOps.product(n, neighPerf[j]) / (l* pnl[j]);
					if (angle>bestAngle){
						best = j;
						bestAngle = angle;
					}
				}
				if (bestAngle>PHI_MAX){
//					Matrix nm = new Matrix(new double[][]{{n[0],n[1],n[2]}});
//					Matrix p = new Matrix(new double[][] { { neighPerf.get(best, 0), neighPerf.get(best, 1),
//							neighPerf.get(best, 2) } });
//					b.plusEquals(nm.transpose().times(p));
//					a.plusEquals(nm.transpose().times(nm));
					
					a[0] += n[0] * neighPerf[best][0]; a[1] += n[0] * neighPerf[best][1]; a[2] += n[0] * neighPerf[best][2];
					a[3] += n[1] * neighPerf[best][0]; a[4] += n[1] * neighPerf[best][1]; a[5] += n[1] * neighPerf[best][2];
					a[6] += n[2] * neighPerf[best][0]; a[7] += n[2] * neighPerf[best][1]; a[8] += n[2] * neighPerf[best][2];
					
					b[0] += n[0] * n[0]; b[1] += n[0] * n[1]; b[2] += n[0] * n[2];
					b[3] += n[1] * n[0]; b[4] += n[1] * n[1]; b[5] += n[1] * n[2];
					b[6] += n[2] * n[0]; b[7] += n[2] * n[1]; b[8] += n[2] * n[2];
				}
			}
			
//			lcm = a.inverse().times(b);
			if (MatrixOps.invert3x3matrix(a, 0.001)){
				//lcm = new Matrix(a).times(new Matrix(b)).getArray();
				
				lcm = new float[9];
				lcm[0] = (float)(a[0] * b[0] + a[1] * b[3] +a[2] * b[6]);
				lcm[1] = (float)(a[0] * b[1] + a[1] * b[4] +a[2] * b[7]);
				lcm[2] = (float)(a[0] * b[2] + a[1] * b[5] +a[2] * b[8]);
				         
				lcm[3] = (float)(a[3] * b[0] + a[4] * b[3] +a[5] * b[6]);
				lcm[4] = (float)(a[3] * b[1] + a[4] * b[4] +a[5] * b[7]);
				lcm[5] = (float)(a[3] * b[2] + a[4] * b[5] +a[5] * b[8]);
				         
				lcm[6] = (float)(a[6] * b[0] + a[7] * b[3] +a[8] * b[6]);
				lcm[7] = (float)(a[6] * b[1] + a[7] * b[4] +a[8] * b[7]);
				lcm[8] = (float)(a[6] * b[2] + a[7] * b[5] +a[8] * b[8]);
			}
			else lcm = new float[]{1f,0f,0f,0f,1f,0f,0f,0f,1f};
		}
		
		private void calculateNye(){
			double[][] neigh = new double[atom.getNeigh().length][3];
			
			for (int i=0; i<atom.getNeigh().length; i++){
				neigh[i][0] = atom.getNeigh()[i].getX() - atom.getX(); 
				neigh[i][1] = atom.getNeigh()[i].getY() - atom.getY();
				neigh[i][2] = atom.getNeigh()[i].getZ() - atom.getZ();
				
				if (pbc_x){
					if (neigh[i][0] < -halfBound_x)
						neigh[i][0] += bounds[0];
					else if (neigh[i][0] > halfBound_x)
						neigh[i][0] -= bounds[0];
				}
				if (pbc_y){
					if (neigh[i][1] < -halfBound_y)
						neigh[i][1] += bounds[1];
					else if (neigh[i][1] > halfBound_y)
						neigh[i][1] -= bounds[1];
				}
				if (pbc_z){
					if (neigh[i][2] < -halfBound_z)
						neigh[i][2] += bounds[2];
					else if (neigh[i][2] > halfBound_z)
						neigh[i][2] -= bounds[2];
				}
			}
			
			double[][] neighT = new double[atom.getNeigh().length][9];
			for (int i=0; i<atom.getNeigh().length; i++){
				//neighT[i]= neigh[i].transpose().times(neigh[i]);
				
				neighT[i][0] = neigh[i][0] * neigh[i][0]; 
				neighT[i][1] = neigh[i][0] * neigh[i][1]; 
				neighT[i][2] = neigh[i][0] * neigh[i][2];
				
				neighT[i][3] = neigh[i][1] * neigh[i][0]; 
				neighT[i][4] = neigh[i][1] * neigh[i][1];
				neighT[i][5] = neigh[i][1] * neigh[i][2];
				
				neighT[i][6] = neigh[i][2] * neigh[i][0];
				neighT[i][7] = neigh[i][2] * neigh[i][1];
				neighT[i][8] = neigh[i][2] * neigh[i][2];
			}
			
			
			double[] grd = new double[27];
			double[] a = new double[9];
			
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					Arrays.fill(a, 0.);
					double c0 = 0., c1 = 0., c2 = 0.;
					double e0 = lcm[i*3+j];
					for (int k=0; k<atom.getNeigh().length; k++){
//						a.plusEquals(neigh[k].transpose().times(neigh[k]));
						//a.plusEquals(neighT[k]);
						a[0] += neighT[k][0];
						a[1] += neighT[k][1];
						a[2] += neighT[k][2];
						                   
						a[3] += neighT[k][3];
						a[4] += neighT[k][4];
						a[5] += neighT[k][5];
						                   
						a[6] += neighT[k][6];
						a[7] += neighT[k][7];
						a[8] += neighT[k][8];
						RbvInfo info = atomToRbvInfoMap.get(atom.getNeigh()[k]);
						if (info == null){
							info = new RbvInfo();
							info.atom = atom.getNeigh()[k];
							atomToRbvInfoMap.put(atom.getNeigh()[k], info);
						}
						if (info.lcm == null) info.calculateLcm();
						double de = info.lcm[i*3+j] - e0;
//						c.plusEquals(neigh[k].times(de));
						c0 += neigh[k][0]*de;
						c1 += neigh[k][1]*de;
						c2 += neigh[k][2]*de;
					}
									
//					Matrix aa = new Matrix(new double[][]{c}).times(new Matrix(a).inverse());
//					grd[i][j][0] = aa.get(0, 0);
//					grd[i][j][1] = aa.get(0, 1);
					
					MatrixOps.invert3x3matrix(a, 0.001);
					grd[i*9 + j*3 + 0] = c0*a[0] + c1*a[1] + c2*a[2];
					grd[i*9 + j*3 + 1] = c0*a[3] + c1*a[4] + c2*a[5];
					grd[i*9 + j*3 + 2] = c0*a[6] + c1*a[7] + c2*a[8];
				}
			}
			
			nyeTensor = new float[9];
//			for (int i=0; i<3; i++){
//				nyeTensor[i] =   (float)(-grd[18 + i*3 + 1] + grd[9 + i*3 + 2]);
//				nyeTensor[3+i] = (float)(-grd[ 0 + i*3 + 2] + grd[18+ i*3 + 0]);
//				nyeTensor[6+i] = (float)(-grd[ 9 + i*3 + 0] + grd[0 + i*3 + 1]);
//		    }
			
			nyeTensor[0] = (float)(grd[11] - grd[19]);
			nyeTensor[1] = (float)(grd[14] - grd[22]);
			nyeTensor[2] = (float)(grd[17] - grd[25]);
			nyeTensor[3] = (float)(grd[18] - grd[2] );
			nyeTensor[4] = (float)(grd[21] - grd[5] );
			nyeTensor[5] = (float)(grd[24] - grd[8] );
			nyeTensor[6] = (float)(grd[1]  - grd[9] );
			nyeTensor[7] = (float)(grd[4]  - grd[12]);
			nyeTensor[8] = (float)(grd[7]  - grd[15]);
		}
	}
	
	private class AnalyseCallable implements Callable<Void> {
		private int start, end;
	
		private List<RbvInfo> infos;
		
		public AnalyseCallable(int start, int end, List<RbvInfo> atoms) {
			this.start = start;
			this.end = end;
			this.infos = atoms;
		}

		public Void call() throws Exception{
			for (int i = start; i < end; i++) {
				if (Thread.interrupted()) return null;
				RbvInfo info = infos.get(i); 
				info.calculateBurgersVector(integrationRadius);
			} 
			return null;
		}
	}
	
}
