// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.dislocationDensity;

import gui.ViewerGLJPanel;
import gui.ViewerGLJPanel.RenderOption;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DecimalFormat;

import javax.media.opengl.GL2;
import javax.swing.*;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import common.ColorTable;

import model.AtomData;
import model.Configuration;
import model.ImportStates;

public class DislocationDensityTensorData {

	private DislocationDensityTensor[][][] ddtFromLatt;
	private DislocationDensityTensor[][][] ddtFromDis;
	
	private DislocationDensityTensor plasticZone;
	
	private int rasterX, rasterY, rasterZ;
	
	public DislocationDensityTensorData(AtomData data) {
		int[] slices = Configuration.getDislocationDensitySlices();
		if (ImportStates.SKELETONIZE.isActive()){
			ddtFromDis = calcDislocationDensityTensors(data, slices[0], slices[1], slices[2]);
			
			if (data.getFileMetaData("extpot") != null && data.getFileMetaData("plasticzone") != null){
				float[] indent = (float[]) data.getFileMetaData("extpot");
				float[] pz = (float[]) data.getFileMetaData("plasticzone");
				
				plasticZone = new DislocationDensityTensor(data.getSkeletonizer(), 
						new HalfShellSector(pz[3], new float[]{indent[0],indent[1],indent[2]},
								indent[3], pz[4], new float[]{pz[0],pz[1],pz[2]}));
			}
		}
		if (ImportStates.LATTICE_ROTATION.isActive()){
			CuboidSectorDensityTensorBuilder csdtb = new CuboidSectorDensityTensorBuilder(data, slices[0], slices[1], slices[2]);
			ddtFromLatt = csdtb.createCuboids();
		}
	}
	
	public DislocationDensityTensor[][][] calcDislocationDensityTensors(AtomData data, int rasterX, int rasterY, int rasterZ){
		this.rasterX = rasterX;
		this.rasterY = rasterY;
		this.rasterZ = rasterZ;
		
		CuboidVolumeElement[][][] ve = new CuboidVolumeElement[rasterX][rasterY][rasterZ];
		
		for (int i=0; i<rasterX; i++){
			for (int j=0;j<rasterY; j++){
				for (int k=0;k<rasterZ; k++){
					float[] low = new float[3];
					float[] high = new float[3];
					low[0] = (data.getBounds()[0]/rasterX)*i;
					low[1] = (data.getBounds()[1]/rasterY)*j;
					low[2] = (data.getBounds()[2]/rasterZ)*k;
					
					high[0] = (data.getBounds()[0]/rasterX)*(i+1);
					high[1] = (data.getBounds()[1]/rasterY)*(j+1);
					high[2] = (data.getBounds()[2]/rasterZ)*(k+1);
					
					ve[i][j][k] = new CuboidVolumeElement(high, low);	
				}
				
			}
		}
		
		DislocationDensityTensor[][][] ddt = new DislocationDensityTensor[rasterX][rasterY][rasterZ];
		for (int i=0; i<rasterX; i++){
			for (int j=0;j<rasterY; j++){
				for (int k=0;k<rasterZ; k++){
					DislocationDensityTensor d = new DislocationDensityTensor(data.getSkeletonizer(), ve[i][j][k]); 
					ddt[i][j][k] = d;
				}
			}
		}
		
		return ddt;
	}
	
	public DislocationDensityTensor[][][] getDdtFromDislocation() {
		return ddtFromDis;
	}
	
	public DislocationDensityTensor[][][] getDdtFromLattice() {
		return ddtFromLatt;
	}
	
	public static class JDislocationDensityTensorControls extends JPanel {
		private enum DDTPlane { X_PLANE, Y_PLANE, Z_PLANE, OFF};
		private static final long serialVersionUID = 1L;
		
		private JCheckBox enableCheckBox = new JCheckBox("Show DDT");
		private JComboBox axisComboBox = new JComboBox(new String[]{"X-Axis","Y-Axis","Z-Axis","Off"});
		private JRadioButton[][] tensorComp = new JRadioButton[3][3];
		private JRadioButton tensorSum;
		private JRadioButton scalarDens;
		
		private JFormattedTextField lowerLimitTextBox = new JFormattedTextField(new DecimalFormat());
		private JFormattedTextField upperLimitTextBox = new JFormattedTextField(new DecimalFormat());
		
		private JRadioButton latticeDDTComboBox = new JRadioButton("DDT from lattice");
		private JRadioButton dislocationNetworkDDTComboBox = new JRadioButton("DDT from dislocation network");
		
		private ViewerGLJPanel viewer;
		private AtomData atomData;
		private DislocationDensityTensorData ddtd;
		
		private boolean lattOrDis = false;
		private DDTPlane ddtPlane = DDTPlane.X_PLANE;
		private int ddtCompX = 0, ddtCompY = 0;
		private int renderStartSlice, renderEndSlice;
		
		
		private static JDislocationDensityTensorControls control;
		public static JDislocationDensityTensorControls getInstance(){
			if (control == null)
				control = new JDislocationDensityTensorControls();
			return control;
		}
		
		private JDislocationDensityTensorControls() {
			this.setBorder(new TitledBorder(new EtchedBorder(1), "DDT"));
			this.setLayout(new GridBagLayout());
			GridBagConstraints gbc = new GridBagConstraints();
			
			ActionListener al = new ActionListener(){
				@Override
				public void actionPerformed(ActionEvent arg0) {
					int i = Integer.parseInt(arg0.getActionCommand());
					if (i < 0){
						ddtCompX = i;
						ddtCompY = 0;
					} else {
						ddtCompX = i/3;
						ddtCompY = i%3;
					}
					viewer.reDraw();
				}
			};
			
			enableCheckBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					RenderOption.DDT.setEnabled(enableCheckBox.isSelected());
				}
			});
			
			gbc.fill = GridBagConstraints.BOTH;
			gbc.gridx = 0; gbc.gridy = 0;
			gbc.gridwidth = 3;
			this.add(enableCheckBox, gbc);
			
			gbc.gridy++; gbc.gridwidth = 1;
			
			gbc.weightx = 0.33;
			ButtonGroup bg = new ButtonGroup();
			for (int i=0; i<3; i++){
				String s1;
				if (i==0) s1="x";
				else if (i==1) s1="y";
				else s1="z";
				for (int j=0; j<3; j++){
					String s2;
					if (j==0) s2="x";
					else if (j==1) s2="y";
					else s2="z";
					
					tensorComp[i][j] = new JRadioButton(String.format("%s,%s", s1, s2));
					tensorComp[i][j].setActionCommand(Integer.toString(3*i+j));
					tensorComp[i][j].addActionListener(al);
					gbc.gridx = j; gbc.gridy = i+1;
					bg.add(tensorComp[i][j]);
					this.add(tensorComp[i][j],gbc);
				}
			}
			
			gbc.gridx = 0; gbc.gridy +=3;
			gbc.gridwidth = 3;
			tensorSum = new JRadioButton("GND (sum of absolutes)");
			tensorSum.setActionCommand("-1");
			tensorSum.addActionListener(al);
			this.add(tensorSum,gbc); gbc.gridy++;
			bg.add(tensorSum);
			scalarDens = new JRadioButton("Dis. density (ignore BV)");
			scalarDens.setActionCommand("-2");
			scalarDens.addActionListener(al);
			this.add(scalarDens,gbc);
			bg.add(scalarDens);
			
			gbc.gridx = 0; gbc.gridy++;
			this.add(axisComboBox, gbc);
			axisComboBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					if (ddtd==null) return;
					int selected = axisComboBox.getSelectedIndex();
					if (selected == 0){
						ddtPlane = DDTPlane.X_PLANE;
						renderStartSlice = 0;
						renderEndSlice = ddtd.rasterX;
						lowerLimitTextBox.setValue(0);
						upperLimitTextBox.setValue(ddtd.rasterX);
					}
					else if (selected == 1){
						ddtPlane = DDTPlane.Y_PLANE;
						renderStartSlice = 0;
						renderEndSlice = ddtd.rasterY;
						lowerLimitTextBox.setValue(0);
						upperLimitTextBox.setValue(ddtd.rasterY);
					}
					else if (selected == 2){
						ddtPlane = DDTPlane.Z_PLANE;
						renderStartSlice = 0;
						renderEndSlice = ddtd.rasterZ;
						lowerLimitTextBox.setValue(0);
						upperLimitTextBox.setValue(ddtd.rasterZ);
					}
					else 
						ddtPlane = DDTPlane.OFF;
					viewer.reDraw();
				}
			});
			
			gbc.gridwidth = 1;
			gbc.gridy++;
			gbc.gridx = 0;
			this.add(new JLabel("min/max slide"),gbc);
			
			gbc.gridx = 1;
			upperLimitTextBox.addPropertyChangeListener(new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent arg0) {
					if (upperLimitTextBox.getValue()==null) return;
					if (ddtd==null) return;
					Number l = (Number)upperLimitTextBox.getValue();
					renderEndSlice = l.intValue();
					viewer.repaint();
				}
			});
			lowerLimitTextBox.addPropertyChangeListener(new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent arg0) {
					if (lowerLimitTextBox.getValue()==null) return;
					if (ddtd==null) return;
					Number l = (Number)lowerLimitTextBox.getValue();
					renderStartSlice = l.intValue();
					viewer.repaint();
				}
			});
			this.add(lowerLimitTextBox,gbc);
			gbc.gridx = 2;
			this.add(upperLimitTextBox,gbc);
			
			gbc.gridx = 0;
			gbc.gridwidth = 3;
			bg = new ButtonGroup();
			bg.add(dislocationNetworkDDTComboBox);
			bg.add(latticeDDTComboBox);
			gbc.gridy++;
			this.add(dislocationNetworkDDTComboBox, gbc);
			dislocationNetworkDDTComboBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					lattOrDis = true;
					viewer.reDraw();
				}
			});
			gbc.gridy++;
			this.add(latticeDDTComboBox, gbc);
			latticeDDTComboBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					lattOrDis = false;
					viewer.reDraw();
				}
			});
		}
		
		public void setViewer(ViewerGLJPanel viewer, AtomData data, boolean init){
			ddtd = data.getDislocationDensityData();
			this.atomData = data;
			
			if (this.viewer==null){
				this.viewer = viewer;
				axisComboBox.setSelectedIndex(0);
				tensorComp[0][0].doClick();
			}
			
			if (init){
				dislocationNetworkDDTComboBox.setEnabled(ImportStates.SKELETONIZE.isActive() && ImportStates.BURGERS_VECTORS.isActive());
				latticeDDTComboBox.setEnabled(ImportStates.LATTICE_ROTATION.isActive());
				enableCheckBox.setEnabled(ImportStates.SKELETONIZE.isActive() && ImportStates.BURGERS_VECTORS.isActive()
						|| ImportStates.LATTICE_ROTATION.isActive());
				if (ImportStates.SKELETONIZE.isActive()) dislocationNetworkDDTComboBox.doClick();
				else if(ImportStates.LATTICE_ROTATION.isActive()) latticeDDTComboBox.doClick();
			}
			this.repaint();
		}
		
		
		public float[] renderDDT(GL2 gl, boolean picking){
			if (ddtPlane == DDTPlane.OFF) return null;
			
			DislocationDensityTensorData ddtd = atomData.getDislocationDensityData();
			DislocationDensityTensor[][][] ddt;
			if (lattOrDis) ddt = ddtd.getDdtFromDislocation();
			else ddt = ddtd.getDdtFromLattice();
			
			float minPower = 14.5f;
			float maxPower = 17f;
			
			if (!picking) gl.glDepthMask(false);
			for (int i=0; i<ddt.length; i++){
				for (int j=0; j<ddt[i].length; j++){
					for (int k=0; k<ddt[i][j].length; k++){
						if (ddtPlane == DDTPlane.X_PLANE){
							if (i<renderStartSlice || i>renderEndSlice) continue;
						} else if (ddtPlane == DDTPlane.Y_PLANE){
							if (j<renderStartSlice || j>renderEndSlice) continue;
						} else {
							if (k<renderStartSlice || k>renderEndSlice) continue;
						}
						
						float c1;
						
						if (ddtCompX==-1){
							c1 = (float)Math.log10(Math.abs(ddt[i][j][k].getDensity()));
							maxPower = 18f;
							minPower = 15f;
						} else if (ddtCompX==-2){
							c1 = (float)Math.log10(Math.abs(ddt[i][j][k].getScalarDensity()));
							maxPower = 18f;
							minPower = 15f;
						} else {
							c1 = (float)Math.log10(Math.abs(ddt[i][j][k].getDensityTensor()[ddtCompX][ddtCompY]));
							maxPower = 17f;
							minPower = 14.5f;
						}
						
						VolumeElement ve = ddt[i][j][k].getVolumeElement();
						float[] f = ColorTable.getIntensityGLColor(minPower, maxPower, c1, 0.6f);
						
						if (c1>0) {
							if (picking) {
								viewer.getNextPicking(gl, ddt[i][j][k]);
							}
							gl.glColor4f(f[0], f[1], f[2], f[3]);
							ve.render(gl, picking);
						}
					}
				}
			}
			
			//Plastic zone rendering
			if (ddtd.plasticZone != null){
				float c1;
				if (ddtCompX==-1){
					c1 = (float)Math.log10(Math.abs(ddtd.plasticZone.getDensity()));
					maxPower = 18f;
					minPower = 15f;
				} else if (ddtCompX==-2){
					c1 = (float)Math.log10(Math.abs(ddtd.plasticZone.getScalarDensity()));
					maxPower = 18f;
					minPower = 15f;
				} else {
					c1 = (float)Math.log10(Math.abs(ddtd.plasticZone.getDensityTensor()[ddtCompX][ddtCompY]));
					maxPower = 17f;
					minPower = 14.5f;
				}
				
				VolumeElement ve = ddtd.plasticZone.getVolumeElement();
				float[] f = ColorTable.getIntensityGLColor(minPower, maxPower, c1, 0.1f);
				
				if (picking) {
					viewer.getNextPicking(gl, ddtd.plasticZone);
				}
				gl.glColor4f(f[0], f[1], f[2], f[3]);
				ve.render(gl, picking);
				gl.glNormal3f(0f, 0f, 1f);
			}
			
			gl.glDepthMask(true);
			return new float[]{minPower, maxPower};
		}
	}
}
