// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.polygrain.mesh;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;

import javax.media.opengl.GL2;

import common.VectorOps;

public class FinalMesh {

	private int[] triangles;
	private float[] vertices;
	private double volume = -1.;
	private float[] normals;
	
	public FinalMesh(Collection<Vertex> vertices, Collection<Triangle> triangles){
		this.triangles = new int[triangles.size()*3];
		this.vertices = new float[vertices.size()*3];
		
		HashMap<Vertex, Integer> vertexMap = new HashMap<Vertex, Integer>();
		
		int i=0;
		for (Vertex v : vertices){
			vertexMap.put(v,i);
			this.vertices[i+0] = v.c[0];
			this.vertices[i+1] = v.c[1];
			this.vertices[i+2] = v.c[2];
			i+=3;
		}
		
		i=0;
		for (Triangle t : triangles){
			this.triangles[i+0] = vertexMap.get(t.neighborEdge.vertexEnd);
			HalfEdge next = t.neighborEdge.next;
			this.triangles[i+1] = vertexMap.get(next.vertexEnd);
			next = next.next;
			this.triangles[i+2] = vertexMap.get(next.vertexEnd);
			i+=3;
		}
		
		makeNormals();
	}
	
	public FinalMesh(int[] triangles, float[] vertices){
		this.triangles = triangles;
		this.vertices = vertices;
		
		makeNormals();
	}
	
	private void makeNormals(){
		this.normals = new float[vertices.length];
		
		for (int i=0; i<triangles.length; i+=3){
			float[] n = getTriangleUnitNormal(i);
			int k = triangles[i+0];
			normals[k] += n[0]; normals[k+1] += n[1]; normals[k+2] += n[2];
			k = triangles[i+1];
			normals[k] += n[0]; normals[k+1] += n[1]; normals[k+2] += n[2];
			k = triangles[i+2];
			normals[k] += n[0]; normals[k+1] += n[1]; normals[k+2] += n[2];
		}
		
		for (int i=0; i<this.normals.length; i+=3){
			float l = (float)(1./Math.sqrt(this.normals[i+0]*this.normals[i+0] + this.normals[i+1]*this.normals[i+1] + 
					this.normals[i+2]*this.normals[i+2]));
			this.normals[i+0] *= l;
			this.normals[i+1] *= l;
			this.normals[i+2] *= l;
		}
	}
	
	public void renderMesh(GL2 gl){
		gl.glBegin(GL2.GL_TRIANGLES);
		for (int i=0; i<triangles.length; i+=3){
			int k = triangles[i+0];
			gl.glNormal3f( normals[k],  normals[k+1],  normals[k+2]);
			gl.glVertex3f(vertices[k], vertices[k+1], vertices[k+2]);
			k = triangles[i+1];
			gl.glNormal3f( normals[k],  normals[k+1],  normals[k+2]);
			gl.glVertex3f(vertices[k], vertices[k+1], vertices[k+2]);
			k = triangles[i+2];
			gl.glNormal3f( normals[k],  normals[k+1],  normals[k+2]);
			gl.glVertex3f(vertices[k], vertices[k+1], vertices[k+2]);
		}
		
		gl.glEnd();
	}
	
	private float[] getTriangleUnitNormal(int index){
		float v2_v1_x = vertices[triangles[index+1]+0] - vertices[triangles[index]+0];
		float v2_v1_y = vertices[triangles[index+1]+1] - vertices[triangles[index]+1];
		float v2_v1_z = vertices[triangles[index+1]+2] - vertices[triangles[index]+2];
		
		float v3_v1_x = vertices[triangles[index+2]+0] - vertices[triangles[index]+0];
		float v3_v1_y = vertices[triangles[index+2]+1] - vertices[triangles[index]+1];
		float v3_v1_z = vertices[triangles[index+2]+2] - vertices[triangles[index]+2];
		
		float[] norm =  new float[]{ v2_v1_y*v3_v1_z-v2_v1_z*v3_v1_y,
				v2_v1_z*v3_v1_x-v2_v1_x*v3_v1_z,
				v2_v1_x*v3_v1_y-v2_v1_y*v3_v1_x};
		
		float l = 1f/VectorOps.length(norm);
		norm[0] *= l; norm[1] *= l; norm[2] *= l;
		return norm;
	}
	
	public float getTriangleArea(int index){
		float v2_v1_x = vertices[triangles[index+1]+0] - vertices[triangles[index]+0];
		float v2_v1_y = vertices[triangles[index+1]+1] - vertices[triangles[index]+1];
		float v2_v1_z = vertices[triangles[index+1]+2] - vertices[triangles[index]+2];
		
		float v3_v1_x = vertices[triangles[index+2]+0] - vertices[triangles[index]+0];
		float v3_v1_y = vertices[triangles[index+2]+1] - vertices[triangles[index]+1];
		float v3_v1_z = vertices[triangles[index+2]+2] - vertices[triangles[index]+2];
		
		float[] norm =  new float[]{ v2_v1_y*v3_v1_z-v2_v1_z*v3_v1_y,
				v2_v1_z*v3_v1_x-v2_v1_x*v3_v1_z,
				v2_v1_x*v3_v1_y-v2_v1_y*v3_v1_x};
		
		return 0.5f*VectorOps.length(norm);
	}
	
	public int getTriangleCount(){
		return triangles.length/3;
	}
	
	public double getVolume(){
		if (volume!=-1) return volume;
		
		this.volume = 0.;
		for (int i=0; i<triangles.length; i+=3){
			float[] c = new float[3];
			c[0] = vertices[triangles[i+0]+0];
			c[1] = vertices[triangles[i+0]+1];
			c[2] = vertices[triangles[i+0]+2];
			
			this.volume += getTriangleArea(i)*VectorOps.product(getTriangleUnitNormal(i), c);
		}		
		this.volume /= 3.;
		
		return this.volume;
	}
	
	public void printMetaData(DataOutputStream dos, int number) throws IOException{
		dos.writeBytes("##mesh "+number+"\n");
		dos.writeBytes("###triangles "+triangles.length+" ");
		for (int i=0; i<triangles.length; i++){
			dos.writeBytes(Integer.toString(triangles[i]));
			dos.writeBytes(" ");
		}
		dos.writeBytes("\n");
		
		dos.writeBytes("###vertices "+vertices.length+" ");
		for (int i=0; i<vertices.length; i++){
			dos.writeBytes(Float.toString(vertices[i]));
			dos.writeBytes(" ");
		}
		dos.writeBytes("\n");
	}
}
