// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.polygrain.mesh;

import java.util.ArrayList;

import model.Coordinate;

import common.VectorOps;

public class HalfEdge implements MeshElement{
	Vertex vertexEnd;
	HalfEdge pair;
	Triangle triangle;
	HalfEdge next;
	
	public boolean isContractable(){
		ArrayList<Vertex> adjacentV2 = pair.vertexEnd.getAdjacentVertices();
		int num = 0;
		
		HalfEdge n = vertexEnd.neighborEdge;
		do {
			if (this.next.vertexEnd != n.vertexEnd && this.pair.next.vertexEnd != n.vertexEnd 
					&& adjacentV2.contains(n.vertexEnd)) {
				return false;
			}
			n = n.pair.next;
			num++;
		} while (n!=vertexEnd.neighborEdge);
		
		if (num<=3) return false;
		return true;
	}
	
	public boolean isContractableForGivenPoint(float[] newCoord){
//		if (!isContractable()) return false;
		
		//Test if contraction does not flip face normals
		//Always contract to the node with the higher order
		Vertex v1 = this.vertexEnd;
		Vertex v2 = this.pair.vertexEnd;
		
		//Always contract to the node with the higher order
		if (v2.compareTo(v1)>0){
			v2 = this.pair.vertexEnd;
			v1 = this.vertexEnd;
		}
		
		float[] c = v1.getCoord().clone();
		
		HalfEdge n = v1.neighborEdge;
		do {
			Triangle t1 = n.triangle;
			//Test only triangles, that are not deleted during edge collapse
			if (t1 != this.triangle  && t1 != this.pair.triangle){
				float[] normal1 = t1.getNormalVector();
				//Move node as in an merging process
				v1.getCoord()[0] = newCoord[0];
				v1.getCoord()[1] = newCoord[1];
				v1.getCoord()[2] = newCoord[2];
				
				boolean badlyShaped = t1.isBadlyShaped();
				
				float[] normal2 = t1.getNormalVector();
				//Move node back
				v1.getCoord()[0] = c[0];
				v1.getCoord()[1] = c[1];
				v1.getCoord()[2] = c[2];
				//Test if the face normal has flipped
				if (VectorOps.product(normal1, normal2)<0 || badlyShaped) 
					return false;
			}
			n = n.pair.next;
		} while (n!=v1.neighborEdge);
		
		
		n = v2.neighborEdge;
		c = v2.getCoord().clone();
		
		do {
			Triangle t1 = n.triangle;
			//Test only triangles, that are not deleted during edge collapse
			if (t1 != this.triangle  && t1 != this.pair.triangle){
			
				float[] normal1 = t1.getNormalVector();
				//Move node as in an merging process
				v2.getCoord()[0] = newCoord[0];
				v2.getCoord()[1] = newCoord[1];
				v2.getCoord()[2] = newCoord[2];
				float[] normal2 = t1.getNormalVector();
				
				boolean badlyShaped = t1.isBadlyShaped();
				
				//Move node back
				v2.getCoord()[0] = c[0];
				v2.getCoord()[1] = c[1];
				v2.getCoord()[2] = c[2];
				//Test if the face normal has flipped
				if (VectorOps.product(normal1, normal2)<0 || badlyShaped) 
					return false;
			}
			n = n.pair.next;
		} while (n!=v2.neighborEdge);
		
		return true;
	}
	
	float getLength(){
		return VectorOps.dist(vertexEnd, this.pair.vertexEnd);
	}
	
	@Override
	public float[] getUnitNormalVector() {
		return VectorOps.normalize(getNormalVector());
	}
	
	@Override
	public float[] getNormalVector() {
		float[] t1n = triangle.getUnitNormalVector();
		float[] t2n = pair.triangle.getUnitNormalVector();
		t1n[0] += t2n[0];
		t1n[1] += t2n[1];
		t1n[2] += t2n[2];
		return t1n;
	}
	
	@Override
	public boolean isPointInMesh(Coordinate p){
		//Decide if the local triangles are convex or concave
		float[] norm = triangle.getNormalVector();
		float[] dir = VectorOps.sub(pair.next.vertexEnd.getCoord(), pair.vertexEnd.getCoord());
		
		float d = norm[0]*dir[0] + norm[1]*dir[1] + norm[2]*dir[2]; 
		
		if (d < 0f){
			//convex
			if (triangle.isPointInMesh(p) && pair.triangle.isPointInMesh(p)) return true;
			else return false;
		} else {
			//concave
			if (triangle.isPointInMesh(p) || pair.triangle.isPointInMesh(p)) return true;
			else return false;
		}
	}
}
