// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.polygrain.mesh;
import model.Coordinate;
import common.Tupel;
import common.VectorOps;

public class Triangle implements MeshElement{
	HalfEdge neighborEdge;
	
	public Vertex[] getVertices(){
		Vertex[] v = new Vertex[3];
		v[0] = neighborEdge.vertexEnd;
		HalfEdge next = neighborEdge.next;
		v[1] = next.vertexEnd;
		next = next.next;
		v[2] = next.vertexEnd;
		return v;
	}
	
	public Vertex getVertex(){
		return neighborEdge.vertexEnd;
	}
	
	@Override
	public float[] getUnitNormalVector() {
		float[] u = getNormalVector();
		float l = 1f/VectorOps.length(u);
		u[0] *= l; u[1] *= l; u[2] *= l;
		return u;
	}
	
	@Override
	public float[] getNormalVector() {
		return VectorOps.crossProduct(
				VectorOps.sub(neighborEdge.next.vertexEnd.getCoord(),neighborEdge.vertexEnd.getCoord()), 
				VectorOps.sub(neighborEdge.next.next.vertexEnd.getCoord(),neighborEdge.vertexEnd.getCoord()));
	}
	
	@Override
	public boolean isPointInMesh(Coordinate p){
		float[] n = this.getNormalVector();
		float[] n2 = VectorOps.sub(p, neighborEdge.vertexEnd);
		
		//If the point is opposite of the face, it cannot be inside
		float d = VectorOps.product(n,n2); 
		if ( d > 0f) return false;
		return true;
	}
	
	public float getArea(){
		Vertex[] v = getVertices();
		float[] v1 = VectorOps.sub(v[1].getCoord(), v[0].getCoord());
		float[] v2 = VectorOps.sub(v[2].getCoord(), v[0].getCoord());
		return 0.5f*VectorOps.length(VectorOps.crossProduct(v1, v2));
	}
	
	public Tupel<Float, MeshElement> getMinSqrDistAndClosestObject(Coordinate p, float[] normal){
		MeshElement closest = this;
		float[] E0 = VectorOps.sub(neighborEdge.next.vertexEnd.getCoord(), neighborEdge.vertexEnd.getCoord());
		float[] E1 = VectorOps.sub(neighborEdge.next.next.vertexEnd.getCoord(), neighborEdge.vertexEnd.getCoord());
		float[] D = VectorOps.sub(neighborEdge.vertexEnd, p);
		
		float a = E0[0]*E0[0]+E0[1]*E0[1]+E0[2]*E0[2];
		float b = E0[0]*E1[0]+E0[1]*E1[1]+E0[2]*E1[2];
		float c = E1[0]*E1[0]+E1[1]*E1[1]+E1[2]*E1[2];
		float d = E0[0]*D[0]+E0[1]*D[1]+E0[2]*D[2];
		float e = E1[0]*D[0]+E1[1]*D[1]+E1[2]*D[2];
		
		float det = a*c-b*b;
		float s = b*e-c*d;
		float t = b*d-a*e;
	
		if (s+t<=det){
			if (s < 0f){
				if (t < 0f){
					//closest to vertex
					return new Tupel<Float, MeshElement> (VectorOps.sqrDist(p, neighborEdge.vertexEnd),
							neighborEdge.vertexEnd);
				} else {
					s = 0f;
					t = (e >= 0f ? 0f : (-e >= c ? 1f : -e/c));
					closest = this.neighborEdge;
				}
			} else if (t < 0f){
				t = 0f;
				s = (d >= 0f ? 0f : (-d >= a ? 1f : -d/a));
				closest = this.neighborEdge.next;
			} else {
				float invDet = 1f/det;
				s *= invDet;
				t *= invDet;
			}
		} else {
			if (s < 0){
				//closest to vertex 
				return new Tupel<Float, MeshElement> (VectorOps.sqrDist(p, neighborEdge.next.next.vertexEnd),
						neighborEdge.next.next.vertexEnd);
				
			} else if (t < 0f){
				//closest to vertex
				return new Tupel<Float, MeshElement> (VectorOps.sqrDist(p, neighborEdge.next.vertexEnd),
						neighborEdge.next.vertexEnd);
			} else {
				//region 1
				float numer = c+e-b-d;
				if (numer <= 0f){
					s = 0f;
				} else {
					float denom = a-2f*b+c;
					s = (numer >= denom ? 1f : numer/denom);
				}
				t = 1f-s;
				closest = this.neighborEdge.next.next;
			}
		}
		
		float f = D[0]*D[0]+D[1]*D[1]+D[2]*D[2];
		return new Tupel<Float, MeshElement> (a*s*s+2f*b*s*t+c*t*t+2f*d*s+2f*e*t+f, closest);
	}
	
	public float getMinSqrDist(Coordinate p, float[] normal){
		float[] E0 = VectorOps.sub(neighborEdge.next.vertexEnd.getCoord(), neighborEdge.vertexEnd.getCoord());
		float[] E1 = VectorOps.sub(neighborEdge.next.next.vertexEnd.getCoord(), neighborEdge.vertexEnd.getCoord());
		float[] D = VectorOps.sub(neighborEdge.vertexEnd, p);
		
		float a = E0[0]*E0[0]+E0[1]*E0[1]+E0[2]*E0[2];
		float b = E0[0]*E1[0]+E0[1]*E1[1]+E0[2]*E1[2];
		float c = E1[0]*E1[0]+E1[1]*E1[1]+E1[2]*E1[2];
		float d = E0[0]*D[0]+E0[1]*D[1]+E0[2]*D[2];
		float e = E1[0]*D[0]+E1[1]*D[1]+E1[2]*D[2];
		
		float det = a*c-b*b;
		float s = b*e-c*d;
		float t = b*d-a*e;
	
		if (s+t<=det){
			if (s < 0f){
				if (t < 0f){
					//closest to vertex
					return VectorOps.sqrDist(p, neighborEdge.vertexEnd);
				} else {
					s = 0f;
					t = (e >= 0f ? 0f : (-e >= c ? 1f : -e/c));
				}
			} else if (t < 0f){
				t = 0f;
				s = (d >= 0f ? 0f : (-d >= a ? 1f : -d/a));
			} else {
				float invDet = 1f/det;
				s *= invDet;
				t *= invDet;
			}
		} else {
			if (s < 0){
				//closest to vertex 
				return VectorOps.sqrDist(p, neighborEdge.next.next.vertexEnd);
			} else if (t < 0f){
				//closest to vertex
				return VectorOps.sqrDist(p, neighborEdge.next.vertexEnd);
			} else {
				//region 1
				float numer = c+e-b-d;
				if (numer <= 0f){
					s = 0f;
				} else {
					float denom = a-2f*b+c;
					s = (numer >= denom ? 1f : numer/denom);
				}
				t = 1f-s;
			}
		}
		
		float f = D[0]*D[0]+D[1]*D[1]+D[2]*D[2];
		return a*s*s+2f*b*s*t+c*t*t+2f*d*s+2f*e*t+f;
	}
	
	
	boolean isBadlyShaped(){
		//Test area vs. edge length, length of the basis should not be 10 times larger than the heigth
		float area = getArea();
		float l = neighborEdge.getLength(); 
		if (area/l < l*0.1f) return true;
		l = neighborEdge.next.getLength(); 
		if (area/l < l*0.1f) return true;
		l = neighborEdge.next.next.getLength(); 
		if (area/l < l*0.1f) return true;
		return false;
	}
}
