// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.polygrain.mesh;

import java.util.ArrayList;

import model.Coordinate;
import model.polygrain.kdTree.*;

import common.VectorOps;

public class Vertex implements Coordinate, Comparable<Vertex>, MeshElement{	
	private final int uniqueNumber;
	float[] c;
	HalfEdge neighborEdge;
	
	public Vertex(float[] c, int uniqueNumber){
		this.c = c;
		this.uniqueNumber = uniqueNumber;
	}
	
	@Override
	public float getX() {
		return c[0];
	}
	
	@Override
	public float getY() {
		return c[1];
	}
	
	@Override
	public float getZ() {
		return c[2];
	}
	
	public float[] getCoord() {
		return c;
	}
	
	public ArrayList<Vertex> getAdjacentVertices(){
		ArrayList<Vertex> v = new ArrayList<Vertex>();
		HalfEdge n = neighborEdge;
		do {
			v.add(n.vertexEnd);
			n = n.pair.next;
		} while (n!=neighborEdge);
		return v;
	}
	
	public ArrayList<Triangle> getAdjacentFaces(){
		ArrayList<Triangle> t = new ArrayList<Triangle>();
		HalfEdge n = neighborEdge;
		do {
			t.add(n.triangle);
			n = n.pair.next;
		} while (n!=neighborEdge);
		return t;
	}
	
	public float[] getLaplacianSmoother(){
		int size = 0;
		float[] s = new float[3];
		
		HalfEdge n = neighborEdge;
		do {
			s[0] += n.vertexEnd.c[0];
			s[1] += n.vertexEnd.c[1];
			s[2] += n.vertexEnd.c[2];
			
			n = n.pair.next;
			size++;
		} while (n!=neighborEdge);

		s[0] = s[0]/size - c[0]; s[1] = s[1]/size - c[1]; s[2] = s[2]/size - c[2];
		
		float[] normal = getUnitNormalVector();
		float sn = s[0]*normal[0] + s[1]*normal[1] + s[2]*normal[2]; 
		
		s[0] -= normal[0]*sn;
		s[1] -= normal[1]*sn;
		s[2] -= normal[2]*sn;
		
		return s;
	}
	
	@Override
	public float[] getUnitNormalVector() {
		float[] u = getNormalVector();
		float l = 1f/VectorOps.length(u);
		u[0] *= l; u[1] *= l; u[2] *= l;
		return u;
	}
	
	@Override
	public float[] getNormalVector() {
		float[] normal = new float[3]; 
		HalfEdge n = neighborEdge;
		do {
			float[] no = n.triangle.getUnitNormalVector();
			normal[0] += no[0];
			normal[1] += no[1];
			normal[2] += no[2];
			n = n.pair.next;
		} while (n!=neighborEdge);
		return normal;
	}
	
	@Override
	public boolean isPointInMesh(Coordinate p){
		HalfEdge n = neighborEdge;
		do {
			if (!n.triangle.isPointInMesh(p)) return false;
			n = n.pair.next;
		} while (n!=neighborEdge);
		return true;
	}
	
	public void shrink(KDTree<? extends Coordinate> tree){
		Coordinate p1 = tree.getNearest(this);
		this.c[0] += 0.4f * (p1.getX()-this.c[0]);
		this.c[1] += 0.4f * (p1.getY()-this.c[1]);
		this.c[2] += 0.4f * (p1.getZ()-this.c[2]);
	}
	
	public void moveByVector(float[] v){
		this.c[0] += v[0];
		this.c[1] += v[1];
		this.c[2] += v[2];
	}
	
	
	public void moveTo(float[] v){
		this.c[0] = v[0];
		this.c[1] = v[1];
		this.c[2] = v[2];
	}
	
	public HalfEdge getNeighborEdge() {
		return neighborEdge;
	}
	
	public int getUniqueNumber() {
		return uniqueNumber;
	}
	
	@Override
	public int compareTo(Vertex o) {
		return this.uniqueNumber-o.uniqueNumber;
	}
}
