// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.skeletonizer;

import java.io.*;
import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.Future;

import common.FastDeletableVector;
import common.ThreadPool;
import common.VectorOps;
import model.*;
import model.skeletonizer.Dislocation.BurgersVectorInformation;
import model.skeletonizer.processors.*;

/**
 * Creates a dislocation skeleton from given set of dislocation core atoms 
 * and additionally, if possible, a set of stacking fault planes 
 */
public class Skeletonizer {	  
	private FastDeletableVector<SkeletonNode> nodes = new FastDeletableVector<SkeletonNode>();
	private ArrayList<Dislocation> dislocations = new ArrayList<Dislocation>();
	
	private ArrayList<PlanarDefect> planarDefects = new ArrayList<PlanarDefect>();
	
	private TreeMap<Atom, PlanarDefect> planarDefectMap;
	
	private final float[] size;
	private final boolean[] pbc;
	
	private float meshingThreshold;
	
	/**
	 * Creates a dislocation skeleton from dislocation core atoms and planes of stacking faults
	 */
	public Skeletonizer(AtomData data) {
		this.pbc = Configuration.getPbc();
		this.size = data.getBounds();
		this.meshingThreshold = Configuration.getCrystalStructure().getSkeletonizationMeshingThreshold();
		
		//reinit numbering of dislocations and surfaces
		Dislocation.reinit();
		PlanarDefect.reinit();
		SkeletonNode.reinit();

		List<Atom> defectAtoms = Configuration.getCrystalStructure().getDislocationDefectAtoms(data);
		//create a initial mesh to skeletonize
		//Each atom becomes a node, edges are derived from nearest neighborhood relationship
		for (Atom a : defectAtoms) {
			nodes.add(new SkeletonNode(a));
		}
		
		//The skeletonizer does not need modifications for polycrystalline materials
		//If atoms are placed in different grains, no neighborhood will be created in "c.buildNeigh(nnb)", thus 
		//the basic algorithm works exactly the same in single- and polycrystals
		final NearestNeighborBuilder<SkeletonNode> nnb = new NearestNeighborBuilder<SkeletonNode>(data.getBounds(), meshingThreshold ,pbc);
		for (SkeletonNode c: nodes) 
			nnb.add(c);
		
		//Parallel build
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			final int j = i;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					int start = (nodes.size() * j)/ThreadPool.availProcessors();
					int end = (nodes.size() * (j+1))/ThreadPool.availProcessors();

					for (int i=start; i<end; i++)
						nodes.get(i).buildNeigh(nnb);
					return null;
				}
			});
		}
		ThreadPool.executeParallel(parallelTasks);
		
		//detect stacking faults if fcc
		if (Configuration.getCrystalStructure().hasStackingFaults()){
			TreeSet<Atom> sfAtoms = Configuration.getCrystalStructure().getStackingFaultAtoms(data);
			if (!sfAtoms.isEmpty()){
				//build nearest neighbor graph between atoms
				data.findNearestNeighbors(meshingThreshold);
				//this map will later contain the relation between atoms and the planes they are part of
				planarDefectMap = new TreeMap<Atom, PlanarDefect>();  
				this.planarDefects = PlanarDefect.createPlanarDefects(sfAtoms, planarDefectMap);
			}
		}
	}
	
	/**
	 * Constructor to directly create a skeleton (e.g. from a file)
	 * @param nodes Collection of all skeletonNodes
	 * @param dislocations Collection of all dislocations
	 * @param stackingFaults Collection of all stacking faults 
	 * @param size box dimension
	 * @param pbc Periodic boundary conditions in xyz-Direction
	 */
	public Skeletonizer(Collection<SkeletonNode> nodes, Collection<Dislocation> dislocations,
			Collection<PlanarDefect> stackingFaults, float[] size, boolean[] pbc) {
		super();
		this.nodes = new FastDeletableVector<SkeletonNode>(nodes);
		this.dislocations = new ArrayList<Dislocation>(dislocations);
		this.planarDefects = new ArrayList<PlanarDefect>(stackingFaults);
		this.size = size.clone();
		this.pbc = pbc.clone();
	}

	/**
	 * Returns all SkeletonNodes in the skeleton
	 * @return all SkeletonNodes
	 */
	public FastDeletableVector<SkeletonNode> getNodes() {
		return nodes;
	}
	
	/**
	 * Returns all identified stacking fault and twin boundary planes in the skeleton
	 * @return all stacking fault and twin boundary planes
	 */
	public ArrayList<PlanarDefect> getPlanarDefects() {
		return planarDefects;
	}
	
	/**
	 * Returns all identified dislocations in the skeleton
	 * @return all dislocations
	 */
	public ArrayList<Dislocation> getDislocations() {
		return dislocations;
	}
	
	/**
	 * Nearest neighborhood threshold distance applied to create the skeleton 
	 * @return Nearest neighborhood threshold distance
	 */
	public float getNearestNeigborDist() {
		return meshingThreshold;
	}
	
	/**
	 * Creates a skeleton with a default set of post- and preprocessors
	 */
	public void transform(){
		ArrayList<SkeletonPreprocessor> preprocessors = new ArrayList<SkeletonPreprocessor>();
		
		preprocessors.addAll(Configuration.getCrystalStructure().getSkeletonizerPreProcessors());
		
		ArrayList<SkeletonMeshPostprocessor> meshPostprocessors = new ArrayList<SkeletonMeshPostprocessor>();
		meshPostprocessors.add(new PruneProcessor());
		
		ArrayList<SkeletonDislocationPostprocessor> dislocationPostProcessors = new ArrayList<SkeletonDislocationPostprocessor>();
		dislocationPostProcessors.add(new DislocationFixingPostprocessor(meshingThreshold));
		dislocationPostProcessors.add(new DislocationSmootherPostprocessor());
		
		if (dislocations.size()==0){
			transform(preprocessors, meshPostprocessors, dislocationPostProcessors);
		}
	}
	
	
	/**
	 * Creates a skeleton with given sets of post- and preprocessors
	 * @param preprocessors Preprocessors are executed on the initial created mesh in the order in which they are inserted in the list 
	 * @param meshPostprocessors These postprocessors are executed on the final shrinked mesh, before it is transformed to dislocations.
	 * They are applied in the order in which they are inserted in the list
	 * @param dislocationPostprocessors These postprocessors are executed after the mesh is transformed to dislocations.
	 * They are applied in the order in which they are inserted in the list
	 */
	public void transform(List<SkeletonPreprocessor> preprocessors, 
			List<SkeletonMeshPostprocessor> meshPostprocessors, 
			List<SkeletonDislocationPostprocessor> dislocationPostprocessors){
		if( this.nodes.size()==0 ) return;
		
		int iterations = 0;
		boolean updated = false;
		
		//Apply pre-processors
		if (preprocessors!=null)
			for (SkeletonPreprocessor pre : preprocessors)
				pre.preProcess(this);


		//The skeletonization core process. Alternate contraction and merging until the process converged
		do {
			updated = false;
			if (contractMesh()) updated = true;
			if (mergeNodes(0.1f*meshingThreshold) && !Thread.interrupted()) updated = true;
			iterations++;
	
		} while (updated && iterations<10000 && !Thread.interrupted());
		if (Thread.interrupted()) return;
		
		//Apply mesh post-processors
		if(meshPostprocessors != null)
			for (SkeletonMeshPostprocessor post : meshPostprocessors)
				post.postProcessMesh(this);
		
		//transform thinned mesh into dislocations
		convertToDislocations();
		
		//Apply dislocation post-processors
		if(dislocationPostprocessors != null)
			for (SkeletonDislocationPostprocessor post : dislocationPostprocessors)
				post.postProcessDislocations(this);
		
		//Create links between stacking faults and dislocations
		if (!planarDefects.isEmpty()){
			for (Dislocation d : dislocations) d.findAdjacentStackingFaults(planarDefectMap);
			for (PlanarDefect s : planarDefects) s.findAdjacentDislocations(dislocations);
			planarDefectMap = null;	//Map is not needed anymore
		}
		//Store dislocation junction information in the junction nodes
		for (Dislocation d : dislocations){
			d.getStartNode().addDislocation(d);
			d.getEndNode().addDislocation(d);
		}
		
		//Analyze the skeleton and find Burgers vectors
		//TODO: Extension required to identify Burgersvectors in poly-phase material 
		//with completely different crystal structures 
		Configuration.getCrystalStructure().analyse(this);
		
		// Remove dislocation with length less than two nodes
		Iterator<Dislocation> disIter = dislocations.iterator();
		while(disIter.hasNext()){
			Dislocation d = disIter.next();
			if (d.getLine().length < 2) 
				disIter.remove();
		}
		
		if (ImportStates.KILL_ALL_ATOMS.isActive())
			for (SkeletonNode n : nodes)
				n.getMappedAtoms().clear();
	}
	
	private boolean contractMesh() {
		CyclicBarrier syncBarrier = new CyclicBarrier(ThreadPool.availProcessors());
		Vector<ContractionCallable> tasks = new Vector<ContractionCallable>();
		//Evaluate all new positions (in parallel)
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			int start = (nodes.size() * i)/ThreadPool.availProcessors();
			int end = (nodes.size() * (i+1))/ThreadPool.availProcessors();
			tasks.add(this.new ContractionCallable(start, end, nodes, syncBarrier));
		}
		
		Vector<Future<Integer>> results = ThreadPool.executeParallel(tasks);

		int nodesToMove = 0;
		//Retrieve sum of all moved nodes
		for (Future<Integer> c : results){
			try {
				nodesToMove += c.get();
			} catch (CancellationException ce){
				//do nothing, jobs have been stopped by interrupt
				return false;
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			}
		}
		
		//If less than 0.5% of all remaining nodes have moved, the mesh is considered as converged  
		return nodesToMove*200 > nodes.size();
	}
	
	private boolean mergeNodes(float mergeTol){
		mergeTol = mergeTol*mergeTol;
		boolean deleted = false;
		
		for (int j=0; j<nodes.size(); j++){
			SkeletonNode t = nodes.get(j);
			if (!t.isCriticalNode()){
				//Find the nearest neighbor for t with a common neighbor
				int nearestIndex = -1;
				float minDist = Float.POSITIVE_INFINITY;
				for (int i=0; i<t.getNeigh().size(); i++){
					SkeletonNode n = t.getNeigh().get(i); 
					if (!n.isCriticalNode()){
						float dist = Configuration.pbcCorrectedDistanceSqr(n, t); 
						if (dist < minDist && t.hasCommonNeighbor(n)){
							minDist = dist;
							nearestIndex = i;
						}
					}
				}
				//If another atom is the merging tolerance range
				//merge the node into the other and delete it afterwards
				if (minDist<mergeTol){
					t.getNeigh().get(nearestIndex).mergeSkeletonNode(t);
					deleted = true;
					nodes.remove(j);
					//During deletion, the last element in the list is copied to the
					//position of the element to be deleted. To continue at the same place in the list
					//with the shifted order, decrease the index by one.
					j--;
				}
			}
		}
		return deleted;
	}
	
	/**
	 * Transform the set of nodes into sets of polylines as dislocations
	 */
	private void convertToDislocations() {
		ArrayList<Dislocation> dislocations = new ArrayList<Dislocation>();
		TreeSet<Edge> edges = new TreeSet<Edge>();
		
		for (SkeletonNode a : nodes){
			//Center the skeletonNode
			a.centerToMappedAtoms(pbc, size);
			//get a unique set of all edges
			for (int j = 0; j < a.getNeigh().size(); j++) {
				SkeletonNode b = a.getNeigh().get(j);
				edges.add(new Edge(a, b));
			}
		}

		while (!edges.isEmpty()) {
			Edge e = edges.first();
			//Find a node with more than two neighbors -> start of a polyline
			while (e!=null && e.a1.getNeigh().size() == 2 && e.a2.getNeigh().size() == 2) {
				e = edges.higher(e);
			}
			//if e is null there are only closed simple loops in the set left
			//take the first element in the set as a random start
			if (e==null){
				e = edges.first();
			}
			// Create the dislocation as a polyline
			ArrayList<SkeletonNode> polyline = new ArrayList<SkeletonNode>();
			SkeletonNode current = null, next = null, start = null;
			if (e.a1.getNeigh().size() != 2) {
				current = e.a1;
				start = e.a1;
				next = e.a2;
			} else {
				current = e.a2;
				start = e.a2;
				next = e.a1;
			}
			polyline.add(current);
			edges.remove(e);
			
			//Traverse along the edges until another node with more than two neighbors is found
			while (next.getNeigh().size() == 2 && next!=start) {
				// Get the next edge
				SkeletonNode overnext = (current == next.getNeigh().get(0)) ? next.getNeigh().get(1) : next.getNeigh().get(0);
				current = next;
				next = overnext;
				edges.remove(new Edge(current, next));
				polyline.add(current);
			}
			//Create and store the polyline as a dislocation
			polyline.add(next);
			dislocations.add(new Dislocation(polyline.toArray(new SkeletonNode[polyline.size()])));
		}
		this.dislocations = dislocations;
	}
	
	/**
	 * 
	 * @param f
	 * @return true if the file is written without errors
	 * @throws FileNotFoundException
	 */
	public boolean writeDislocationSkeleton(File f) throws FileNotFoundException{
		PrintWriter pw = new PrintWriter(f);
		
		pw.println("#total number of nodes");
		pw.println(nodes.size());
		pw.println("#number x y z");
		for (SkeletonNode n : nodes)
			pw.println(n.getNumber()+" "+n.x+" "+n.y+" "+n.z);
		
		pw.println("#total number of dislocations");
		pw.println(dislocations.size());
		pw.println("#number numberOfNodes n_1 n_2 ... n_n BV_x BV_y BV_z BV_identified");
		pw.println("#BV_identified: (y) if Burgers vector is identified, (n) if just a numerical average is known");
		for (Dislocation d : dislocations){
			pw.print(d.getNumber()+" "+d.getLine().length+" ");
			for (SkeletonNode n : d.getLine())
				pw.print(n.getNumber()+" ");
			if (d.getBurgersVectorInfo() != null){
				BurgersVectorInformation bvInfo = d.getBurgersVectorInfo();
				if (bvInfo.getBurgersVector().getType() == BurgersVector.BurgersVectorType.UNDEFINED) {
					CrystalRotationTools crt;
					if (ImportStates.POLY_MATERIAL.isActive()) crt = d.getGrain().getCystalRotationTools();
					else crt = Configuration.getCrystalRotationTools();
						
					float[] abv = crt.getInCrystalCoordinates(bvInfo.getAverageResultantBurgersVector());
					pw.println(String.format("%.4f %.4f %.4f n", abv[0],abv[1], abv[2]));
					
				} else {
					pw.println(String.format("%.4f %.4f %.4f y", 
							bvInfo.getBurgersVector().getInCrystalCoordinates()[0],
							bvInfo.getBurgersVector().getInCrystalCoordinates()[1],
							bvInfo.getBurgersVector().getInCrystalCoordinates()[2]));
				}
			} else {
				pw.println("0.0 0.0 0.0 n");
			}
		}
		
		boolean error = pw.checkError();
		pw.close();
		return error;
	}
	
	private class Edge implements Comparable<Edge>{
		public SkeletonNode a1, a2;
		public Edge(SkeletonNode a1, SkeletonNode a2){
			if (a1.getNumber() == a2.getNumber()) 
				throw new IllegalArgumentException("Needs two unique atoms");
			if (a1.getNumber() < a2.getNumber()){
				this.a1 = a1;
				this.a2 = a2;
			} else {
				this.a2 = a1;
				this.a1 = a2;
			}
		}
		
		@Override
		public boolean equals(Object obj) {
			if (obj == null) return false;
			if (!(obj instanceof Edge)) return false;
			Edge e = (Edge)obj;
			if (this.a1.getNumber()==e.a1.getNumber() && this.a2.getNumber()==e.a2.getNumber()) return true;
			else return false;
		}
		
		@Override
		public int compareTo(Edge o) {
			if (a1.getNumber()<o.a1.getNumber()) return 1;
			if (a1.getNumber()>o.a1.getNumber()) return -1;
			if (a2.getNumber()<o.a2.getNumber()) return 1;
			if (a2.getNumber()>o.a2.getNumber()) return -1;
			return 0;
		}
	}
	
	
	private class ContractionCallable implements Callable<Integer> {
		private int start, end;
		private List<SkeletonNode> nodes;
		private CyclicBarrier syncBarrier;
		
		
		public ContractionCallable(int start, int end, List<SkeletonNode> nodes, CyclicBarrier barrier) {
			this.start = start;
			this.end = end;
			this.nodes = nodes;
			this.syncBarrier = barrier;
		}

		@Override
		public Integer call() throws Exception {
			List<SkeletonNode> nodesMoved = new ArrayList<SkeletonNode>();
			List<float[]> nodesMovedTo = new ArrayList<float[]>();
			
			float[][] dirs = new float[100][];
			float[] dirsLenght = new float[100];
			
			for (int j=start; j<end; j++){
				if (Thread.interrupted()) return null;
				SkeletonNode a = nodes.get(j);
				if (!a.isCriticalNode()) {
					//Calculate the new position with an inverse distance weighting scheme
					//Do not contract towards critical nodes where the mesh is thinned out
					//to prevent singularities
					
					int dirsSize = 0;
					float sumLength = 0f;
					
					for (SkeletonNode n : a.getNeigh()) {
						if (!n.isCriticalNode()){
							float[] dir = Configuration.pbcCorrectedDirection(n, a);
							float l = 1f/VectorOps.lengthSqr(dir);
							dirsLenght[dirsSize] = l;
							dirs[dirsSize++] = dir;
							sumLength += l;
						}
					}
					sumLength = 0.33f / sumLength;
					
					float[] move = new float[3];
					for (int i = 0; i < dirsSize; i++) {
						float[] dir = dirs[i];
						float l = dirsLenght[i] * sumLength;
						move[0] += dir[0] * l;
						move[1] += dir[1] * l;
						move[2] += dir[2] * l;
					}

					//Store new coordinate if it differs significantly from the current  
					if (Math.abs(move[0]) > 0.001f || Math.abs(move[1]) > 0.001f || Math.abs(move[2]) > 0.001f) {
						nodesMoved.add(a);
						nodesMovedTo.add(move);
					}
				}
			}
		
			syncBarrier.await();
			
			for (int i=0; i<nodesMoved.size(); i++){
				SkeletonNode n = nodesMoved.get(i);
				float[] m = nodesMovedTo.get(i);
				
				n.x += m[0];
				n.y += m[1];
				n.z += m[2];
				
				//Ensure each node is inside the box under periodic boundary conditions
				if (pbc[0]){
					if (n.x < 0f) n.x += size[0];
					else if (n.x > size[0]) n.x -= size[0];
				}
				if (pbc[1]) {
					if (n.y < 0f) n.y += size[1];
					else if (n.y > size[1]) n.y -= size[1];
				}
				if (pbc[2]) {
					if (n.z < 0f) n.z += size[2];
					else if (n.z > size[2]) n.z -= size[2];
				}
			}
			
			return nodesMoved.size();
		}
	}
}
