// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.skeletonizer.processors;

import java.util.Vector;
import java.util.concurrent.Callable;

import Jama.Matrix;
import common.ThreadPool;
import common.VectorOps;

import model.Atom;
import model.Configuration;
import model.ImportStates;
import model.skeletonizer.*;

/**
 * Mesh contraction that is based on the derived line sense of the RBV data
 * Each node to the point on the plane orthogonal to the line sense vector that minimizes
 * the distance to its neighbors. Furthermore neighbors are weighted in this process by the magnitude of the RBV.
 * Yields quickly better results compared to the standard mesh contraction scheme and is physically reasonable.
 */
public class MeshLineSenseCenteringPreprocessor implements SkeletonPreprocessor {

	@Override
	public void preProcess(Skeletonizer skel) {		
		if (!ImportStates.BURGERS_VECTORS.isActive()) return;
		if (skel.getNodes().size() == 0) return;
		
		float[][] moveTo = new float[skel.getNodes().size()][3];
		Vector<SkeletonNode> nodes = skel.getNodes();
		
		Vector<MeshLineSenseCenteringCallable> tasks = new Vector<MeshLineSenseCenteringCallable>();

		//Evaluate all new positions (in parallel)
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			int start = (nodes.size() * i)/ThreadPool.availProcessors();
			int end = (nodes.size() * (i+1))/ThreadPool.availProcessors();
			tasks.add(this.new MeshLineSenseCenteringCallable(start, end, nodes, moveTo));
		}
		ThreadPool.executeParallel(tasks);
		
		//Move all node to the new position
		for (int i=0; i<nodes.size(); i++){
			nodes.get(i).moveTo(moveTo[i][0], moveTo[i][1], moveTo[i][2]);
		}
	}
	
	
	private class MeshLineSenseCenteringCallable implements Callable<Void> {
		private int start, end;
		private Vector<SkeletonNode> nodes;
		private float[][] moveTo;
		
		
		public MeshLineSenseCenteringCallable(int start, int end, Vector<SkeletonNode> nodes, float[][] moveTo) {
			this.start = start;
			this.end = end;
			this.nodes = nodes;
			this.moveTo = moveTo;
		}

		@Override
		public Void call() throws Exception {
			double defaultLength = 1./Configuration.getCrystalStructure().getPerfectBurgersVectorLength();
			
			for (int j=start; j<end;j++){
				if (Thread.interrupted()) return null;
				SkeletonNode n = nodes.get(j);
				Atom central = n.getMappedAtoms().get(0);
				
				//Compute a orthogonal basis with the line direction as one normal
				float[] line = central.getRBV().getLineDirectionAsVector();
				float[] normalA;
				if (Math.abs(line[0]) < Math.abs(line[1]) && Math.abs(line[0]) < Math.abs(line[2])){
					normalA = new float[]{0f, line[1], -line[2]};
				} else if (Math.abs(line[1]) < Math.abs(line[2])) {
					normalA = new float[]{line[0], 0f, -line[2]};
				} else normalA = new float[]{line[0], -line[1], 0f};
				
				float[] normalB = VectorOps.crossProduct(line, normalA);
				
				//Create the system of equation, including the possible movement directions (a+b) and the
				//distance vector towards the nearest neighbors, also include the center point itself
				double[][] x = new double[n.getNeigh().size()*3 + 3][2];
				double[][] y = new double[n.getNeigh().size()*3 + 3][1];
				
				for (int i=0; i < n.getNeigh().size(); i++){
					Atom neigh = n.getNeigh().get(i).getMappedAtoms().get(0);
					
					double w = Math.sqrt(neigh.getRBV().getSquaredRBVLengh())*defaultLength;
					
					x[i*3+0][0] = normalA[0] * w; x[i*3+0][1] = normalB[0] * w;
					x[i*3+1][0] = normalA[1] * w; x[i*3+1][1] = normalB[1] * w;
					x[i*3+2][0] = normalA[2] * w; x[i*3+2][1] = normalB[2] * w;
					
					float[] dir = Configuration.pbcCorrectedDirection(n.getNeigh().get(i), n);
					y[i*3+0][0] = dir[0];
					y[i*3+1][0] = dir[1];
					y[i*3+2][0] = dir[2];
				}
				
				x[n.getNeigh().size()*3+0][0] = normalA[0]; x[n.getNeigh().size()*3+0][1] = normalB[0];
				x[n.getNeigh().size()*3+1][0] = normalA[1]; x[n.getNeigh().size()*3+1][1] = normalB[1];
				x[n.getNeigh().size()*3+2][0] = normalA[2]; x[n.getNeigh().size()*3+2][1] = normalB[2];
				
				y[n.getNeigh().size()*3+0][0] = 0f;
				y[n.getNeigh().size()*3+1][0] = 0f;
				y[n.getNeigh().size()*3+2][0] = 0f;
				
				//Solve equation by least square fitting and move the atom to the coordinate that minimizes the
				//average distance
				Matrix xM = new Matrix(x);
				Matrix yM = new Matrix(y);
				
				Matrix b = xM.solve(yM);
				
				moveTo[j][0] = (float)(n.getX() + b.get(0, 0)*normalA[0] + b.get(1, 0)*normalB[0]);
				moveTo[j][1] = (float)(n.getY() + b.get(0, 0)*normalA[1] + b.get(1, 0)*normalB[1]);
				moveTo[j][2] = (float)(n.getZ() + b.get(0, 0)*normalA[2] + b.get(1, 0)*normalB[2]);
			}
			
			return null;
		}
	}
}
